/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#ifndef NODE_H
#define NODE_H 1

#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>

#include "ipaddress.h"
#include "net_cable.h"
#include "variable_tree.h"

using namespace std;

class Node
/** abstrakte Basisklasse f&uuml;r alle Netzknoten
 *  (Net_Switch, L3_Router, Host) */
{
	protected:

	public:
		Node(void);
		virtual ~Node(void);

		/** stellt den Variablenbaum f&uuml;r das Dialogfenster
		 *  zusammen, Zeiger darauf in 'infos' zur&uuml;ck */
		virtual void	getInfos(Variable_Tree* &infos) = 0;
		/** liest die Werte aus dem Variablenbaum ein */
		virtual void	setInfos(Variable_Tree* infos) = 0;
		/** l&auml;dt die Einstellungen und Pakete der Knoten aus
		 *  der XML-Datei 'infile' */
		virtual void	loadXML(xmlTextReaderPtr infile) = 0;
		/** speichert die Einstellungen und Pakete der Knoten in die
		 *  XML-Datei 'outfile' */
		virtual void	saveXML(xmlTextWriterPtr outfile) const = 0;

		//@{ @name Knotenfunktionen
		/** das Netzwerkkabel 'new_cable' soll mit dem Knoten
		 *  verbunden werden, damit dar&uuml;ber Pakete versandt werden
		 *  k&ouml;nnen */
		virtual void	attachCable(Net_Cable *new_cable) = 0;
		/** alle angeschlossenen Kabel entfernen, zur Neustrukturierung
		 *  des Netzwerkes nach &Auml;nderung des Aufbaus */
		virtual void	detachCables(void) = 0;
		// returns true if there is a packet in one of the cards/queues
		/** gibt die Position des 'untersten' (ersten) Paketes
		 *  zur&uuml;ck, zur Fokussierung in der Animation;
		 *  'true' wenn ein Paket vorhanden ist, sonst 'false' */
		virtual bool	getFirstPacketPos(
					Point_3D &packet_pos) const = 0;
		/** f&uuml;hrt ein Schritt der Netzwerksimulation im Knoten
		 *  aus, 'speed_factor' beeinflusst die Geschwindigkeit der
		 *  Drehung und des Pakettransports */
		virtual void	executeNet(double speed_factor) = 0;
		/** importiert Paket in die Simulation, d.h. es kommt in die
		 *  entsprechende Warteschlange zum Senden */
		virtual void	executeNetImportPacket(
					Packet* cable_packet) = 0;
		/** gibt den Namen des Knotens zur&uuml;ck */
		virtual void	getName(string &node_name) const = 0;
		/** gibt 'true' zur&uuml;ck, wenn eine der Netzwerkkarten
		 *  des Knotens die IP-Adresse 'searched_address' hat, sonst
		 *  'false' */
		virtual bool	hasIpAddress(
					IpAddress &searched_address) const = 0;
		/** gibt 'true' zur&uuml;ck, wenn eine der Netzwerkkarten oder
		 *  Warteschlangen (bei Switches) des Knotens die
		 *  Ethernet-Adresse 'searched_address' hat, sonst 'false' */
		virtual bool	hasMacAddress(
					MacAddress &searched_address) const = 0;
		// return if (one) interface has to be turned
		/** gibt 'true' zur&uuml;ck, wenn eine der Netzwerkkarten des
		 *  Knoten eine Drehung anfordert, damit sie in die richtige
		 *  Richtung absenden/empfangen kann; sonst 'false'; <br>
		 *  zur Abstimmung zwischen den Kabeln und Knoten in der
		 *  Animation */
		virtual bool	needsTurn(void) const = 0;
		//@}

		//@{ @name Position, Winkel, Farbe, Gr&ouml;&szlig;e
		/** gibt die aktuelle Position des Knoten zur&uuml;ck
		 *  (Mittelpunkt des K&ouml;rpers) */
		virtual void	getPos(double &current_centerx, 
					double &current_centery,
					double &current_centerz) const = 0;
		/** gibt die aktuelle Position in 'current_pos' und den
		 *  Rotationswinkel in 'current_angle' zur&uuml;ck */
		virtual void	getPosAngle(Point_3D &current_pos,
					double &current_angle) const = 0;
		/** setzt die Farbe, mit der der Knoten dargestellt werden
		 *  soll, Rot / Gr&uuml;n / Blau / Alpha(Transparenz) im
		 *  Bereich 0.0 - 1.0 */
		virtual void	setColor(double new_red,  double new_green, 
						double new_blue,
						double new_alpha) = 0;
		/** setzt die Position des Knoten
		 *  (Mittelpunkt des K&ouml;rpers)*/
		virtual void	setPos(double new_centerx, double new_centery,
					double new_centerz) = 0;
		/** setzt die Gr&ouml;&szlig;e des Knoten,
		 *  Breite / H&ouml;he / Tiefe als Bounding-Box */
		virtual void	setSize(double new_width, double new_height,
					double new_depth) = 0;
		//@}

		/** zeichnet den Knoten */
		virtual void	draw(void) const = 0;
		/** &uuml;bergibt dem Knoten den aktuellen Zustand von
		 *  Level of Detail (an/aus) 'level_of_details_on' und die
		 *  aktuelle Kameraposition 'camera_pos',
		 *  Knoten zeichnet die internen Displaylisten neu, welche
		 *  abh&auml;ngig von Level of Detail sind */
		virtual void	drawSetCameraPos(bool level_of_details_on,
						Point_3D camera_pos) = 0;
};

#endif
