/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <string>
#include <vector>

#ifndef VARIABLE_H
#define VARIABLE_H 1

using namespace std;

class Variable
/** abstrakte Basisklasse zur Speicherung einer Variable mit Name und Wert,<br>
 *  zur &Uuml;bergabe von einem Programm-Modul an andere Module, welche die
 *  Schnittstelle des einen Moduls (einstellbare Werte und deren Datentypen)
 *  nicht kennen */
{
	protected:
		/** zeigt an, ob der Variablenwert konstant ist,
		 *  also unver&auml;nderbar */
		bool	constant;
		/** der Name der Variablen, zur Identifizierung */
		string	name;
	public:
		Variable(void);
		virtual ~Variable(void);
		/** zeigt an, ob der Wert der Variable konstant ist, also nicht
		 *  mehr &auml;nderbar mit setValue() */
		virtual	bool	isConstant(void) = 0;
		/** gibt den Namen der Variable zur&uuml;ck */
		virtual string	getName(void) = 0;
		/** gibt den Wert der Variablen als String zur&uuml;ck */
		virtual string	getValue(void) = 0;
		/** setzt den Wert der Variable auf konstant, nicht mehr
		 *  &auml;nderbar mit setValue() */
		virtual void	setConstant(bool is_constant) = 0;
		/** setzt den Namen der Variable auf 'new_name' */
		virtual void	setName(string new_name) = 0;
		/** setzt den Wert der Variable auf den in 'new_value' */
		virtual bool	setValue(string new_value) = 0;
};

template <class ValueType>
class Numeric_Variable : public Variable
/** zur Speicherung eines Zahlenwertes oder eines Feldes von Zahlen,
 *  der Wertebereich und die maximale Feldl&auml;nge ist festlegbar, <br>
 *  ValueType sollte mit Zuweisungs- (=), Vergleichs- (<=, >=) und
 *  Streamoperatoren (<<, >>) nutzbar sein */
{
	private:
		/** minimale (min_length) und maximale (max_length)
		 *  Feldl&auml;nge, also Anzahl von Elementen im Feld */
		size_t			min_length, max_length;
		/** minimaler (min_value) und maximaler (max_value) Wert eines
		 *  Elementes dieser Variable */
		ValueType		min_value, max_value;
		/** Feld vom Datentyp 'ValueType' mit Elementen der Variable */
		vector<ValueType>	value;
	public:
		Numeric_Variable(void);
		~Numeric_Variable(void);
		bool	isConstant(void);
		string	getName(void);
		void	getValue(size_t index, ValueType &current_value);
		string	getValue(void);
		void	setConstant(bool is_constant);
		bool	setLength(size_t new_max, size_t new_min = 0);
		void	setName(string new_name);

		//@{ @name Wert(e) und Wertebereich setzen
		bool	setMaxmin(ValueType new_max, ValueType new_min);
		bool	setValue(ValueType new_value);
		bool	setValue(size_t index, ValueType new_value);
		bool	setValue(vector<ValueType> new_value);
		bool	setValue(string new_value);
		//@}

};

template <class ValueType>
ostream& operator << (ostream& output,
			Numeric_Variable<ValueType> &var);


class String_Variable : public Variable
/** zur Speicherung eines Textfeldes (string) vorgegebener L&auml;nge */
{
	private:
		/** minimale (min_length) und maximale (max_length)
		 *  Feldl&auml;nge, also Anzahl von Zeichen des Textes */
		size_t	min_length, max_length;
		/** das Textfeld */
		string	value;
	public:	
		String_Variable(void);
		~String_Variable(void);
		bool	isConstant(void);
		string	getName(void);
		string	getValue(void);
		void	setConstant(bool is_constant);
		bool	setLength(size_t new_max, size_t new_min = 0);
		void	setName(string new_name);
		bool	setValue(string new_value);
		friend ostream& operator << (ostream& output, 
						String_Variable &var);
};

#include "../src/variable_template.cpp"

#endif

