/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <math.h>

#include "bezier.h"

Bezier::Bezier(void)
/** Konstruktor, ist leer */
{
};

Bezier::~Bezier(void)
/** Destruktor, ist leer */
{
};

unsigned int Bezier::factorial(unsigned int x)
/** berechnet die Fakult&auml;t von x, also x!, und gibt das Ergebnis
 *  zur&uuml;ck */
{
	unsigned int	i, result;

	result = 1;
	for(i = x; i > 0; i--) result *= i;
	return result;
};

unsigned int Bezier::binomialCoefficient(unsigned int n, unsigned int k)
/** berechnet den Binomialkoeffizienten 'n &uuml;ber k' = n! / (k! * (n - k)!)
 *  und gibt das Ergebnis zur&uuml;ck */
{
	unsigned int	result;

	result = factorial(n) / (factorial(k) * factorial(n - k));
	return result;
};

double Bezier::bernstein(unsigned int n, unsigned int i, double t)
/** berechnet das Bernsteinpolynom(n, i, t) =
 *  (n &uuml;ber i) * t^i * (1 - t)^(n - i) und gibt das Ergebnis zur&uuml;ck
 */
{
	double	result;

	result = binomialCoefficient(n, i) * pow(t, i * 1.0) *
			pow(1 - t, n - i * 1.0);
	return result;
};

void Bezier::calcBezier(vector<Point_3D> &control_points,
			size_t num_curve_points,
			vector<Point_3D> &curve_points)
/** berechnet die Bezierkurve aus den Hilfspunkten 'control_points',
 *  zur&uuml;ckgegeben werden eine Anzahl von 'num_curve_points' Punkten
 *  im Feld 'curve_point' entlang dieser berechneten Kurve, <br>
 *  dabei ist der erste Kurvenpunkt gleich dem ersten Hilfspunkt und der
 *  letzte Kurvenpunkt gleich dem letzten Hilfspunkt, <br>
 *  alle Hilfspunkte sollten in der gleichen horizontalen Ebene liegen, es
 *  werden nur die X- und Z-Werte der Punkte zur Berechnung herangezogen,
 *  die berechnete Kurve liegt in der gleichen horizontalen Ebene, <br>
 *  die Farbwerte des ersten Kontrollpunktes werden f&uuml;r alle Kurvenpunkte
 *  &uuml;bernommen (genauso wie dessen H&ouml;henlage, s.o.) */
{
	size_t	i, j;
	double	result_bernstein, curve_part, x, y, z;

	// resize output array
	curve_points.resize(num_curve_points, Point_3D());
	for(j = 0; j < num_curve_points; j++)
	{
		// on which part of curve are we
		curve_part = j * 1.0 / num_curve_points;
		x = 0.0;
		y = control_points[0].getY();	// all points same height
		z = 0.0;
		// calculate the curve point (x,z) out of all control points
		for(i = 0; i < control_points.size(); i++)
		{
			result_bernstein = bernstein(control_points.size() - 1,
							 i, curve_part);
			x += control_points[i].getX() * result_bernstein;
			z += control_points[i].getZ() * result_bernstein;
		};
		// set point to result, take color of control points
		curve_points[j].setPos(x, y, z);
		curve_points[j].setColor(control_points[0].getR(),
						control_points[0].getG(),
						control_points[0].getB(),
						control_points[0].getA());
		// cout << "Bezier::calcBezier: " << curve_points[j] << endl;
	};
	// first/last curve point is same as first/last control point
	curve_points[0].setPos(0, 0, 0);
	curve_points[0] += control_points[0];
	curve_points[num_curve_points - 1].setPos(0, 0, 0);
	curve_points[num_curve_points - 1] +=
				control_points[control_points.size() - 1];
};

