/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <string.h>

#include "checksums.h"

Checksums::Checksums()
/** Konstruktor, ist leer */
{
};

Checksums::~Checksums()
/** Destruktor, ist leer */
{
};

// taken from RFC 1071
u_int16_t Checksums::calcChecksum(const unsigned char *buffer, 
					const size_t num_bytes)
/** berechnet die Pr&uuml;fsumme aus dem Feld "buffer" der L&auml;nge
 *  'num_bytes' und gibt sie zur&uuml;ck, 
 *  Berechnung erfolgt nach dem (f&uuml;r alle Internet-Protokolle verwendeten)
 *  Algorithmus aus RFC 1071: das Einerkomplement der Summe aller
 *  Einerkomplemente der 16-Bit-Elemente des Feldes */
{
	u_int16_t	return_sum;
	size_t		i;
	unsigned int	sum;

	sum = 0;
	for(i = 0; i < num_bytes/2; i++)
	{
		sum += (buffer[i * 2 + 1] << 8) + buffer[i * 2];
	};
	if ((num_bytes % 2) == 1) sum += buffer[num_bytes - 1];
	while (sum >> 16) sum = (sum & 0xffff) + (sum >> 16);
	return_sum = ~sum;
	return return_sum;
};

// RFC 791 (IPv4)
void Checksums::updateIpChecksum(struct ip* const ip_header)
/** berechnet die Pr&uuml;fsumme des IPv4-Kopfes 'ip_header' und tr&auml;gt sie
 *  dort ein */
{
	u_int16_t	sum;

	ip_header->ip_sum = 0;
	sum = calcChecksum(reinterpret_cast<unsigned char*>(ip_header),
				sizeof(ip_header));
	ip_header->ip_sum = sum;
};

// RFC 793 (TCP)
void Checksums::updateTcpChecksum(const struct ip* const ip_header, 
					struct tcphdr* const tcp_header)
/** berechnet die Pr&uuml;fsumme des TCP-Kopfes 'tcp_header' im IP-Paket
 *  mit dem IPv4-Kopf 'ip_header' (wird zur Berechnung ben&ouml;tigt) und
 *  tr&auml;gt sie im TCP-Kopf ein */
{
	const size_t	pseudo_tcp_size = 12 + sizeof(tcp_header);
	u_int16_t	sum, tcp_length;
	unsigned char	buffer_pseudo_and_tcp_header[ pseudo_tcp_size ];

	tcp_header->check = 0;
	// concat pseudo header out of ip and tcp header
	memset(buffer_pseudo_and_tcp_header, 0, pseudo_tcp_size);
	memcpy(buffer_pseudo_and_tcp_header, &(ip_header->ip_src),
		sizeof(in_addr_t));
	memcpy(&(buffer_pseudo_and_tcp_header[sizeof(in_addr_t)]),
		&(ip_header->ip_dst), sizeof(in_addr_t));
	memcpy(&(buffer_pseudo_and_tcp_header[2 * sizeof(in_addr_t) + 1]),
		&(ip_header->ip_p), sizeof(ip_header->ip_p));
	tcp_length = htons(sizeof(tcp_header));
	memcpy(&(buffer_pseudo_and_tcp_header[2 * sizeof(in_addr_t) + 2]),
		&tcp_length, sizeof(tcp_length));
	memcpy(&(buffer_pseudo_and_tcp_header[2 * sizeof(in_addr_t) + 4]),
		tcp_header, sizeof(tcp_header));

	sum = calcChecksum(buffer_pseudo_and_tcp_header, pseudo_tcp_size);	
	tcp_header->check = sum;
};

// RFC 768 (UDP)
void Checksums::updateUdpChecksum(const struct ip* const ip_header,
					struct udphdr* const udp_header)
/** berechnet die Pr&uuml;fsumme des UDP-Kopfes 'udp_header' im IP-Paket
 *  mit dem IPv4-Kopf 'ip_header' (wird zur Berechnung ben&ouml;tigt) und
 *  tr&auml;gt sie im UDP-Kopf ein */
{
	const size_t	pseudo_udp_size = 12 + sizeof(udp_header);
	u_int16_t	sum;
	unsigned char	buffer_pseudo_and_udp_header[ pseudo_udp_size ];

	udp_header->check = 0;
	// concat pseudo header out of udp and ip header
	memset(buffer_pseudo_and_udp_header, 0, pseudo_udp_size);
	memcpy(buffer_pseudo_and_udp_header, &(ip_header->ip_src),
		sizeof(in_addr_t));
	memcpy(&(buffer_pseudo_and_udp_header[sizeof(in_addr_t)]),
		&(ip_header->ip_dst), sizeof(in_addr_t));
	memcpy(&(buffer_pseudo_and_udp_header[2 * sizeof(in_addr_t) + 1]),
		&(ip_header->ip_p), sizeof(ip_header->ip_p));
	memcpy(&(buffer_pseudo_and_udp_header[2 * sizeof(in_addr_t) + 2]),
		&(udp_header->len), sizeof(udp_header->len));
	memcpy(&(buffer_pseudo_and_udp_header[2 * sizeof(in_addr_t) + 4]),
		udp_header, sizeof(udp_header));

	sum = calcChecksum(buffer_pseudo_and_udp_header, pseudo_udp_size);
	udp_header->check = sum;
};

// RFC 792 (ICMP)
void Checksums::updateIcmpChecksum(struct icmp* const icmp_header)
/** ungetestet: berechnet die Pr&uuml;fsumme des ICMP-Kopfes und tr&auml;gt
 *  sie dort ein */
{
	u_int16_t	sum;

	icmp_header->icmp_cksum = 0;
	sum = calcChecksum(reinterpret_cast<unsigned char*>(icmp_header),
				sizeof(icmp_header));
	icmp_header->icmp_cksum = sum;
	// with internet header (like defined in RFC 792) ?
};

// RFC 2236 (IGMP v2)
void Checksums::updateIgmpChecksum(struct igmp* const igmp_header)
/** ungetestet: berechnet die Pr&uuml;fsumme des IGMPv2-Kopfes und tr&auml;gt
 *  sie dort ein */
{
	u_int16_t	sum;

	igmp_header->igmp_cksum = 0;
	sum = calcChecksum(reinterpret_cast<unsigned char*>(igmp_header),
				sizeof(igmp_header)); // ??!
	igmp_header->igmp_cksum = sum;
};

