/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"
#include "win_net.h"

#include <iostream>
#include <stdio.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <sys/socket.h>

#include "checksums.h"
#include "constants.h"
#include "packet_client.h"

Packet_Client::Packet_Client(void) :
		is_connected(false), client_socket(-1),
		server_port(0), server_ip(INADDR_ANY), service_name("")
/** Konstruktor, Initialisierung der Variablen: is_connected, client_socket,
 *  server_port, server_ip, service_name */
{
};

Packet_Client::~Packet_Client(void)
/** Konstruktor, ruft nur close() zum Schlie&szlig;en der Verbindung zum
 *  Server, falls sie noch offen ist */
{
	close();
};

int Packet_Client::init(const char new_server_ip[], 
			unsigned short new_server_port)
/** Initialisierung: &uuml;bernimmt IP-Adresse in 'new_server_ip' und
 *  Portnummer 'new_server_port' als Verbindungseinstellungen zum Server,
 *  Paketstrukturen werden mit '0' initialisiert, ether_type im Ethernet-Kopf
 *  auf ETHERTYPE_IP gesetzt, <br>
 *  gibt '0' zur&uuml;ck, wenn kein Fehler auftrat,
 *  sonst '-1', wenn der Text in 'new_server_ip' nicht in eine IP-Adresse
 *  umgewandelt werden konnte */
{
	int		error;
	struct in_addr	converted_server_ip;

	error = 0;
	if (new_server_ip)	// text in 'new_server_ip' to address
		if (inet_aton(new_server_ip, &converted_server_ip) == 0)
		{
			cout << "Packet_Client: wrong server_ip format."
				<< endl;
			error = -1;
		}
		else server_ip = converted_server_ip.s_addr;
	server_port = new_server_port;
	// init fill with '0'
	memset(&eth_header, 0, sizeof(eth_header));
	memset(&ip_header, 0, sizeof(ip_header));
	memset(&tcp_header, 0, sizeof(tcp_header));
	memset(&udp_header, 0, sizeof(udp_header));
	eth_header.ether_type = htons(ETHERTYPE_IP);
	return error;
};

int Packet_Client::connect(void)
/** Verbindung zum Server aufnehmen, eine bestehende Verbindung wird vorher
 *  durch close() beendet, <br>
 *  IP-Adresse und Portnummer des Servers sollten vorher per init()
 *  eingestellt werden, <br>
 *  R&uuml;ckgabe '0' wenn fehlerfrei, '-1' bei Fehlern (Verbindung
 *  fehlgeschlagen) */
{
	int			error;
	struct sockaddr_in	socket_address;

	error = 0;
	close();	// if older connection is open.
	#ifdef WIN32
	init_socket();
	#endif
	client_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (client_socket == -1)
	{
		error = -1;
		cout << "Packet_Client: error opening socket." << endl;
	};
	if (!error)
	{
		memset(&socket_address, 0, sizeof(socket_address));
		socket_address.sin_family = AF_INET;
		socket_address.sin_addr.s_addr = server_ip;
		socket_address.sin_port = htons(server_port);
		error = ::connect(client_socket,
				reinterpret_cast<sockaddr*>(&socket_address),
				sizeof(socket_address));
		if (error) cout << "Packet_Client: error connecting." << endl;
	};
	if (!error)
	{
		cout << "Packet_Client: Connection successful." << endl;
		is_connected = true;
	};
	return error;
};

int Packet_Client::sendServiceName(void) const
/** sendet den aktuell eingestellten Dienstnamen zum Server, sollte vor jedem
 *  Paket (dieses Dienstes) geschehen, <br>
 *  gibt '-1' bei Fehler zur&uuml;ck, sonst '0', siehe dazu sendRawPacket() */
{
	int	error;
	char	packet_buffer[service_name_length];

	error = 0;
	memset(packet_buffer, 0, service_name_length);
	snprintf(packet_buffer, service_name_length,
			service_name_format_string, service_name.c_str());
	if (!error) error = sendRawPacket(packet_buffer, service_name_length);
	return error;
};

int Packet_Client::sendRawPacket(const char *packet_buffer,
					size_t packet_size) const
/** sendet das Paket im Puffer 'packet_buffer' mit der Gr&ouml;&szlig;e
 *  in Bytes 'packet_size' &uuml;ber eine bestehende Verbindung zum Server,<br>
 *  liefert '-1' bei Fehler (keine Verbindung / Senden fehlgeschlagen),
 *  sonst '0' wenn fehlerfrei */
{
	int	error, num_send;

	error = 0;
	if ((client_socket > 0) && (is_connected))
	{
		num_send = send(client_socket, packet_buffer,
				packet_size, 0);
		if (num_send != static_cast<int>(packet_size)) error = -1;
	}
	else error = -1;
	if (error) cout << "Packet_Client: error sending packet." << endl;
	return error;
};

int Packet_Client::sendRawWithService(const char *packet_buffer,
					size_t packet_size) const
/** sendet den aktuell eingestellten Dienstnamen und das Paket im Puffer
 *  'packet_buffer' mit der Gr&ouml;&szlig;e in Bytes 'packet_size'
 *  &uuml;ber eine bestehende Verbindung zum Server,<br>
 *  liefert '-1' bei Fehler (keine Verbindung / Senden fehlgeschlagen),
 *  sonst '0' wenn fehlerfrei, siehe sendServiceName() und sendRawPacket() */
{
	int	error;

	error = 0;
	if (!error) error = sendServiceName();
	if (!error) error = sendRawPacket(packet_buffer, packet_size);
	return error;
};

int Packet_Client::sendTcpPacket(void)
/** sendet Dienstname und ein TCP-Paket inklusive IP- und Ethernet-Kopf an den
 *  Server, Pr&uuml;fsummen werden berechnet und ins Paket eingetragen; <br>
 *  IP-Adressen und Portnummern des Paketes sollten vorher eingestellt werden,
 *  siehe setSourceIpAddress(), setDestinationIpAddress(), setSourcePort(),
 *  setDestinationPort(), setIpHeader(), setTcpHeader() und setTcpFlags(); <br>
 *  liefert '-1' bei Fehler (keine Verbindung / Senden fehlgeschlagen),
 *  sonst '0' wenn fehlerfrei, siehe sendServiceName() und sendRawPacket() */
{
	const size_t	packet_size = sizeof(ether_header) + sizeof(ip) +
					sizeof(tcphdr);
	int		error;
	char		packet_buffer[ packet_size ];

	error = 0;
	// fill correct values in packet
	ip_header.ip_p = IPPROTO_TCP;
	ip_header.ip_len = htons(sizeof(ip) + sizeof(tcphdr));
	updateIpChecksum();
	updateTcpChecksum();

	// concat data in structures to packet
	memset(packet_buffer, 0, packet_size);
	memcpy(packet_buffer, &eth_header, sizeof(eth_header));
	memcpy(&(packet_buffer[ sizeof(eth_header)] ),
		&ip_header, sizeof(ip_header));
	memcpy(&(packet_buffer[ sizeof(eth_header) + sizeof(ip_header) ]),
		&tcp_header, sizeof(tcp_header));

	// send packet to server
	if (!error) error = sendServiceName();
	if (!error) error = sendRawPacket(packet_buffer, packet_size);
	return error;
};

int Packet_Client::sendUdpPacket(void)
/** sendet ein UDP-Paket inklusive IP- und Ethernet-Kopf an den Server,
 *  Pr&uuml;fsummen werden berechnet und ins Paket eingetragen; <br>
 *  IP-Adressen und Portnummern des Paketes sollten vorher eingestellt werden,
 *  siehe setSourceIpAddress(), setDestinationIpAddress(), setSourcePort() und
 *  setDestinationPort(); <br>
 *  liefert '-1' bei Fehler (keine Verbindung / Senden fehlgeschlagen),
 *  sonst '0' wenn fehlerfrei, siehe sendServiceName() und sendRawPacket() */
{
	const size_t	packet_size = sizeof(ether_header) + sizeof(ip) +
					sizeof(udphdr);
	int		error;
	char		packet_buffer[ packet_size ];

	error = 0;
	// fill correct values in packet
	ip_header.ip_p = IPPROTO_UDP;
	ip_header.ip_len = htons(sizeof(ip) + sizeof(udphdr));
	updateIpChecksum();
	updateUdpChecksum();

	// concat data in structures to packet
	memset(packet_buffer, 0, packet_size);
	memcpy(packet_buffer, &eth_header, sizeof(eth_header));
	memcpy(&(packet_buffer[ sizeof(eth_header)] ),
		&ip_header, sizeof(ip_header));
	memcpy(&(packet_buffer[ sizeof(eth_header) + sizeof(ip_header) ]),
		&udp_header, sizeof(udp_header));

	// send packet to server
	if (!error) error = sendServiceName();
	if (!error) error = sendRawPacket(packet_buffer, packet_size);
	return error;
};

int Packet_Client::close(void)
/** schlie&szlig;t die Verbindung zum Server, falls eine besteht, <br>
 *  gibt '0' zur&uuml;ck, wenn fehlerfrei, sonst '-1' bei Fehler,
 *  close() ohne connect() (d.h. ohne bestehende Verbindung) ist kein Fehler */
{
	int	error;

	error = 0;
	if (client_socket > 0) error = ::close(client_socket);
	if (error) cout << "Packet_Client: Error closing." << endl;
	client_socket = -1;
	is_connected = false;
	return error;
};

void Packet_Client::updateIpChecksum(void)
/** berechnet die Pr&uuml;fsumme im IP-Kopf neu */
{
	Checksums::updateIpChecksum(&ip_header);
};

int Packet_Client::setSourceIpAddress(const char new_source_ip[])
/** setzt die Adresse im Text 'new_source_ip' als Quelladresse in den IP-Kopf
 *  ein, Pr&uuml;fsumme wird neu berechnet, <br>
 *  gibt '0' zur&uuml;ck, wenn kein Fehler auftrat,
 *  sonst '-1', wenn der Text in 'new_source_ip' nicht in eine IP-Adresse
 *  umgewandelt werden konnte */
{
	int		error;
	struct in_addr	converted_source_ip;

	error = 0;
	if (new_source_ip)
		if (inet_aton(new_source_ip, &converted_source_ip) == 0)
		{
			cout << "Packet_Client: wrong source_ip format."
				<< endl;
			error = -1;
		}
		else ip_header.ip_src.s_addr = converted_source_ip.s_addr;
	return error;
	updateIpChecksum();
};

int Packet_Client::setDestinationIpAddress(const char new_destination_ip[])
/** setzt die Adresse im Text 'new_destination_ip' als Zieladresse in den
 *  IP-Kopf ein, Pr&uuml;fsumme wird neu berechnet, <br>
 *  gibt '0' zur&uuml;ck, wenn kein Fehler auftrat,
 *  sonst '-1', wenn der Text in 'new_destination_ip' nicht in eine IP-Adresse
 *  umgewandelt werden konnte */
{
	int		error;
	struct in_addr	converted_destination_ip;

	error = 0;
	if (new_destination_ip)
		if (inet_aton(new_destination_ip,
			&converted_destination_ip) == 0)
		{
			cout << "Packet_Client: wrong destination_ip format."
				<< endl;
			error = -1;
		}
		else ip_header.ip_dst.s_addr = converted_destination_ip.s_addr;
	return error;
	updateIpChecksum();
};

void Packet_Client::setIpHeader(u_int8_t tos, u_int16_t id, u_int8_t ttl)
/** setzt die restlichen Werte (ohne die Quell-/Zieladressen) im IP-Kopf,
 *  Pr&uuml;fsumme wird neu berechnet, <br>
 *  der Wert von 'tos' kommt ins 'Type of Service'-Feld, 'id' ins
 *  'Identification'-Feld, 'ttl' ins Feld 'time to live' */
{
	ip_header.ip_v = IPVERSION;
	ip_header.ip_hl = 5;
	ip_header.ip_tos = tos;
	ip_header.ip_len = htons(20);
	ip_header.ip_id = htons(id);
	ip_header.ip_off = 0;
	ip_header.ip_ttl = ttl;
	ip_header.ip_p = IPPROTO_TCP;
	updateIpChecksum();
};

// RFC 793 (TCP)
void Packet_Client::updateTcpChecksum()
/** berechnet die Pr&uuml;fsumme im TCP-Kopf neu */
{
	Checksums::updateTcpChecksum(&ip_header, &tcp_header);
};

// RFC 768 (UDP)
void Packet_Client::updateUdpChecksum()
/** setzt die L&auml;ngenangabe im UDP-Kopf auf '8' (L&auml;nge des Kopfes)
 *  und berechnet die Pr&uuml;fsumme im UDP-Kopf neu */
{
	udp_header.len = htons(8);
	Checksums::updateUdpChecksum(&ip_header, &udp_header);
};

void Packet_Client::setSourcePort(in_port_t new_source_port)
/** setzt den Quellport 'new_source_port' im TCP- und im UDP-Kopf,
 *  Pr&uuml;fsummen werden neu berechnet */
{
	tcp_header.source = htons(new_source_port);
	udp_header.source = htons(new_source_port);
	updateTcpChecksum();
	updateUdpChecksum();
};

void Packet_Client::setDestinationPort(in_port_t new_destination_port)
/** setzt den Zielport 'new_destination_port' im TCP- und im UDP-Kopf,
 *  Pr&uuml;fsummen werden neu berechnet */
{
	tcp_header.dest = htons(new_destination_port);
	udp_header.dest = htons(new_destination_port);
	updateTcpChecksum();
	updateUdpChecksum();
};

void Packet_Client::setTcpHeader(u_int32_t seq, u_int32_t ack,
				 u_int16_t window, u_int16_t urgent_pointer)
/** setzt die restlichen Werte (ohne die Quell-/Zielports) im TCP-Kopf,
 *  Pr&uuml;fsumme wird neu berechnet, <br>
 *  der Wert von 'seq' kommt ins Feld 'Sequenznummer', 'ack' zu
 *  'Acknowledgement-Nummer', 'window' zu 'window size' und 'urgent_pointer'
 *  ins Feld 'urgent pointer' */
{
	tcp_header.seq = htonl(seq);
	tcp_header.ack_seq = htonl(ack);
	tcp_header.res1 = 0;
	tcp_header.doff = 5;
	tcp_header.res2 = 0;
	tcp_header.window = htons(window);
	tcp_header.urg_ptr = htons(urgent_pointer);
	updateTcpChecksum();
};

void Packet_Client::setTcpFlags(bool fin, bool sync, bool reset, bool push,
				bool ack, bool urgent)
/** setzt die TCP-Flags im TCP-Kopf, Pr&uuml;fsumme wird neu berechnet,
 *  'fin' f&uuml;r 'finish', 'sync' f&uuml;r 'synchronize', 'reset', 'push'
 *  'ack' f&uuml;r 'acknowledgement', 'urgent' f&uuml;r
 *  'urgent pointer gesetzt' */
{
	tcp_header.fin = fin;
	tcp_header.syn = sync;
	tcp_header.rst = reset;
	tcp_header.psh = push;
	tcp_header.ack = ack;
	tcp_header.urg = urgent;
	updateTcpChecksum();
};

void Packet_Client::setServiceName(const char new_service_name[])
/** setzt den Dienstnamen, der dann mit sendRawWithService(), sendTcpPacket()
 *  oder sendUdpPacket() gesendet wird, <br>
 *  Server (TcpQueue / UdpQueue::findServicePort()) wandelt ihn in Portnummer
 *  um, wenn Quellport == 0 */
{
	service_name = new_service_name;
};

