/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <iostream>

#include "arp_packet.h"
#include "constants.h"
#include "net_queue.h"
#include "packet_loader_xml.h"
#include "tcp_service_packet.h"
#include "udp_service_packet.h"

Packet_Loader_XML::Packet_Loader_XML(void) :
	packet_size(0), service_name(""), infile(NULL)
/** Konstruktor, Initialisierung der Variablen: packet_size, service_name,
 *  infile */
{
};

Packet_Loader_XML::~Packet_Loader_XML(void)
/** Destruktor, ruft nur closeFile() zum Schlie&szlig;en einer eventuell
 *  ge&ouml;ffneten Datei */
{
	closeFile();
};

bool Packet_Loader_XML::getPacket(char* &current_packet_buffer,
					size_t &current_packet_size,
					string &current_service_name)
/** Zeiger 'current_packet_buffer' auf den Paketpuffer mit dem letzten
 *  geladenen Paket (siehe loadPacket()) bzw. NULL, wenn keins vorhanden ist,
 *  dessen Gr&ouml;&szlig;e in Bytes 'current_packet_size' und
 *  der aktuell gesetzte Dienstname 'current_service_name' (aus importPacket())
 *  wird zur&uuml;ckgegeben, <br>
 *  gibt 'true' zur&uuml;ck, wenn das Paket l&auml;nger als 0 Bytes ist,
 *  d.h. wenn ein Paket vorhanden ist */
{
	current_packet_buffer = NULL;
	current_packet_size = packet_size;
	current_service_name = service_name;
	if (packet_size > 0)
		current_packet_buffer = packet_buffer;
	return (packet_size > 0);
};

void Packet_Loader_XML::load(const char filename_to_load[])
/** schlie&szlig;t eine schon ge&ouml;ffnete Datei (closeFile()), &ouml;ffnet
 *  die XML-Datei mit dem Namen 'filename_to_load' und liest
 *  solange, bis die erste Paketbeschreibung folgen sollte,
 *  d.h. sucht die Tags "<netview>" und "<up_queue>" */
{
	bool			found_top_tag, found_queue_tag;
	int			state, depth;
	string			filename;

	// close old file, open new
	closeFile();
	filename = "";
	if (filename_to_load != NULL)
		if (filename_to_load[0] != 0) // strlen > 0
			filename = filename_to_load;
	cout << "Packet_Loader_XML: Loading: " << filename << endl;
	infile = xmlNewTextReaderFilename(filename.c_str());
	// cout << "Packet_Loader_XML: infile: " << infile << endl;

	// read until '<netview>'
	found_top_tag = false; state = 1;
	while ((!found_top_tag) && (state >= 0))
	{
		state = xmlTextReaderRead(infile);
		if (xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_netview) == 0)
			found_top_tag = true;
	};
	// if (found_top_tag)
	//	cout << "Packet_Loader_XML: Found <netview>." << endl;

	// read until '<up_queue>'
	depth = xmlTextReaderDepth(infile);
	found_queue_tag = false;
	while ((!found_queue_tag) && (found_top_tag) && (state >= 0))
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if (xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_up_queue) == 0)
			found_queue_tag = true;
	};
	// if (found_queue_tag)
	//	cout << "Packet_Loader_XML: Found <up_queue>." << endl;
};

void Packet_Loader_XML::importPacket(Packet *cable_packet)
/** das Paket hinter dem Zeiger 'cable_packet' wird in den Paketpuffer
 *  eingelesen, alle Kopfinformationen werden herauskopiert
 *  (Ethernet_Packet::exportHeader() usw.), bei Service-Paketen wird
 *  der Dienstname entnommen (Service_Packet::getServiceName()); <br> 
 *  neue Paketklassen sollten hier bekannt gemacht werden, zur Zeit sind
 *  folgende Klassen nutzbar: Ethernet_Packet, IPv4_Packet, TCP_Packet,
 *  UDP_Packet, TCP_Service_Packet, UDP_Service_Packet */
{
	Ethernet_Packet		*eth_packet;
	struct ether_header	eth_header;
	IPv4_Packet		*ipv4_packet;
	struct ip		ip_header;
	TCP_Packet		*tcp_packet;
	struct tcphdr		tcp_header;
	UDP_Packet		*udp_packet;
	struct udphdr		udp_header;
	TCP_Service_Packet	*tcp_service_packet;
	UDP_Service_Packet	*udp_service_packet;

	if (cable_packet)
	{
		packet_size = 0;
		// Ethernet and IPv4 packets
		eth_packet = NULL;
		if (cable_packet->isProtocol(ETHERNET))
			eth_packet = reinterpret_cast<Ethernet_Packet*>
					(cable_packet);
		if (eth_packet)
		{
			eth_packet->exportHeader(&eth_header);
			memcpy(packet_buffer, &eth_header, sizeof(eth_header));
			packet_size += sizeof(eth_header);
			cout << "Packet_Loader_XML: loaded ethernet header."
				<< endl;
		};
		ipv4_packet = NULL;
		if (cable_packet->isProtocol(IPv4))
			ipv4_packet = reinterpret_cast<IPv4_Packet*>
					(cable_packet);
		if (ipv4_packet)
		{
			ipv4_packet->exportHeader(&ip_header);
			memcpy(&(packet_buffer[sizeof(eth_header)]),
				&ip_header, sizeof(ip_header));
			packet_size += sizeof(ip_header);
			cout << "Packet_Loader_XML: loaded IPv4 header."
				<< endl;
		};

		// tcp and udp packets
		tcp_packet = NULL;
		if (cable_packet->isProtocol(TCP))
			tcp_packet = reinterpret_cast<TCP_Packet*>
					(cable_packet);
		if (tcp_packet)
		{
			tcp_packet->exportHeader(&tcp_header);
			memcpy(&(packet_buffer[sizeof(eth_header) + 
				sizeof(ip_header)]), &tcp_header,
				sizeof(tcp_header));
			packet_size += sizeof(tcp_header);
			cout << "Packet_Loader_XML: loaded TCP header."
				<< endl;
		};
		udp_packet = NULL;
		if (cable_packet->isProtocol(UDP))
			udp_packet = reinterpret_cast<UDP_Packet*>
					(cable_packet);
		if (udp_packet)
		{
			udp_packet->exportHeader(&udp_header);
			memcpy(&(packet_buffer[sizeof(eth_header) +
				sizeof(ip_header)]), &udp_header,
				sizeof(udp_header));
			packet_size += sizeof(udp_header);
			cout << "Packet_Loader_XML: loaded UDP header."
				<< endl;
		};

		// service packets
		service_name = "";
		tcp_service_packet = NULL;
		if ((cable_packet->isProtocol(SERVICE_PACKET)) && (tcp_packet))
			tcp_service_packet = reinterpret_cast
					<TCP_Service_Packet*>(cable_packet);
		if (tcp_service_packet)
		{
			service_name = tcp_service_packet->getServiceName();
			cout << "Packet_Loader_XML: loaded service name: "
				<< service_name << endl;
		};
		udp_service_packet = NULL;
		if ((cable_packet->isProtocol(SERVICE_PACKET)) && (udp_packet))
			udp_service_packet = reinterpret_cast
					<UDP_Service_Packet*>(cable_packet);
		if (udp_service_packet)
		{
			service_name = udp_service_packet->getServiceName();
			cout << "Packet_Loader_XML: loaded service name: "
				<< service_name << endl;
		};
	};
};

bool Packet_Loader_XML::loadPacket(void)
/** l&auml;dt ein Paket aus der ge&ouml;ffneten XML-Datei in den internen
 *  Paketpuffer, sollte erst nach load() gerufen werden; <br>
 *  gibt 'true' zur&uuml;ck, wenn ein Paket gelesen wurde, sonst 'false'; <br>
 *  ruft u.a. Net_Queue::loadPacketClasses() und importPacket() */
{
	int	state, depth;
	Packet	*cable_packet;

	cable_packet = NULL;
	if (!infile) return false;
	do
	{
		state = xmlTextReaderReadState(infile);
		depth = xmlTextReaderDepth(infile);
		// cout << "Packet_Loader_XML: state: " << state << " "
		// 	<< "depth: " << depth << endl;
		if (depth >= 1)
		{
			state = xmlTextReaderRead(infile);
			depth = xmlTextReaderDepth(infile);
		};
		// cout << "Packet_Loader_XML: infile: " << infile << endl;
		// if (state >= 1)
		//	cout << "Packet_Loader_XML: node: "
		//		<< xmlTextReaderConstName(infile) << endl;
		if ((state == 1) && (depth > 1))
			Net_Queue::loadPacketClasses(infile, cable_packet);
	} while ((state >= 1) && (depth > 1) && (cable_packet == NULL));

	// read -> import it
	if (cable_packet)
	{
		importPacket(cable_packet);
		delete cable_packet;
	};
	return (cable_packet != NULL);
};

void Packet_Loader_XML::closeFile(void)
/** schlie&szlig;t die ge&ouml;ffnete XML-Datei mit den Paketbeschreibungen */
{
	cout << "Packet_Loader_XML: closing: " << infile << endl;
	if (infile)
	{
		xmlTextReaderClose(infile);
		xmlFreeTextReader(infile);
	};
};

