/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"
#include "win_net.h"

#include <errno.h>
#include <fcntl.h>
#include <iostream>
#include <netdb.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <netinet/ether.h>
#include <netpacket/packet.h>

#include "constants.h"
#include "packet_server.h"
#include "tcp_service_packet.h"
#include "udp_service_packet.h"

/* size of the receive buffer, only the header data is interesting now */
/** L&auml;nge des Empfangspuffers, zur Zeit interessieren nur die K&ouml;pfe
 *  der Pakete */
const size_t	packet_server_buffer_size = 2048;
/** Zeit im Mikrosekunden als Timeout bei select in serveClient() */
const long	packet_server_timeout_usec = 50;

PacketServer::PacketServer(void) : server_port(0), server_socket(-1),
					current_client_service_name("")
/** Konstruktor, Initialisierung von Variablen: server_port, server_socket,
 *  current_client_service_name */
{
};

PacketServer::~PacketServer(void)
/** Destruktor, ruft nur stopServer() zum Beenden des Servers und
 *  delAllPackets() zum L&ouml;schen aller Pakete */
{
	stopServer();
	delAllPackets();
};

void PacketServer::init(unsigned short new_server_port)
/** Initialisierung: stoppt den laufenden Server (stopServer()) und startet
 *  einen neuen (startServer()) mit 'new_server_port' als Port, mit dem sich
 *  die Klienten verbinden k&ouml;nnen */
{
	stopServer();
	server_port = new_server_port;
	startServer();
};

int PacketServer::startServer(void)
/** startet den Server auf dem aktuell eingestellten Port 'server_port', <br>
 *  mit Socketoption SO_REUSEADDR zum schnellen Neustart, wenn noch Daten von
 *  alten Verbindungen in den Betriebsystempuffern stehen, <br>
 *  und O_NONBLOCK f&uuml;r nicht-blockierendes Lesen vom Socket; <br>
 *  gibt '0' zur&uuml;ck, wenn fehlerfrei, '-1' bei Fehler */
{
	int			socket_option, error;
	struct sockaddr_in	socket_address;

	error = 0;
	// open new socket, set socket option SO_REUSEADDR
	#ifdef WIN32
	init_socket();
	#endif
	server_socket = socket(AF_INET, SOCK_STREAM, 0);
	if (server_socket == -1)
	{
		cout << "PacketServer: Error opening socket." << endl;
		error = -1;
	};
	if (!error)
	{
		socket_option = 1;
		error = setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR,
				reinterpret_cast<const char*>(&socket_option),
				sizeof(socket_option));
		if (error != 0)
			cout << "PacketServer: Error setsockopt REUSEADDR."
				<< endl;
	};
	// bind() and listen()
	if (!error)
	{
		memset(&socket_address, 0, sizeof(socket_address));
		socket_address.sin_family = AF_INET;
		socket_address.sin_addr.s_addr = INADDR_ANY;
		socket_address.sin_port = htons(server_port);
		error = bind(server_socket,
				reinterpret_cast<sockaddr*>(&socket_address),
				sizeof(socket_address));
		if (error != 0) cout << "PacketServer: Error on bind." << endl;
	};
	if (!error)
	{
		error = listen(server_socket, 5);
		if (error != 0)
			cout << "PacketServer: Error on listen." << endl;
	};
	// set O_NONBLOCK for non blocking read access
	if (!error)
	{
		error = fcntl(server_socket, F_SETFL, O_NONBLOCK);
		if (error != 0)
		{
			cout << "PacketServer: ";
			cout << "Error setting non-blocking mode." << endl;
		};
	};
	return error;
};

int PacketServer::stopServer(void)
/** beendet den Server, schlie&szlig;t alle Verbindungen zu den Klienten und
 *  dann den Server(-socket) selber; <br>
 *  gibt '0' zur&uuml;ck, wenn fehlerfrei, '-1' bei Fehler */
{
	int	error, close_return;
	size_t	i;

	error = 0;
	// close all connections to clients
	close_return = 0;
	for(i = 0; i < client_sockets.size(); i++)
	{
		if (client_sockets[i] > 0)
			close_return = close(client_sockets[i]);
		client_sockets[i] = 0;
		if (close_return == -1) error = -1;
	};
	// deleting all entries
	client_sockets.clear();
	// close server socket
	if (server_socket > 0) close_return = close(server_socket);
	server_socket = -1;
	if (close_return == -1) error = -1;
	return error;
};

void PacketServer::delAllPackets()
/**
* L&ouml;scht alle Pakete des Servers.
*/
{
	list<Packet*>::iterator	all_packets;

	all_packets = incoming_packets.begin();
	while (all_packets != incoming_packets.end())
	{
		if (*all_packets) delete *all_packets;
		*all_packets = NULL;
		all_packets++;
	};
	incoming_packets.erase(incoming_packets.begin(),
				incoming_packets.end());
};

void PacketServer::serverStep(void)
/** nimmt die Verbindung zu einem neuen Klienten auf (soweit vorhanden)
 *  und empf&auml;ngt Daten von bestehenden Verbindungen (serveClient()),
 *  l&ouml;scht alle Eintr&auml;ge &uuml;ber vom Klienten geschlossene
 *  Verbindungen */
{
	int	new_client_socket;
	size_t	i;

	// reset error variable, has to be done !!! we need it !!!
	errno = 0;
	new_client_socket = accept(server_socket, 0, 0);
	if (new_client_socket > 0)
	{
		cout << "PacketServer: new client" << endl;
		client_sockets.insert(client_sockets.end(), new_client_socket);
	};
	for(i = 0; i < client_sockets.size(); i++)
	{
		if (client_sockets[i] > 0) serveClient(client_sockets[i]);
		if (client_sockets[i] < 0)
			client_sockets.erase(client_sockets.begin() + i);
	};
};

void PacketServer::readEthernetPacket(unsigned char* &packet_buffer,
				size_t &buffer_length, Packet* &cable_packet)
/** liest aus dem Puffer 'packet_buffer' ein Ethernet-Paket ein,
 *  'buffer_length' ist die Anzahl der benutzten Bytes im Puffer (empfangene
 *  Bytes), das Paket wird danach im Zeiger 'cable_packet' zur&uuml;ckgegeben,
 *  <br> ruft readIpPacket(), falls ein IP-Paket enthalten ist */
{
	size_t			new_buffer_length;
	struct ether_header	*eth_header;
	unsigned char		*new_packet_buffer;
	Ethernet_Packet		*eth_packet;

	new_packet_buffer = NULL;
	eth_packet = NULL;
	if (buffer_length < sizeof(ether_header)) packet_buffer = NULL;
	if (packet_buffer == NULL) return;

	eth_header = reinterpret_cast<ether_header*>(packet_buffer);
	if (eth_header)
	{
		// reposition buffer pointer and size for next call
		new_packet_buffer = &(packet_buffer[sizeof(ether_header)]);
		new_buffer_length = buffer_length - sizeof(ether_header);
		// cout << "Packet_Server: Found ethernet header: "
		//	<< eth_header << " " << new_buffer_length << " "
		//	<< "type: " << eth_header->ether_type << endl;
		if (eth_header->ether_type == htons(ETHERTYPE_IP))
			readIpPacket(new_packet_buffer, new_buffer_length,
					eth_packet);
		packet_buffer = new_packet_buffer;	// go to next packet
		buffer_length = new_buffer_length;
		if (eth_packet)
		{	// import header data
			eth_packet->importHeader(eth_header);
			cable_packet = reinterpret_cast<Packet*>(eth_packet);
		};
	};
};

void PacketServer::readIpPacket(unsigned char* &packet_buffer,
			size_t &buffer_length, Ethernet_Packet* &eth_packet)
/** liest aus dem Puffer 'packet_buffer' ein IP-Paket ein,
 *  'buffer_length' ist die Anzahl der benutzten Bytes im Puffer (empfangene
 *  Bytes), das Paket wird danach im Zeiger 'eth_packet' zur&uuml;ckgegeben,
 *  <br> ruft readTcpPacket(), falls ein TCP-Paket enthalten ist, und
 *  readUdpPacket(), falls ein UDP-Paket enthalten ist */
{
	size_t		packet_length, new_buffer_length;
	struct ip	*ipv4_header;
	unsigned char	*new_packet_buffer;
	IPv4_Packet	*ip_packet;

	new_packet_buffer = NULL;
	ip_packet = NULL;
	if (buffer_length < sizeof(struct ip)) packet_buffer = NULL;
	if (packet_buffer == NULL) return;

	ipv4_header = NULL;
	if (((packet_buffer[0] & 0xf0) >> 4) == IPVERSION)	// IPv4
		ipv4_header = reinterpret_cast<struct ip*>(packet_buffer);
	else cout << "Packet_Server: unknown IP version." << endl;
	if (ipv4_header)
	{
		// reposition buffer pointer and size for next call
		packet_length = ntohs(ipv4_header->ip_len);
							// without ether_header
		// cout << "Packet_Server: IPv4 Packet length: "
		//	<< packet_length << endl;
		new_packet_buffer = &(packet_buffer[sizeof(struct ip)]);
		if (ipv4_header->ip_p == IPPROTO_TCP)
			readTcpPacket(new_packet_buffer,
					packet_length - sizeof(struct ip),
					ip_packet);
		if (ipv4_header->ip_p == IPPROTO_UDP)
			readUdpPacket(new_packet_buffer,
					packet_length - sizeof(struct ip),
					ip_packet);
		// reposition buffer pointer and size for next call
		new_packet_buffer = &(packet_buffer[ packet_length ]);
		new_buffer_length = buffer_length - packet_length;
		packet_buffer = new_packet_buffer;
		buffer_length = new_buffer_length;
		if (ip_packet)
		{	// import header data
			ip_packet->importHeader(ipv4_header);
			eth_packet = reinterpret_cast<Ethernet_Packet*>
					(ip_packet);
		};
	};
};

bool PacketServer::findServiceNameForPort(unsigned short port, 
						const char protocol[],
						string &service_name)
/** sucht zur Portnummer 'port' des Protokolls 'protocol' (hier "tcp" oder
 *  "udp") den Dienstnamen 'service_name' per getservbyport(),
 *  zur&uuml;ckgegeben wird der gefundene Dienstname und 'true' bzw. ein
 *  Leerstring ("") und 'false', wenn keiner gefunden wurde*/
{
	bool		found;
	struct servent	*service_entry;

	found = false;
	service_entry = getservbyport(htons(port), protocol);
	service_name = "";
	if (service_entry)
	{
		service_name = service_entry->s_name;
		if (service_entry->s_aliases[0] != NULL)
			service_name += string("/") + 
					service_entry->s_aliases[0];
	};
	if (service_name != "") found = true;
	cout << "PacketServer: findService: Port: " << port << " ";
	cout << "Name: " << service_name << endl;
	return found;
};

void PacketServer::readTcpPacket(unsigned char* packet_buffer,
				size_t packet_length, IPv4_Packet* &ip_packet)
/** liest aus dem Puffer 'packet_buffer' ein TCP-Service-Paket ein,
 *  'buffer_length' ist die Anzahl der benutzten Bytes im Puffer (empfangene
 *  Bytes), das Paket wird danach im Zeiger 'ip_packet' zur&uuml;ckgegeben,
 *  <br> wenn kein Dienstname bekannt ist, wird findServiceNameForPort()
 *  gerufen, dessen Ergebnis wird als Dienstname eingetragen */
{
	struct tcphdr		*tcp_header;
	TCP_Service_Packet	*tcp_service_packet;
	string			service_name;

	tcp_service_packet = NULL;
	if ((packet_length < sizeof(struct tcphdr)) || (packet_buffer == NULL))
		return;

	tcp_header = reinterpret_cast<struct tcphdr*>(packet_buffer);
	if (tcp_header)
	{
		tcp_service_packet = new TCP_Service_Packet;
		if (tcp_service_packet)
		{
			// set service name
			if (current_client_service_name != "")
				tcp_service_packet->setServiceName(
					current_client_service_name.c_str());
			else
			{
				if (!findServiceNameForPort(
						ntohs(tcp_header->dest),
						"tcp", service_name))
				{
					if (findServiceNameForPort(
						ntohs(tcp_header->source),
						"tcp", service_name))
					{
						tcp_service_packet->
							setServiceName(
							service_name.c_str());
					};
				}
				else
					tcp_service_packet->setServiceName(
						service_name.c_str());
			};
			// import header data
			tcp_service_packet->importHeader(tcp_header);
			ip_packet = reinterpret_cast<IPv4_Packet*>
					(tcp_service_packet);
		};
		// cout << "Packet_Server: Received TCP packet: " << " "
		//	<< "Source port: " << ntohs(tcp_header->source) << " "
		//	<< "dest. port:  " << ntohs(tcp_header->dest) << endl;
	};
};

void PacketServer::readUdpPacket(unsigned char* packet_buffer,
				size_t packet_length, IPv4_Packet* &ip_packet)
/** liest aus dem Puffer 'packet_buffer' ein UDP-Service-Paket ein,
 *  'buffer_length' ist die Anzahl der benutzten Bytes im Puffer (empfangene
 *  Bytes), das Paket wird danach im Zeiger 'ip_packet' zur&uuml;ckgegeben,
 *  <br> wenn kein Dienstname bekannt ist, wird findServiceNameForPort()
 *  gerufen, dessen Ergebnis wird als Dienstname eingetragen */
{
	struct udphdr		*udp_header;
	UDP_Service_Packet	*udp_service_packet;
	string			service_name;

	if ((packet_length < sizeof(struct udphdr)) || (packet_buffer == NULL))
		return;

	udp_header = reinterpret_cast<struct udphdr*>(packet_buffer);
	if (udp_header)
	{
		udp_service_packet = new UDP_Service_Packet();
		if (udp_service_packet)
		{
			// set service name
			if (current_client_service_name != "")
				udp_service_packet->setServiceName(
					current_client_service_name.c_str());
			else
			{
				if (!findServiceNameForPort(
						ntohs(udp_header->dest),
						"udp", service_name))
				{
					if (findServiceNameForPort(
						ntohs(udp_header->source),
						"udp", service_name))
					{
						udp_service_packet->
							setServiceName(
							service_name.c_str());
					};
				}
				else
					udp_service_packet->setServiceName(
						service_name.c_str());
			};
			// import header data
			udp_service_packet->importHeader(udp_header);
			ip_packet = reinterpret_cast<IPv4_Packet*>
					(udp_service_packet);
		};
		// cout << "Packet_Server: Received UDP packet: " << " "
		//	<< "Source port: " << ntohs(udp_header->source) << " "
		//	<< "dest. port:  " << ntohs(udp_header->dest) << endl;
	};

};

void PacketServer::readCommandPacket(unsigned char* &packet_buffer,
					size_t &buffer_length)
/** versucht, ein Kommandopaket aus dem Puffer 'packet_buffer' der L&auml;nge
 *  'buffer_length' zu lesen, <br>
 *  zur Zeit werden nur Dienstnamen-Kommandos akzeptiert, die vor jedem
 *  Paket des Dienstes mitgeschickt werden */
{
	char	*service_name_command;
	char	new_service_name[service_name_length];

	if (buffer_length >= service_name_length)
	{
		memset(new_service_name, 0, service_name_length);
		service_name_command = reinterpret_cast<char*>(packet_buffer);
		// cout << "PacketServer: " << service_name_command << endl;
		// try to find service_name_format_string
		if (strncmp(service_name_command, service_name_format_string,
				strlen(service_name_format_string) - 2) == 0)
		{
			// cout << "PacketServer: found service name command.";
			// cout << endl;
			sscanf(service_name_command,
				service_name_format_string,
				new_service_name);	// get service name
			packet_buffer = &(packet_buffer[service_name_length]);
			buffer_length -= service_name_length;
		};
		current_client_service_name = new_service_name;
		cout << "PacketServer: Client service name: ";
		cout << current_client_service_name << endl;
	};
};

// Vorlesung "Evs Kapitel 3.5 "Parallelität in einem Prozeß - select"
int PacketServer::serveClient(int &current_client_socket)
/** "bedient" einen Klienten auf der Verbindung 'current_client_socket',
 *  d.h. empf&auml;ngt Pakete von ihm und h&auml;ngt sie an die interne Liste
 *  an, die durch getPacket() von Network::executeNet() gelesen wird; <br>
 *  gibt '-1' bei Fehler zur&uuml;ck, sonst '0' (fehlerfrei) */
{
	int		number, error;
	bool		stop_receiving;
	size_t		num_received, sum_num_received, rest_length;
	unsigned char	buffer[packet_server_buffer_size], *buffer_pos;
	Packet*		cable_packet;
	fd_set		rfds;
	struct timeval	time_value;

	// init all
	stop_receiving = false;
	sum_num_received = 0;
	error = 0;
	memset(buffer, 0, packet_server_buffer_size);
	FD_ZERO(&rfds);
	FD_SET(current_client_socket, &rfds);
	time_value.tv_sec = 0;
	time_value.tv_usec = packet_server_timeout_usec;
	number = current_client_socket + 1;
	do
	{
		buffer_pos = &(buffer[sum_num_received]);
		rest_length = packet_server_buffer_size - sum_num_received;
		num_received = static_cast<size_t>(-1);
		// are there any data to receive -> receive it?
		if (select(number, &rfds, NULL, NULL, &time_value) < 0)
			stop_receiving = true;
		if (FD_ISSET(current_client_socket, &rfds))
			num_received = recv(current_client_socket,
					reinterpret_cast<char*>(buffer_pos),
					rest_length, 0);
		else stop_receiving = true;
		// cout << "PacketServer: " << num_received << endl;
		// error output
		if (num_received == static_cast<size_t>(-1))
		{
			stop_receiving = true;
			if ((errno != EAGAIN) && (errno != 0)) error = -1;
			if (error)
			{
				cout << "PacketServer: Error: " << errno;
				cout << " " << error;
				cout << " " << strerror(errno) << endl;
			};
		}
		else sum_num_received += num_received;
		if (sum_num_received >= packet_server_buffer_size)
			stop_receiving = true;		
	} while ((num_received > 0) && (!stop_receiving));
	FD_ZERO(&rfds);
	// socket closed by client ?
	if (num_received == 0)
	{
		close(current_client_socket);
		current_client_socket = -1;	// mark for deletion
	};
	if ((!error) && (sum_num_received > 0))
	{
		// interpret received packet
		cout << "PacketServer: Received packet (size: ";
		cout << sum_num_received << " )." << endl;
		buffer_pos = buffer;
		rest_length = sum_num_received;
		while (buffer_pos != NULL)
		{
			cable_packet = NULL;
			readCommandPacket(buffer_pos, rest_length);
			readEthernetPacket(buffer_pos, rest_length,
						cable_packet);
			if (cable_packet)
				incoming_packets.insert(incoming_packets.end(),
							cable_packet);
		};
	};
	return error;
};

void PacketServer::getPacket(Packet* &cable_packet)
/** gibt erstes Paket 'cable_packet' der internen Liste der empfangenen
 *  Pakete zur&uuml;ck, NULL wenn keins da ist */
{
	cable_packet = NULL;
	if (incoming_packets.size() > 0)
	{
		cable_packet = *(incoming_packets.begin());
		// cout << "PacketServer::getPacket: " << cable_packet << endl;
		incoming_packets.erase(incoming_packets.begin());
	};
};

size_t PacketServer::numPackets(void) const
/** gibt die Anzahl der empfangenen Datenpakete zur&uuml;ck, die durch
 *  getPacket() gelesen werden k&ouml;nnen */
{
	return incoming_packets.size();
};

