/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#ifdef WIN32
#include "getopt.h"
#endif

#include <iostream>
#include <unistd.h>

#include "packet_client.h"
#include "packet_loader_xml.h"

using namespace std;

// For getopt
/** Text aller Optionen dieses Programms, f&uuml;r readArgs() */
const char		test_client_option_string[] = "f:i:p:";
/** Buchstabe der Option zur Server-IP-Adresse, siehe readArgs() */
const int		test_client_option_server_ip = 'i';
/** Buchstabe der Option zum Serverport, siehe readArgs() */
const int		test_client_option_server_port = 'p';
/** Buchstabe der Option zum Name einer XML-Datei mit Paketbeschreibungen,
 *  siehe readArgs() */
const int		test_client_option_xml_file = 'f';

/** Standard-Adresse des Servers, falls keine andere angegeben wurde */
const char		test_client_standard_address[] = "127.0.0.1";
/** Standard-Portnummer des Servers, falls keine andere angegeben wurde */
const unsigned short	test_client_standard_port = 9000;
/** Standard-Dateiname f&uuml;r XML-Paketbeschreibungen, falls kein anderer
 *  angegeben wurde */
const char		test_client_standard_xml_filename[] = "";
/** Hilfetext mit allen Optionen, wird bei jedem Start ausgegeben */
const char		test_client_usage_text[] =
		"Usage: test_client [-i ipaddress] [-p port] [-f filename]";

void connect(Packet_Client &packet_sender, const char *server_ip,
		unsigned short server_port)
/** verbindet sich mit Hilfe des Objektes 'packet_sender' mit dem Server
 *  der IP-Adresse 'server_ip' auf Port 'server_port' */
{
	cout << "test_client: Connecting." << endl;
	packet_sender.init(server_ip, server_port);
	packet_sender.connect();
};

void closeConnection(Packet_Client &packet_sender)
/** schlie&szlig;t mit Hilfe des Objektes 'packet_sender' die Verbindung zum
 *  Server */
{
	cout << "test_client: Closing." << endl;
	packet_sender.close();
};

void settingData(Packet_Client &packet_sender)
/** initialisiert das Objekt 'packet_sender': Standardwerte f&uuml;r die
 *  Pakete, die 'test_client' sendet, <br>
 *  Quell- und Ziel-IP-Adresse, Quell- und Zielport, IP-Kopf, TCP-Kopf,
 *  TCP-Flags und Dienstname */
{
	cout << "test_client: Setting data." << endl;
	packet_sender.setSourceIpAddress("192.168.2.2");
	packet_sender.setDestinationIpAddress("192.168.1.2");
	packet_sender.setIpHeader(0, 1, 7);

	packet_sender.setSourcePort(18000);
	packet_sender.setDestinationPort(80);
	packet_sender.setTcpHeader(14023324, 1553553, 32767, 0);
	packet_sender.setTcpFlags(0, 1, 0, 0, 1, 0);

	packet_sender.setServiceName("HTTP");
};

void readArgs(int argc, char* argv[], const char* &server_ip,
		unsigned short &server_port, const char* &xml_filename)
/** liest von der Kommandozeile Anzahl 'argc' Argumente im Feld 'argv' ein,
 *  zur&uuml;ckgegeben werden die neuen Werte von 'server_ip' (Adresse des
 *  Servers), 'server_port' (Portnummer des Servers) und 'xml_filename'
 *  (Name einer Datei mit Paketbeschreibungen), <br>
 *  diese Werte kommen entweder aus internen Konstanten oder von den Argumenten
 */
{
	int	current_opt;

	// init with standard values if no arg given
	server_ip = test_client_standard_address;
	server_port = test_client_standard_port;
	xml_filename = test_client_standard_xml_filename;
	cout << test_client_usage_text << endl;
	do
	{	// test all args
		current_opt = getopt(argc, argv, test_client_option_string);
		if (current_opt == test_client_option_server_ip)
			server_ip = optarg;
		if (current_opt == test_client_option_server_port)
			server_port = atoi(optarg);
		if (current_opt == test_client_option_xml_file)
			xml_filename = optarg;
	} while (current_opt != -1);
	cout << "test_client: Using: address: " << server_ip << " ";
	cout << "port: " << server_port << " file: " << xml_filename << endl;
};

void loadSendPacket(Packet_Client &packet_sender,
			Packet_Loader_XML &loader_xml)
/** l&auml;dt mit Hilfe des Objektes 'loader_xml' ein Paket aus der
 *  XML-Beschreibungsdatei und sendet es (inklusive Dienstname) mit der Hilfe
 *  von 'packet_sender' an den Server */
{
	bool	loaded, get_ok;
	char	*current_packet_buffer;
	size_t	current_packet_size;
	string	current_service_name;

	loaded = loader_xml.loadPacket();
	cout << "test_client: Loaded packet?: " << loaded << endl;
	get_ok = false;
	if (loaded)
		get_ok = loader_xml.getPacket(current_packet_buffer,
				current_packet_size, current_service_name);
	if (get_ok)
	{
		packet_sender.setServiceName(current_service_name.c_str());
		packet_sender.sendRawWithService(current_packet_buffer,
							current_packet_size);
	};
};

void mainLoop(Packet_Client &packet_sender, Packet_Loader_XML &loader_xml,
		const char *server_ip,
		unsigned short server_port, const char *xml_filename)
/** liest Tastendr&uuml;cke vom Nutzer ein und interpretiert sie als Befehle,
 *  bei Bedarf wird:
 *  - eine Verbindung zum Server hergestellt ('packet_sender',
 *  'server_ip': Adresse des Servers, 'server_port': Portnummer des Servers),
 *  - die Verbindung geschlossen,
 *  - ein Paket aus der Datei gelesen und gesendet ('loader_xml'), 
 *  - die Verbindung erneuert (geschlossen und neu hergestellt),
 *  - ein TCP-Paket (aus internen Daten) gesendet
 *  - ein UDP-Paket (aus internen Daten) gesendet
 *  - das Programm beendet,
 *
 *  der Name der Paketbeschreibungsdatei 'xml_filename' wird zur Zeit nicht
 *  ben&ouml;tigt */
{
	bool	do_quit;
	char	letter;

	do_quit = false;
	// read letter and do the command belonging to it
	while (!do_quit)
	{
		cout << "test_client: Press a key and Return to continue."
			<< endl;
		cin >> letter;
		switch (letter)
		{
			case 'c':	closeConnection(packet_sender);
					break;
			case 'l':	loadSendPacket(packet_sender,
						loader_xml);
					break;
			case 'q':	do_quit = true;
					break;
			case 'r':	cout << "test_client: Reconnecting"
						<< endl;
					closeConnection(packet_sender);
					connect(packet_sender, server_ip,
						server_port);
					break;
			case 'u':	cout << "test_client: Sending "
						<< "UDP packet." << endl;
					packet_sender.sendUdpPacket();
					break;
			default:	cout << "test_client: Sending "
						<< "TCP packet." << endl;
					packet_sender.sendTcpPacket();
					break;
		};
	};
};

int main(int argc, char *argv[])
/** Hauptprogramm:
 *  - liest die Argumente von der Kommandozeile (readArgs()),
 *  - verbindet sich mit dem Server (connect()),
 *  - initialisiert die Paketeinstellungen (settingData()),
 *  - wartet auf Kommandos (mainLoop()) und
 *  - schlie&szlig;t die Verbindung zum Server (closeConnection())
 *
 *  gibt immer '0' zur&uuml;ck */
{
	const char		*server_ip, *xml_filename;
	unsigned short		server_port;
	Packet_Client		packet_sender;
	Packet_Loader_XML	loader_xml;

	// cout << "test_client: reading arguments." << endl;
	readArgs(argc, argv, server_ip, server_port, xml_filename);
	// cout << "test_client: loading file." << endl;
	if (xml_filename[0] != 0)	// strlen > 0
		loader_xml.load(xml_filename);
	connect(packet_sender, server_ip, server_port);
	settingData(packet_sender);
	cout << "test_client: c: close; l: load packet; q: quit;" << endl;
	cout << "test_client: r: reconnect; u: send UDP; "
		<< "any other key: send TCP" << endl;
	mainLoop(packet_sender, loader_xml,
			server_ip, server_port, xml_filename);
	closeConnection(packet_sender);
	return 0;
};

