/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <iostream>
#include <math.h>
#include <GL/glu.h>

#include "camera.h"

/** minimale Geschwindigkeit der Kamera */
const double	camera_max_speed = 0.9;
/** maximale Geschwindigkeit der Kamera */
const double	camera_min_speed = -0.9;
/** Gr&ouml;&szlig;e der Pyramide, die die Kamera in der Karte markiert */
const double	camera_draw_size[3] = {1.0, 1.5, 1.0};
/** Radius der Spitze der Pyramide, die die Kamera in der Karte markiert */
const double	camera_draw_top_radius = 0.1;
/** Farbe der Pyramide, die die Kamera in der Karte markiert */
const double	camera_draw_color[4] = {0.0, 1.0, 0.0, 1.0};
/** Geschwindigkeit der Kamera-Drehung wenn sie selbst rotiert */
const double	camera_turn_step_factor = 0.25;
/** Abstand der Kamera zum Objekt bei der Animation */
const double	camera_animated_cam_distance = -1.8;

using namespace std;

Camera::Camera(void) :
	speed(0.0), max_speed(camera_max_speed), min_speed(camera_min_speed),
	x_angle(0.0), y_angle(0.0), y_turn_to_angle(0.0),
	x_position(0.0), y_position(0.0), z_position(0.0),
	center_x(0.0), center_y(0.0), center_z(0.0),
	up_x(0.0), up_y(1.0), up_z(0.0)
/** Konstruktor, Werte initialisieren */
{
};

Camera::~Camera(void)
/** Destruktor, leer */
{
};

void Camera::draw(void) const
/** Ausgabe der Kamera-Pyramide f&uuml;r die Kartenansicht */
{
	glTranslated(x_position, y_position, z_position);
	glRotated(90.0, 0, 0, 1);
	glRotated(y_angle * 180.0 / M_PI + 180.0, 1, 0, 0);
	glRotated(x_angle * 180.0 / M_PI, 0, 0, 1);
	camera_figure.draw();
	glRotated(- x_angle * 180.0 / M_PI, 0, 0, 1);
	glRotated(- (y_angle * 180.0 / M_PI + 180.0), 1, 0, 0);
	glRotated(-90.0, 0, 0, 1);
	glTranslated(-x_position, -y_position, -z_position);
};

void Camera::gotoPos(void) const
/** Blickfeld in der 3D-Welt setzen, entsprechend Kamera-Position und
 *  -Drehwinkel */
{
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(x_position, y_position, z_position, 
			center_x, center_y, center_z, up_x, up_y, up_z);
};

double Camera::getXAngle(void) const
/** gibt den Winkel zur Horizontalen (X-Achse) zur&uuml;ck,
 *  alle Winkel 0.0 - 360.0 */
{
	return x_angle * 180.0 / M_PI;
};

double Camera::getYAngle(void) const
/** gibt den Winkel um die Senkrechte (Y-Achse) zur&uuml;ck, 0.0 - 360.0 */
{
	return y_angle * 180.0 / M_PI;
};

void Camera::getPos(double &x_pos, double &y_pos, double &z_pos) const
/** gibt die Position der Kamera zur&uuml;ck (x, y, z) */
{
	x_pos = x_position;
	y_pos = y_position;
	z_pos = z_position;
};

void Camera::getPos(Point_3D &camera_pos) const
/**  gibt die Position der Kamera zur&uuml;ck (als Punkt) */
{
	camera_pos.setPos(x_position, y_position, z_position);
};

void Camera::init(void)
/** Kamera-Werte initialisieren und Pyramide (Markierung in Karte) einstellen
 *  (Gr&ouml;&szlig;e, Farbe, Position) */
{
	speed = 0.0;
	x_angle = 0.0, y_angle = 0.0; y_turn_to_angle = 0.0;
	x_position = 0.0; y_position = 0.0, z_position = 0.0;
	center_x = 0.0; center_y = 0.0; center_z = 0.0;
	up_x = 0.0; up_y = 1.0; up_z = 0.0;

	camera_figure.setPos(0.0, 0.0, 0.0);
	camera_figure.setSize(camera_draw_size[0], camera_draw_size[1],
				camera_draw_size[2]);
	camera_figure.setTopRadius(camera_draw_top_radius);
	camera_figure.setColor(camera_draw_color[0], camera_draw_color[1],
				camera_draw_color[2], camera_draw_color[3]);
};

bool Camera::is_moving(void) const
/** gibt zur&uuml;ck, ob sich die Kamera bewegt
 * (true, Geschwindigkeit nicht 0) */
{
	return (speed != 0.0); 
};

void Camera::movePos(double add_x_pos, double add_y_pos, double add_z_pos)
/** verschiebt die Kamera um die angegebenen Werte in X/Y/Z-Richtung */
{
	double	rotated_add_x, rotated_add_y, rotated_add_z;

	// x_angle used correct?
	// cout << "Camera::movePos: " << add_x_pos << " " << add_y_pos
	//	<< " " << add_z_pos << endl;
	rotated_add_x = add_x_pos * sin(y_angle) * cos(x_angle);
	rotated_add_x += add_y_pos * sin(x_angle);
	rotated_add_x -= add_z_pos * cos(y_angle) * cos(x_angle);
	rotated_add_y = add_y_pos * cos(x_angle);
	rotated_add_z = add_x_pos * cos(y_angle) * cos(x_angle);
	rotated_add_z += add_y_pos * sin(x_angle);
	rotated_add_z += add_z_pos * sin(y_angle) * cos(x_angle);
	x_position += rotated_add_x;
	y_position += rotated_add_y;
	z_position += rotated_add_z;
	// cout << "Camera: position: " << x_position << " " << y_position
	//	<< " " << z_position << " " << " speed: " << speed << " "
	//	<< "angles: " << x_angle << " " << y_angle << endl;
	turn(0.0, 0.0);		// to adjust center
};

void Camera::move(double speed_step, double speed_factor)
/** verschiebt die Kamera in die aktuelle Richtung (X- und Y-Winkel) 
 *  mit der <br>
 *  neuen Geschwindigkeit = (alte Geschw. + speed_step) * speed_factor, <br>
 *  um den Weg = neue Geschwindigkeit * 1 Zeiteinheit */
{	
	double	old_speed;

	// angles in radians
	old_speed = speed;
	speed = (speed + speed_step) * speed_factor;
	if ((old_speed == 0.0) && (speed == 0.0)) return;	// no change
	if (speed > max_speed) speed = max_speed; 
	if (speed < min_speed) speed = min_speed;
	x_position += speed * cos(y_angle) * cos(x_angle);
	y_position += speed * sin(x_angle);
	z_position -= speed * sin(y_angle) * cos(x_angle);
	// cout << "position: " << x_position << " " << y_position << " ";
	// cout << z_position << endl;
	// cout << "speed: " << speed << " step: " << speed_step << endl;
	turn(0.0, 0.0);		// to adjust center
};

void Camera::setPos(double x_pos, double y_pos, double z_pos)
/** setzt die Position der Kamera, Drehwinkel werden beibehalten */
{
	speed = 0;
	x_position = x_pos;
	y_position = y_pos;
	z_position = z_pos;
	// cout << "Camera: position: " << x_position << " "
	//	<< y_position << " " << z_position << endl;
	turn(0.0, 0.0);
};

void Camera::setAngle(double new_x_angle, double new_y_angle)
/** setzt die neue Richtung der Kamera (Winkel um X- und Y-Achse),
 *  alle Winkel 0.0 bis 360.0 */
{
	x_angle = fmod(new_x_angle + 360.0, 360.0) * M_PI / 180.0;
	y_angle = fmod(new_y_angle + 360.0, 360.0) * M_PI / 180.0;
	y_turn_to_angle = y_angle;
	// cout << "x_angle: " << new_x_angle << " ";
	// cout << "y_angle: " << new_y_angle << endl;
	turn(0.0, 0.0);
};

void Camera::turn(double x_angle_step, double y_angle_step)
/** Dreht die Kamera um die vorgegebenen Winkel um die X- und Y-Achse,
 *  alle Winkel 0.0 bis 360.0 */
{
	double	x_angle_round, y_angle_round;

	x_angle_step = x_angle_step * M_PI / 180.0;
	y_angle_step = y_angle_step * M_PI / 180.0;
	x_angle = fmod(x_angle + x_angle_step + 2*M_PI, 2*M_PI);
	if (((y_angle < y_turn_to_angle) && 
		(y_angle + y_angle_step >= y_turn_to_angle)) ||
		((y_angle > y_turn_to_angle) &&
		(y_angle + y_angle_step <= y_turn_to_angle)))
	{	// needed to get right to value of turn_to_angle
		y_angle = y_turn_to_angle; 
	}
	else y_angle = fmod(y_angle + y_angle_step + 2*M_PI, 2*M_PI);
	if (fabs(y_turn_to_angle - y_angle) < 0.001) // rounding errors
		y_angle = y_turn_to_angle;
	// caculate center
	center_x = x_position + 1 * cos(y_angle) * cos(x_angle);
	center_y = y_position + 1 * sin(x_angle);
	center_z = z_position - 1 * sin(y_angle) * cos(x_angle);
	if (fabs(center_x) < 0.000001) center_x = 0.0;	// rounding
	if (fabs(center_y) < 0.000001) center_y = 0.0;
	if (fabs(center_z) < 0.000001) center_z = 0.0;
	x_angle_round = fmod(rint(x_angle / M_PI * 2), 4.0); // -> 0 - 4
	y_angle_round = fmod(rint(y_angle / M_PI * 2), 4.0); 
	// cout << "xar yar: " << x_angle_round << " " << y_angle_round << endl;
	// finding up vector
	if (x_angle_round == 0.0)
		{ up_x = 0.0; up_y = 1.0; up_z = 0.0; }
	if (x_angle_round == 2.0)
		{ up_x = 0.0; up_y = -1.0; up_z = 0.0; }
	if (((x_angle_round == 1.0) && (y_angle_round == 0.0)) ||
		((x_angle_round == 3.0) && (y_angle_round == 2.0)))
		{ up_x = -1.0; up_y = 0.0; up_z = 0.0; }
	if (((x_angle_round == 3.0) && (y_angle_round == 0.0)) ||
		((x_angle_round == 1.0) && (y_angle_round == 2.0)))
		{ up_x = 1.0; up_y = 0.0; up_z = 0.0; }
	if (((x_angle_round == 1.0) && (y_angle_round == 1.0)) ||
		((x_angle_round == 3.0) && (y_angle_round == 3.0)))
		{ up_x = 0.0; up_y = 0.0; up_z = -1.0; }
	if (((x_angle_round == 3.0) && (y_angle_round == 1.0)) ||
		((x_angle_round == 1.0) && (y_angle_round == 3.0)))
		{ up_x = 0.0; up_y = 0.0; up_z = 1.0; }
	// if ((x_angle_step != 0.0) || (y_angle_step != 0.0))
	//	speed = 0;	// stop if turning
	// cout << "center: " << center_x << " " << center_y << " " << center_z;
	// cout << endl;
	// cout << "Camera: x_angle: " << x_angle * 180.0 / M_PI 
	//	<< " y_angle: " << y_angle * 180.0 / M_PI << endl;
};

bool Camera::needsTurn(void) const
/** zeigt an ob sich die Kamera noch drehen muss, um den angestrebten Winkel
 *  (siehe turnToAngle() ) zu erreichen (true); oder schon da ist (false) */
{
	// cout << "Camera: " << "current angle: " << y_angle << " "
	//	<< "needed angle: " << y_turn_to_angle << " "
	//	<< "diff: " << y_turn_to_angle - y_angle << endl;
	return (y_turn_to_angle != y_angle);
};

void Camera::turnToAngle(double new_y_angle)
/** setzt den neuen angestrebten Winkel um die Y-Achse, der zu erreichen ist */
{
	y_turn_to_angle = fmod(new_y_angle * M_PI / 180.0 + 2 * M_PI,
				2 * M_PI);
};

void Camera::turnToAngleStep(void)
/** dreht die Kamera um einen Schritt Richtung angestrebtem Winkel,
 *  Drehgeschwindigkeit ist eine Hysterese, abh&auml;ngig von Konstanten
 *  und Differenz zum angestrebten Winkel */
{
	double	turn_step_angle, diff_angle;

	if (fabs(y_angle - y_turn_to_angle) > M_PI)
	{
		if (y_angle < M_PI)
			diff_angle = y_turn_to_angle - (y_angle + 2 * M_PI);
		else diff_angle = (y_turn_to_angle + 2 * M_PI) - y_angle;
		turn_step_angle = camera_turn_step_factor * (180.0 / M_PI) *
					diff_angle;
	}
	else turn_step_angle = camera_turn_step_factor * (180.0 / M_PI) *
				(y_turn_to_angle - y_angle);
	// cout << "Camera: " << (y_angle * 180.0 / M_PI) << " "
	//	<< (y_turn_to_angle * 180.0 / M_PI) << " "
	//	<< turn_step_angle << endl;
	turn(0.0, turn_step_angle);
};

void Camera::updateCam(Point_3D &new_camera_pos, double new_camera_y_angle,
			bool follow_angle)
/** setzt Kamerawinkel und Position anhand der &uuml;bergebenen Werte
 *  (Abstand entsprechend Konstante, Winkel zur Horizontalen 30 Grad nach
 *  unten) soweit follow_angle false ist, sonst <br>
 *  bei follow_angle true wird nur der Y-Winkel angepasst, damit die Kamera
 *  auf das Objekt an der angegebenen Position schaut */
{
	double		new_camera_x_angle;
	Point_3D	current_camera_pos;

	// cout << "Camera::updateCam: new_y_angle: " << new_camera_y_angle
	//	<< endl;
	if (!follow_angle)
	{
		setPos(new_camera_pos.getX(), new_camera_pos.getY(),
			new_camera_pos.getZ());
		new_camera_x_angle = - 30;
		new_camera_y_angle = new_camera_y_angle - 270;
		setAngle(new_camera_x_angle, getYAngle());
		turnToAngle(new_camera_y_angle);
		move(camera_animated_cam_distance, 1.0);
		move(0.0, 1.0);
		move(0.0, 0.0);
	}
	else
	{
		getPos(current_camera_pos);
		new_camera_x_angle = getXAngle();
		new_camera_y_angle = 0 +
			current_camera_pos.getAngleXZ(new_camera_pos);
		setAngle(new_camera_x_angle, getYAngle());
		turnToAngle(new_camera_y_angle);
	};
	// cout << "Camera: camera angle (x y): "
	//	<< camera_x_angle << " " << camera_y_angle << endl;
};

Camera& Camera::operator =(const Camera	&new_values)
/** &Uuml;bergabe der Werte einer anderen Kamera bei Operator '=' */
{
	speed = new_values.speed;
	min_speed = new_values.min_speed;
	max_speed = new_values.max_speed;
	x_angle = new_values.x_angle;
	y_angle = new_values.y_angle;
	y_turn_to_angle = new_values.y_turn_to_angle;
	x_position = new_values.x_position;
	y_position = new_values.y_position;
	z_position = new_values.z_position;
	turn(0.0, 0.0);
	return *this;
};

