/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include "point_3d.h"

#include <iomanip>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <GL/gl.h>

#include "constants.h"

Point_3D::Point_3D(void) :	posx(0.0), posy(0.0), posz(0.0), 
				red(0.0), green(0.0), blue(0.0), alpha(0.0)
/** Konstruktor: Alle Variablen auf 0.0 setzen */
{
};

Point_3D::~Point_3D(void)
/** Destruktor: ist leer */
{
};

double Point_3D::getX(void) const
/** gibt Position auf der X-Achse zurck */
{
   return posx;
};

double Point_3D::getY(void) const
/** gibt Position auf der Y-Achse zurck */
{
   return posy;
};

double Point_3D::getZ(void) const
/** gibt Position auf der Z-Achse zurck */
{
   return posz;
};

double Point_3D::getR(void) const
/** gibt Rot-Anteil der Farbe zurck, Wertebereich 0.0 - 1.0 */
{
   return red;
};

double Point_3D::getG(void) const
/** gibt Grn-Anteil der Farbe zurck, Wertebereich 0.0 - 1.0 */
{
   return green;
};

double Point_3D::getB(void) const
/** gibt Blau-Anteil der Farbe zurck, Wertebereich 0.0 - 1.0 */
{
   return blue;
};

double Point_3D::getA(void) const
/** gibt Alpha-Anteil der Farbe zurck, Wertebereich 0.0 - 1.0, 
 * Alpha-Wert funktionert nur mit Blending */
{
   return alpha;
};

double Point_3D::getDistance(const Point_3D &end_point, bool with_y) const
/** Berechnet den Abstand zwischen diesem (this) und bergebenen Punkt 
 * (end_point), bei with_y == true wird auch der Y-Wert (Hhe) mit 
 * einbezogen, sonst nur X- und Z-Achse. <BR>
 * Formel: distance = sqrt( (x2 - x1)^2 + (y2 - y1)^2 + (z2 - z1)^2) */
{
   double	distance;

   distance = fabs(posx - end_point.getX()) * fabs(posx - end_point.getX());
   distance += fabs(posz - end_point.getZ()) * fabs(posz - end_point.getZ());
   if (with_y)	distance += fabs(posy - end_point.getY()) * 
	   			fabs(posy - end_point.getY());
   distance = sqrt(distance);
   return distance;
};

double Point_3D::getAngleXZ(const Point_3D &end_point) const
/** Berechnet den Winkel zwischen diesem (this) und bergebenem Punkt 
 * (end_point) bezogen auf X- und Z-Achse. Ist Endpunkt genau rechts von 
 * diesem, dann ist Ergebnis 0 Grad; genau links ergibt 180 Grad, genau 
 * dahinter sind 90 Grad. <BR>
 * Formel: Winkel = arcus_tan((z2 - z1) / (x2 - x1)) */
{
   double	angle;

   angle = atan((end_point.getZ() - posz) / fabs(end_point.getX() - posx));
   angle = angle * 180.0 / M_PI + 180.0;
   if (end_point.getX() >= posx) angle = 180.0 - angle;
   return angle;
};

void Point_3D::parseXML(xmlNode * const parse_child)
/** Liest XML-Tags mit dem Attribut 'point' ein. Darstellung: "x, y, z",
 *  gegeben z.B. als "1.0, -1.1, -1.2" */
{
   int          len, i;
   char         *to_d_begin, *to_d_end;
   double	pos_number;
   xmlAttr      *property;
   xmlChar      *compare, *content, *begin, *end;
   xmlNode      *child;
	        
   child = parse_child;
   posx = 0.0; posy = 0.0; posz = 0.0;
   if (child)		// existiert Knoten im Baum ?
   {
	property = child->properties;	// Attribute des Tags lesen
        // cout << "parsePoint : " << child->name << "\t";
	while (property)		// Fr alle Attribute
        {
		// Vergleich mit vorgegebenem Namen 'point'
		compare = xmlStrdup(xml_syntax_point);
		content = 0;
		if (xmlStrcasecmp(property->name, compare) == 0)
			if (property->children)
			{
				content = property->children->content;
				// cout << property->children->content << endl;
			};
		// Falls Attribut "point" einen Wert hat
		if (content)
		{
			begin = content;
			len = -1;
			// String aufteilen bei den Kommas, umwandeln in Zahlen
			for(i = 0; i < 3; i++)
			{
				begin = xmlStrsub(begin, len + 1, 
						xmlStrlen(begin));
				end = xmlStrdup(xmlStrchr(begin, ','));
				len = xmlStrlen(begin) - xmlStrlen(end);
				to_d_begin = reinterpret_cast<char*>(begin);
				to_d_end = 0;
				pos_number = strtod(to_d_begin, &to_d_end);
				if (i == 0) posx = pos_number;
				if (i == 1) posy = pos_number;
				if (i == 2) posz = pos_number;
			};
			// cout << point[0] << " " << point[1] 
			//	<< " " << point[2] << endl;
		};
		property = property->next;	// nchstes Attribut
	};	
   };

};

void Point_3D::loadXML(xmlTextReaderPtr infile)
/** Liest XML-Tags mit dem Attribut 'point' aus einer XML-Datei 'infile' ein,
 *  <br> Darstellung: "x, y, z", gegeben z.B. als "1.0, -1.1, -1.2" */
{
	int	len, i;
	char	*to_d_begin, *to_d_end;
	double	pos_number;
	xmlChar	*content, *begin, *end;

	// cout << "Point_3D::loadXML" << endl;
	if (xmlTextReaderHasAttributes(infile))
	{
		content = xmlTextReaderGetAttribute(infile, xml_syntax_point);
		// cout << "attribute of pos: " << content << endl;
		if (content)
		{
			begin = content;
			len = -1;
			// String aufteilen bei den Kommas, umwandeln in Zahlen
			for(i = 0; i < 3; i++)
			{
				begin = xmlStrsub(begin, len + 1, 
						xmlStrlen(begin));
				end = xmlStrdup(xmlStrchr(begin, ','));
				len = xmlStrlen(begin) - xmlStrlen(end);
				to_d_begin = reinterpret_cast<char*>(begin);
				to_d_end = 0;
				pos_number = strtod(to_d_begin, &to_d_end);
				if (i == 0) posx = pos_number;
				if (i == 1) posy = pos_number;
				if (i == 2) posz = pos_number;
				#ifndef WIN32
				delete [] end;
				#endif
			};
			// cout <<  posx << " " << posy 
			//	<< " " << posz << endl;
			#ifndef WIN32
			delete [] content;
			#endif
		};
	};
};

void Point_3D::saveXML(xmlTextWriterPtr outfile) const
/** speichert einen Punkt in der XML-Datei 'outfile' */
{
	xmlTextWriterStartElement(outfile, xml_syntax_pos);
	xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_point, 
						"%f, %f, %f", 
						posx, posy, posz);
	xmlTextWriterEndElement(outfile);
};

void Point_3D::setPos(double new_posx, double new_posy, double new_posz)
/** neue Positionswerte setzen */
{
   posx = new_posx; posy = new_posy; posz = new_posz;
};

void Point_3D::setColor(double new_red, double new_green, 
			double new_blue, double new_alpha)
/** neue Farbwerte setzen, RGBA, Wertebereich 0.0 - 1.0, alpha nur mit
 *  Blending nutzbar */
{
   red = new_red; green = new_green; blue = new_blue; alpha = new_alpha;
};

void Point_3D::translate(double mult) const
/** Aktuelle Darstellungsmatrix um diesen Punkt-Vektor verschieben,
 * multipliziert mit 'mult'. mult = -1: in entgegengesetzte Richtung */
{
   glTranslatef(posx * mult, posy * mult, posz * mult);
};

void Point_3D::colorForDraw(void) const
/** Farbe des Punktes als Pinselfarbe setzen, verwendet fr die n&auml;chsten 
 * Figuren (Linien, Rechtecke, ...) */
{
	GLfloat	material[] = {red, green, blue, alpha}; 
   
	// Farbwert ohne Licht
	glColor4d(red, green, blue, alpha);
	// Farbwert bei eingeschaltetem Licht (Reflektion von ambientem und 
	// diffusem und spekularem Licht
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, material);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, material);
	memcpy(material, material_specular_std, sizeof(material));
	material[3] = alpha;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, material_shininess_std);
};

void Point_3D::vertexDraw(double add_x, double add_y, double add_z) const
/** diesen Punkt zeichnen (kann auch Teil einer geometrischen Figur sein) */
{
   glVertex3d(posx + add_x, posy + add_y, posz + add_z);
};

Point_3D& Point_3D::operator +=(const Point_3D &point)
/** addiert zu diesem Punkt den Punkt 'point' dazu, Ergebnis wird hier
 *  gespeichert, Referenz auf diesen Punkt (this) wird zur&uuml;ckgegeben */
{
	posx += point.posx;
	posy += point.posy;
	posz += point.posz;
	red += point.red;
	green += point.green;
	blue += point.blue;
	alpha += point.alpha;
	return *this;
};

ostream& operator << (ostream &output, const Point_3D &point)
/** gibt den Punkt 'point' auf dem Ausgabestream 'output' aus, Referenz auf den
 *  Ausgabestream wird zur&uuml;ckgegeben */
{
	output << "X, Y, Z: " << rint(point.getX() * 100000) / 100000 << ", "
		<< rint(point.getY() * 100000) / 100000 << ", " 
		<< rint(point.getZ() * 100000) / 100000 << "; ";
	/* output << "R, G, B, A: " << point.red << ", " << point.green << ", " 
		<< point.blue << ", " << point.alpha;*/
	return output;
};

