/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include "viewport.h"

#include <math.h>
#include <iostream>
#include <GL/gl.h>
#include <GL/glu.h>

using namespace std;

Viewport::Viewport(void)
/** Konstruktor: Darstellungsmatrizen mit 0.0 initialiseren */
{
   int i;

   for(i = 0; i < 16; i++)
   {
	model_matrix[i] = 0.0; 
	projection_matrix[i] = 0.0;
   };
   drawContent = NULL;
};

Viewport::~Viewport(void)
/** Destruktor: ist leer */
{
};

void Viewport::scale(double new_width, double new_height)
/** Skalierung des Bereiches auf neue Breite und Hhe */
{
   left *= new_width; right *= new_width ;
   top *= new_height; bottom *= new_height;
   /* cout << "Viewport: " << left << " " << top << " " << right - left << " " 
	<< bottom - top << endl;*/
};

void Viewport::scalePoint(double window_height, double size[2], 
			  double point[2], double scaled_point[2])
/** Skalieren eines Punktes vom Fenstermanager-Darstellung auf Punkt innerhalb
 * des Fensterbereiches. <BR>
 * Parameter: <BR>
 * Fensterhhe window_height in Bildschirm-Pixeln, <BR>
 * Gre des Fensterbereiches size im Zielmastab <BR>
 * umzurechnender Punkt point <BR>
 * Rckgabe des berechneten Punktes in scaled_point */
{
   double my_top, my_bottom, my_height;

   // berechnen der Ausmae des Fensterbereiches
   my_top = window_height - bottom;
   my_bottom = window_height - top;
   my_height = my_bottom - my_top;
   // umrechnen und skalieren
   scaled_point[0] = (point[0] - left) * size[0] / (right - left);
   scaled_point[1] = size[1] - (point[1] - my_top) * size[1] / my_height;
};

void Viewport::setPos(double new_left, double new_top,
			double new_right, double new_bottom)
/** Setzen der Eckpunkte left, top, right, bottom wie in Rectangle
 */
{
	Rectangle::setPos(new_left, new_top, new_right, new_bottom);
};

void Viewport::setPos(const double new_pos[4], const double scaling[4])
/** Setzen der Position mit Feldern als Parameter, inklusive Skalierung <BR>
 * Parameter: <BR>
 * neue Position new_pos {left, top, right, bottom}
 * mit Skalierung scaling (scaling[0] * left usw.) */
{
	setPos(new_pos[0] * scaling[0], new_pos[1] * scaling[1],
		new_pos[2] * scaling[2], new_pos[3] * scaling[3]);
};

void Viewport::setModelMatrix(void)
/** aktuell in OpenGL gesetzte Matrix fr Verschiebung und Rotation in Viewport
 * speichern, Matrix wird in draw() verwendet */
{
   glMatrixMode(GL_MODELVIEW);
   glGetDoublev(GL_MODELVIEW_MATRIX, model_matrix);
};

void Viewport::setProjectionMatrix(void)
/** aktuell in OpenGL gesetzte Matrix fr Pojektion (Sichtbereich, Verzerrung)
 * in Viewport speichern, Matrix wird in draw() verwendet */
{
   glMatrixMode(GL_PROJECTION);
   glGetDoublev(GL_PROJECTION_MATRIX, projection_matrix);
};

void Viewport::setDrawContent(int (*new_drawContent)(void))
/** neure Darstellungsfunktion setzen */
{
   drawContent = new_drawContent;
};

void Viewport::setglOrtho(const double ortho_values[6],
				const double scaling[6])
/** ruft glOrtho mit den Werten aus bergebenem Feld, skaliert mit scaling,
 * Werte werden per setProjectionMatrix gleich bernommen */
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(ortho_values[0] * scaling[0], ortho_values[1] * scaling[1],
		ortho_values[2] * scaling[2], ortho_values[3] * scaling[3],
		ortho_values[4] * scaling[4], ortho_values[5] * scaling[5]);
	setProjectionMatrix();
};

void Viewport::setgluPerspective(double fovy, double aspect,
					double zNear, double zFar)
/** ruft gluPerspective mit den bergebenen Werten, durch eingebautes
 * setProjectionMatrix wird die Matrix gleich bernommen */
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(fovy, aspect, zNear, zFar);
	setProjectionMatrix();
};

bool Viewport::isInViewport(int window_height, int x, int y) const
/** Testet, ob gegebener Punkt (x, y) in Bildschirmpixeln 
 * bei aktueller Fensterhhe window_height in Bildschirmpixeln
 * innerhalb des Fensterbereiches liegt. Rand zhlt als auerhalb. */
{ 
   return ((x > left) && (x < right) && 
	   (y > window_height - bottom) && (y < window_height - top));
};

void Viewport::draw(void) const
/** Zeichenfunktion, ldt gespeicherte Matrizen und ruft vorher gesetzte
 * Darstellungsfunktion */
{
   // Bereich des Fensters festlegen, damit nichts darber hinaus 
   // gezeichnet wird
   glViewport(static_cast<int>(rint(left)), 
              static_cast<int>(rint(top)),
              static_cast<int>(rint(right - left)), 
              static_cast<int>(rint(bottom - top)));
   // Hintergrundfarbe setzen (einstellbar ber setColor() )
   glClearColor(red, green, blue, alpha);
   // Matrizen laden
   glMatrixMode(GL_PROJECTION); glLoadMatrixd(projection_matrix);
   glMatrixMode(GL_MODELVIEW); glLoadMatrixd(model_matrix);

   // gesetzte Darstellungsfunktion rufen
   if (drawContent != NULL) (*drawContent)();
};

