/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include "constants.h"
#include "ipaddress.h"

#include <string>
#include <string.h>
#include <arpa/inet.h>

IpAddress::IpAddress(void)
/** Konstruktor, setzt Adresse auf "0.0.0.0" */
{
	address.s_addr = INADDR_ANY;
};

IpAddress::IpAddress(in_addr_t new_address)
/** Konstruktor mit &Uuml;bergabe der Adresse 'new_address' als Initialwert */
{
	address.s_addr = new_address;
};

IpAddress::~IpAddress(void)
/** Destruktor, ist leer */
{
};

void IpAddress::getAddress(struct in_addr &current_address)
/** gibt die aktuelle Adresse in 'current_address' zur&uuml;ck */
{
	current_address.s_addr = address.s_addr;
};

char* IpAddress::getAddress(void) const
/** gibt die aktuelle Adresse als Text zur&uuml;ck (inet_ntoa()) */
{
	return inet_ntoa(address);
};

void IpAddress::correct_iptext(char *ip_text)
/** &uuml;berpr&uuml;ft und korrigiert den Text in 'ip_text', der eine
 *  IP-Adresse darstellen soll:
 *  - alle Zeichen ab dem 4. '.' werden gel&ouml;scht (auf 0 gesetzt)
 *  - mehr als drei Zeichen (ungleich '.') hinter dem letzten Punkt:
 *    diese und die Zeichen dahinter werden gel&ouml;scht (auf 0)
 *  - ein Zeichen ungleich Ziffern oder Punkt: dieses und alle folgenden
 *    Zeichen  werden gel&ouml;scht (auf 0)
 *  - zwei Nullen hinter einem Punkt bzw. am Anfang: die Zeichen ab der
 *    zweiten Null werden gel&ouml;scht (auf 0) */
{
	char	num_points, pos, last_point;
	
	num_points = 0;
	pos = 0;
	last_point = -1;
	while(pos < 15)
	{
		if ((ip_text[pos] == '.') && (num_points <= 3))
		{
			last_point = pos;
			num_points++;
		};
		// more than three digits after last point
		if ((num_points == 3) && (pos - last_point > 3))
			num_points = 4;		// erase rest
		// non digit characters (without '.')
		if ((ip_text[pos] != '.') && ((ip_text[pos] < '0') ||
				(ip_text[pos] > '9')))
			num_points = 4;		// erase rest
		if ((ip_text[pos] == '0') && (last_point == pos - 1) &&
			(!((ip_text[pos + 1] == '.') ||
				(ip_text[pos + 1] == 0))))
		{
			// cout << "too many 0" << endl;
			num_points = 4;		// erase rest
			pos++;
		};	
		if (num_points > 3) ip_text[pos] = 0;
		pos++;
	};

};

bool IpAddress::setAddress(const char *ip_string)
/** setzt die interne Adresse auf den Wert in 'ip_string',
 *  dieser wird vorher mit correct_iptext() kontrolliert und bei Bedarf
 *  korrigiert; <br>
 *  wenn nach Korrektur bei der Umwandlung des Textes in eine IP-Adresse mit
 *  inet_addr() immer noch Fehler auftreten, wird 'false' zur&uuml;ckgegeben,
 *  sonst 'true' (fehlerfrei) */
{
	bool		error;
	size_t		string_length;
	char		ip_text[16];
	struct in_addr	test_addr;

	// init values
	error = false;
	memset(ip_text, 0, sizeof(ip_text));
	// get length, copy, correct
	string_length = strlen(ip_string);
	if (string_length > 15) string_length = 15;
	strncpy(ip_text, ip_string, string_length);
	correct_iptext(ip_text);
	// try to convert ascii text to address
	test_addr.s_addr = INADDR_NONE;
	if ((inet_addr(ip_text) == INADDR_NONE) &&
		strncmp(ip_text, inet_ntoa(test_addr), 15) != 0) 
			error = true;
	if (error)
		cout << "IpAddress::setAddress(char*) error? " << error << endl;
	if (!error) address.s_addr = inet_addr(ip_text);
	if (error)
		cout << ip_text << " " << inet_ntoa(test_addr) << " "
			<< address.s_addr << endl;
	return !error;
};

void IpAddress::setAddress(const struct in_addr &new_address)
/** setzt die interne Adresse auf den Wert in 'new_address' */
{
	address.s_addr = new_address.s_addr;
};

void IpAddress::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt eine IP-Adresse aus der XML-Datei 'infile' (Attribut 'ipaddr') */
{
	char	*ip_string;
	xmlChar	*content;

	if (xmlTextReaderHasAttributes(infile))
	{
		content = xmlTextReaderGetAttribute(infile, 
							xml_syntax_ipaddress);
		if (content)
		{
			ip_string = reinterpret_cast<char*>(content);
			setAddress(ip_string);
			//cout << "IpAddress::loadXML " << getAddress() << endl;
			#ifndef WIN32
			delete [] content;
			#endif
		};
	};
};

void IpAddress::saveXML(xmlTextWriterPtr outfile) const
/** speichert die Adresse in der XML-Datei 'outfile' (Attribut 'ipaddr') */
{
	xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_ipaddress, "%s",
						inet_ntoa(address));
};

bool IpAddress::operator <= (const IpAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' <= 'right', sonst 'false' */
{
	return address.s_addr <= right.address.s_addr;
};

bool IpAddress::operator >= (const IpAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' >= 'right', sonst 'false' */
{
	return address.s_addr >= right.address.s_addr;
};

bool IpAddress::operator == (const IpAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' == 'right', sonst 'false' */
{
	return address.s_addr == right.address.s_addr;
};

bool IpAddress::operator < (const IpAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' < 'right', sonst 'false' */
{
	return address.s_addr < right.address.s_addr;
};

bool IpAddress::operator > (const IpAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' > 'right', sonst 'false' */
{
	return address.s_addr > right.address.s_addr;
};

ostream& operator << (ostream &output, const IpAddress &ipaddr)
/** Ausgabe der Adresse 'ipaddr' auf dem Stream 'output' mit inet_ntoa(), <br>
 *  Referenz auf den Stream wird zur&uuml;ckgegeben */
{
	output << ipaddr.getAddress();
	return output;
};

istream & operator >> (istream &input, IpAddress &ipaddr)
/** Einlesen der Adresse 'ipaddr' aus dem Stream 'input' mit setAddress(), <br>
 *  Referenz auf den Stream wird zur&uuml;ckgegeben */
{
	string	input_text;
	
	input >> input_text;
	ipaddr.setAddress(input_text.c_str());
	return input;
};

