/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <fstream>
#include <iostream>
#include <limits.h>
#include <string>

#include "glut.h"
#include "key_control.h"

/** Name der Einstellungsdatei zur Tastaturbelegung */
const char keycode_filename[] = "etc/keycodes.txt";

/** Bezeichnung f&uuml;r die Taste "Cursor links" */
const char cursor_left_text[] = "LEFT";
/** Bezeichnung f&uuml;r die Taste "Cursor rechts" */
const char cursor_right_text[] = "RIGHT";
/** Bezeichnung f&uuml;r die Taste "Cursor auf" */
const char cursor_up_text[] = "UP";
/** Bezeichnung f&uuml;r die Taste "Cursor ab" */
const char cursor_down_text[] = "DOWN";
/** Kommentarzeichen, eine Zeile mit diesem und einem Leerzeichen am Anfang
 *  ('# ') wird als Kommentar gewertet, darf nicht hinter der Definition einer
 *  Tastenbelegung benutzt werden (in der gleichen Zeile wie diese) */
const char keycode_file_comment_char = '#';

/** vordefinierte Tasten als Anfangsbelegung der Tabelle der normalen Tasten */
unsigned char init_normal_keys[] = {27, 'q'};
/** Aktionen zu den vordefinierten Tasten als Anfangsbelegung */
Actions init_normal_keys_actions[] = {Action_Exit, Action_Exit};

Key_Control::Key_Control(void)
/** Konstruktor, ist leer */
{
};

Key_Control::~Key_Control(void)
/** Destruktor, l&ouml;scht alle Tabelleneintr&auml;ge */
{
	normal_keys.clear();
	special_keys.clear();
};

void Key_Control::init(void)
/** Initialisierung: aus den Konstanten 'init_normal_keys' und
 *  'init_normal_keys_actions' wird eine Vorbelegung erstellt,
 *  danach wird die Datei 'keycode_filename' (Tastaturbelegung) eingelesen */
{
	size_t	i;

	for(i = 0; i < sizeof(init_normal_keys); i++)
		normal_keys[init_normal_keys[i]] = init_normal_keys_actions[i];
	load(keycode_filename);
};

Actions Key_Control::stringToAction(const char action_string[])
/** wandelt den Text 'action_string' in eine Aktions(-nummer) um,
 *  Zuordnung erfolgt &uuml;ber das Feld 'Action_Strings' in actions.h, <br>
 *  die gefundene Aktionsnumer oder 'Action_Dummy' (Aktion ohne Wirkung,
 *  keine Aktion zu diesem Text gefunden) wird zur&uuml;ckgegeben */
{
	size_t		i;
	const char	*test_string;
	Actions		found_action;

	found_action = Action_Dummy;
	test_string = Action_Strings[0];
	i = 0;
	while(strlen(test_string) > 0)
	{
		test_string = Action_Strings[i];
		// compare given action_string with entry in Action_Strings
		if ((strncmp(test_string, action_string, 
			strlen(action_string)) == 0) &&
			(strlen(action_string) == strlen(test_string)))
				found_action = Actions(i);
		i++;
	};
	return found_action;
};

bool Key_Control::load(const char filename[])
/** l&auml;dt die Tastenbelegung aus der Datei 'filename',
 *  gibt 'false' zur&uuml;ck, wenn ein Fehler aufgetreten ist (Datei nicht
 *  gefunden), sonst 'true' (fehlerfrei), <br>
 *  ruft u.a. stringToAction() */
{
	bool		error;
	char		*end_ptr;
	unsigned char	new_normal_keycode, new_special_keycode;
	string		action_name, keycode_text, line;
	ifstream	keycode_file;

	error = false;
	cout << "Loading key commands from: " << filename << endl;
	keycode_file.open(filename, ios::in);
	if (!keycode_file.is_open()) error = true;
	if (!error)
	{
		while(keycode_file)
		{
			// test first character for comment
			if (keycode_file.peek() == keycode_file_comment_char)
				keycode_file.ignore(INT_MAX, '\n');
				// std::numeric_limits<int>::max(), '\n');
			// read in keycode and action
			keycode_file >> keycode_text;
			keycode_file >> action_name;
			new_normal_keycode = 0;
			new_special_keycode = 0;
			end_ptr = NULL;
			if (keycode_text == "" + keycode_file_comment_char)
				continue;
			// test for special keycode
			if (keycode_text.compare(cursor_left_text) == 0)
				new_special_keycode = GLUT_KEY_LEFT;
			else if (keycode_text.compare(cursor_right_text) == 0)
				new_special_keycode = GLUT_KEY_RIGHT;
			else if (keycode_text.compare(cursor_up_text) == 0)
				new_special_keycode = GLUT_KEY_UP;
			else if (keycode_text.compare(cursor_down_text) == 0)
				new_special_keycode = GLUT_KEY_DOWN;
			else if ((keycode_text.size() > 1) &&
				 (keycode_text[0] == 'F'))
			{	// F1 to F12
				keycode_text = keycode_text.substr(1, 2);
				new_special_keycode = strtoul(
					keycode_text.c_str(), &end_ptr, 10);
			}
			// normal keycode
			else if ((keycode_text.size() == 1) &&
				((keycode_text[0] < '0') ||
				(keycode_text[0] > '9')))
					new_normal_keycode = keycode_text[0];
			else if ((keycode_text.size() == 3) && 
				(keycode_text[0] == '\''))
					new_normal_keycode = keycode_text[1];
			else	new_normal_keycode = 
					strtoul(keycode_text.c_str(), 
						&end_ptr, 0) % 255;
			// keycode to action
			if (new_normal_keycode != 0)
				normal_keys[new_normal_keycode] =
					stringToAction(action_name.c_str());
			if (new_special_keycode != 0)
				special_keys[new_special_keycode] =
					stringToAction(action_name.c_str());
			// cout << "Keycode: " << new_normal_keycode << " ";
			// cout << "action number: " << 
			//	stringToAction(action_name.c_str()) << " ";
			// cout << "action: " << action_name << endl;
		};
	};
	if (keycode_file.is_open()) keycode_file.close();
	return !error;
};

Actions Key_Control::translateKeycode(unsigned char normal_key)
/** gibt zur normalen Taste 'normal_key' die zugeh&ouml;rige Aktion laut
 *  interner Tabelle 'normal_keys' zur&uuml;ck */
{
	return normal_keys[normal_key];
};

Actions Key_Control::translateKeycode(int special_key)
/** gibt zur Sondertaste 'special_key' die zugeh&ouml;rige Aktion laut
 *  interner Tabelle 'special_keys' zur&uuml;ck */
{
	return special_keys[special_key];
};

