/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include "constants.h"
#include "macaddress.h"

MacAddress::MacAddress()
/** Konstruktor, ruft nur init() */
{
	init();
};

MacAddress::MacAddress(const char mac_string[])
/** Konstruktor mit &Uuml;bergabe des Adresswertes in 'mac_string',
 *  ruft nur setAddress() */ 
{
	setAddress(mac_string);
};

MacAddress::~MacAddress()
/** Destruktor, ist leer */
{
};

void MacAddress::init(void)
/** Initialisierung: setzt die interne Adresse auf '00:00:00:00:00:00' */
{
	memset(&address, 0, sizeof(address));
};

void MacAddress::getAddress(struct ether_addr &current_address) const
/** gibt die aktuelle interne Adresse in 'current_address' zur&uuml;ck */
{
	memcpy(&current_address, &address, sizeof(address));
};

string MacAddress::getAddress(void) const
/** gibt die aktuelle interne Adresse als Textstring zur&uuml;ck */
{
	size_t		i;
	char		mac_text[18];
	string		mac_string;

	// cout << "ETH_ALEN: " << ETH_ALEN << endl;
	memset(mac_text, 0, sizeof(mac_text));
	for(i = 0; i < ETH_ALEN; i++)
	{
		mac_text[3 * i] = '0';
		mac_text[3 * i + 1] = '0';
		// octet to ascii text
		snprintf(&(mac_text[3 * i]), 3, "%.2x", 
				address.ether_addr_octet[i]);
		// insert ':'
		if ((i + 1) < ETH_ALEN) mac_text[3 * i + 2] = ':';
	};
	mac_string = mac_text;
	// cout << "Mac Address: " << mac_string << "|" << endl;
	return mac_string;
};

bool MacAddress::setAddress(const char mac_string[])
/** setzt die interne Adresse auf den Wert in 'mac_string', <br>
 *  gibt 'false' zur&uuml;ck, wenn Fehler aufgetreten ist (keine richtige
 *  Ethernetadresse in 'mac_string'), sonst 'true' (fehlerfrei) */
{
	bool		error;
	size_t		i;
	char		end_char, *end;
	char		mac_text[18];
	const char	*begin;

	if (mac_string == NULL) return false;
	// test length of text
	if (strlen(mac_string) < 17) return false;// too short AA:BB:CC:DD:EE:FF
	memset(mac_text, 0, sizeof(mac_text));
	strncpy(mac_text, mac_string, sizeof(mac_text) - 1);
	error = false;
	for(i = 0; i < 17; i++)
	{
		if ((i + 1) % 3 == 0)
		{
			if (mac_text[i] != ':') error = true;
		}
		else
		{	// wrong character, only
			// '0' - '9', 'A' - 'F' and 'a' - 'f' is allowed
			if (mac_text[i] < '0') error = true;
			if ((mac_text[i] > '9') && (mac_text[i] < 'A'))
				error = true;
			if ((mac_text[i] > 'F') && (mac_text[i] < 'a'))
				error = true;
			if (mac_text[i] > 'f') error = true;
		};
	}
	if (error)
		cout << "MacAddress::setAddress(string) Error? " << error 
			<< endl;
	if (!error)
		for(i = 0; i < 6; i++)
		{
			begin = &(mac_text[i * 3]);
			end = &end_char;
			// ascii text to octet
			address.ether_addr_octet[i] = strtol(begin, &end, 16);
		};
	if (error)
	       cout << "address: " << mac_text << " " << getAddress() << endl;
	return !error;
};

void MacAddress::setAddress(struct ether_addr &new_address)
/** setzt die interne Adresse auf den Wert in 'new_address' */
{
	memcpy(&address, &new_address, sizeof(address));
};

void MacAddress::loadXML(xmlTextReaderPtr infile)
/** liest eine Ethernet-Adresse aus der XML-Datei 'infile' ein
 *  (Attribut macaddr) */
{
        char    *mac_string;
        xmlChar *content;
                                                                                
        if (xmlTextReaderHasAttributes(infile))
        {
                content = xmlTextReaderGetAttribute(infile,
                                                        xml_syntax_macaddress);
                if (content)
                {
                        mac_string = reinterpret_cast<char*>(content);
                        setAddress(mac_string);
                        // cout << "MacAddress::loadXML " 
			//	<< getAddress() << endl;
			#ifndef WIN32
                        delete [] content;
			#endif
                };
        };
};

void MacAddress::saveXML(xmlTextWriterPtr outfile) const
/** speichert die Ethernetadresse in die XML-Datei 'outfile' (Attribut macaddr)
 */
{
	string	mac_string;

	mac_string = getAddress();
        xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_macaddress, "%s",
                                                mac_string.c_str());
};

bool MacAddress::operator <= (const MacAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' <= 'right', sonst 'false' */
{
	return (memcmp(&address, &right.address, sizeof(address)) <= 0);
};

bool MacAddress::operator >= (const MacAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' >= 'right', sonst 'false' */
{
	return (memcmp(&address, &right.address, sizeof(address)) >= 0);
};

bool MacAddress::operator == (const MacAddress &right) const
/** Vergleich, liefert 'true', wenn 'left' == 'right', sonst 'false' */
{
	return (memcmp(&address, &right.address, sizeof(address)) == 0);
};

ostream& operator << (ostream &output, const MacAddress &macaddr)
/** Ausgabe der Adresse 'macaddr' auf dem Stream 'output' mit getAddress(),
 *  <br> Referenz auf den Stream wird zur&uuml;ckgegeben */
{
	output << macaddr.getAddress();
	return output;
};

istream& operator >> (istream &input, MacAddress &macaddr)
/** Einlesen der Adresse 'macaddr' aus dem Stream 'input' mit setAddress(),
 *  <br> Referenz auf den Stream wird zur&uuml;ckgegeben */
{
	string	input_text;

	input >> input_text;
	macaddr.setAddress(input_text.c_str());
	return input;
};

