/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <math.h>

#include "arpcache.h"
#include "constants.h"

// needed to set Stroke_Font size for output
/** maximale Anzahl Textzeichen je Zeile im dargestellten ARP-Cache
 *  (15 Zeichen (4*3+3) IP-Adresse und 17 Zeichen (6*2+5) Ethernet-Adresse
 *  plus Leerzeichen) */
const unsigned short	arp_cache_max_text_length = 35;
/** maximale Anzahl Textzeilen im dargestellten ARP-Cache (nur dargestellter
 *  Bereich, Tabelle kann gr&ouml;&szlig;er sein) */
const unsigned short	arp_cache_max_text_lines = 8;

/** Farbe des Textes der dargestellten Tabelle, RGBA 0.0 - 1.0, <br>
 *  hier: wei&szlig; und undurchsichtig */
const double		arp_cache_text_color[4] = {0.95, 0.95, 0.95, 1.0};
/** Gr&ouml;&szlig;e des dargestellten Textes gesamt (derzeit 35 Zeichen mal
 *  8 Zeilen), zur Berechnung der Zeilenanfangspositionen */
const double		arp_cache_text_size[3] = {2.0, 1.0, 0.0};
/** maximaler Abstand der dargestellten Tabelle zur Kamera, bei dem mit Level
 *  of Detail noch etwas dargestellt wird */
const double		arp_cache_max_detail_distance = 3.0;
/** Farbe des Texthintergrundes einer Zeile, RGBA 0.0 - 1.0, wenn Tabellenzeile
 *  nicht den eben verwendeten Eintrag (last_searched_ip) enth&auml;lt <br>
 *  hier: gr&uuml;n und teildurchsichtig */
const double		arp_cache_color_not_same[4] = {0.0, 0.8, 0.0, 0.8};
/** Farbe des Texthintergrundes einer Zeile, RGBA 0.0 - 1.0, wenn Tabellenzeile
 *  mit dem gerade gesuchten Eintrag (last_searched_ip) &uuml;bereinstimmt;
 *  <br> hier: rot und teildurchsichtig */
const double		arp_cache_color_found[4] = {1.0, 0.0, 0.0, 0.8};

ArpCache::ArpCache(void)
/** Konstruktor, ist leer */
{
};

ArpCache::~ArpCache(void)
/** Destruktor, ist leer */
{
};

void ArpCache::drawColorBackground(bool is_same_ip,
					class Rectangle &background) const
/** setzt Farbe des Rechteckes 'background' f&uuml;r den Texthintergrund,
 *  je nachdem, ob die Zeile gerade zur &Uuml;bersetzung verwendet wurde,
 *  'is_same_ip' auf 'true' -> color_found,
 *  oder nicht ('false') -> color_not_same */
{
	if (is_same_ip)
	{	// is_same_ip -> color_found
		background.setColor(arp_cache_color_found[0],
					arp_cache_color_found[1],
					arp_cache_color_found[2],
					arp_cache_color_found[3]);
	}
	else
	{	// not is_same_ip -> color_not_same
		background.setColor(arp_cache_color_not_same[0],
					arp_cache_color_not_same[1],
					arp_cache_color_not_same[2],
					arp_cache_color_not_same[3]);
	}
};

void ArpCache::draw(bool level_of_details_on, Point_3D camera_pos) const
/** zeichnet den ARP-Cache, wenn Zustand 'level_of_details_on' auf aus oder
 *  Kamera 'camera_pos' nah genug dran ist */
{
	double						x_pos, y_pos, z_pos,
							sign, char_width;
	size_t						i;
	string						entry_line;
	IpAddress					entry_ip;
	MacAddress					entry_mac;
	Point_3D					start, pos;
	class Rectangle					background;
	map<IpAddress, MacAddress>::const_iterator	entry_iterator;
	pair<IpAddress, MacAddress>			pair_ip_mac;

	if ((!level_of_details_on) || (camera_pos.getDistance(text_pos) <= 
					arp_cache_max_detail_distance))
	{
		// calculate top left position, where begin to draw
		x_pos = text_pos.getX() - text_size.getX() / 2.0;
		y_pos = text_pos.getY() + text_size.getY() / 2.0;
		z_pos = text_pos.getZ();
		start.setPos(x_pos, y_pos, z_pos);
		start.translate(1);
		// character width and line height
		char_width = text_size.getX() / arp_cache_max_text_length;
		y_pos = - text_size.getY() / arp_cache_max_text_lines;
		// sign: draw background before or in front of text
		sign = text_pos.getZ() / fabs(text_pos.getZ());
		pos.setPos(0.0, y_pos, 0.0);	// only line height
		// set size of background rectangle for first line
		background.setPos(0.0, -y_pos, text_size.getX() / 2, 0.0);
		glTranslated(0.0, -0.01, -0.001 * sign);
		drawColorBackground(true, background);
		background.draw();
		glTranslated(0.0, 0.01, 0.001 * sign);
		// first line of output: last searched IP address
		cache_text_output.print(last_searched_ip.getAddress());
		// set size of background for rest of the lines
		background.setPos(0.0, 0.0, text_size.getX(), y_pos);
		i = 0;
		entry_iterator = ip_mac_map.begin();
		while(entry_iterator != ip_mac_map.end())
		{
			pos.translate(i * 1.0);
			glTranslated(0.0, -0.01, -0.001 * sign);
			pair_ip_mac = *entry_iterator;
			// choose color and draw background
			drawColorBackground(
				pair_ip_mac.first == last_searched_ip,
				background);
			background.draw();
			glTranslated(0.0, 0.01, 0.001 * sign);
			glTranslated(char_width, 0.0, 0.0);
			pos.translate(1.0);
			// concat text line and draw it
			entry_line = pair_ip_mac.first.getAddress();
			entry_line += " " + pair_ip_mac.second.getAddress();
			cache_text_output.print(entry_line.c_str());
			pos.translate((i + 1) * (-1.0));
			glTranslated(-char_width, 0.0, 0.0);
			entry_iterator++;
			i++;
		};
		start.translate(-1);
	};
};

void ArpCache::init(void)
/** Initialisierung: setColor(), setSize(), Position der Textausgabe */
{
	setColor(arp_cache_text_color[0], arp_cache_text_color[1],
		arp_cache_text_color[2], arp_cache_text_color[3]);
	setSize(arp_cache_text_size[0], arp_cache_text_size[1],
		arp_cache_text_size[2]);
	cache_text_output.setPos(0.0, 0.0, 0.0);
};

void ArpCache::setColor(double new_red, double new_green,
				double new_blue, double new_alpha)
/** setzt die Farbe des dargestellten Textes auf 'new_red' (Rot),
 *  'new_green' (Gr&uuml;n), 'new_blue' (Blau) und 'new_alpha'
 *  (Transparenz, 1.0 undurchsichtig), Wertebereich jeweils 0.0 - 1.0 */
{
	cache_text_output.setColor(new_red, new_green, new_blue, new_alpha);
};

void ArpCache::setPos(double new_centerx, double new_centery,
				double new_centerz)
/** setzt die Position des dargestellten Textes (Mittelpunkt des rechteckigen
 *  Bereiches) x,y,z */
{
	text_pos.setPos(new_centerx, new_centery, new_centerz);
};

void ArpCache::setSize(double new_width, double new_height,
				double new_depth)
/** setzt die Gr&ouml;&szlig;e des dargestellten Textes (Bounding-Box),
 *  'new_width' = Breite, 'new_height' = H&ouml;he, 'new_depth' = Tiefe, <br>
 *  die Ausma&szlig;e des gesamten Textbereiches, nicht nur eines Zeichens */
{
	cache_text_output.setSize(new_width / arp_cache_max_text_length,
				new_height / arp_cache_max_text_lines / 1.5);
	text_size.setPos(new_width, new_height, new_depth);
};

bool ArpCache::findMac(const IpAddress &ipaddress,
			MacAddress &macaddress)
// search in the table for the translation of 'ipaddress' to 'macaddress'
/** sucht in der Tabelle des ARP-Caches nach der Ethernet-Adresse 'macaddress'
 *  zur bekannten IP-Adresse 'ipaddress'; gibt 'true' zur&uuml;ck, wenn eine
 *  gefunden wurde, sonst 'false' */
{
	bool						found;
	map<IpAddress, MacAddress>::const_iterator	found_ip;
	pair<IpAddress, MacAddress>			pair_ip_mac;

	found = false;
	cout << "Searching Mac for " << ipaddress << endl;
	last_searched_ip = ipaddress;
	found_ip = ip_mac_map.find(ipaddress);
	if (found_ip != ip_mac_map.end()) found = true;
	// cout << "Found: " << found << endl;
	if (found)
	{
		pair_ip_mac = *found_ip;
		macaddress = pair_ip_mac.second;
	};
	// if (found) cout << "Found Mac: " << macaddress << endl;
	return found;
};

void ArpCache::insertIpMac(IpAddress &ipaddress, MacAddress &macaddress)
// fill in a translation entry ipaddress to macaddress, after an ARP reply
/** erweitert die Tabelle um einen Eintrag: zur IP-Adresse 'ipaddress'
 *  geh&ouml;rt die Ethernet-Adresse 'macaddress'; normalerweise nach einem
 *  ARP-Antwortpaket */
{
	// cout << "Insert in arp cache: " << ipaddress << " - " << macaddress
	//	<< endl;
	ip_mac_map[ipaddress] = macaddress;
};

void ArpCache::getInfos(Variable_Tree* &infos)
/** zur Zeit leer, gibt nur Zeiger auf den eigenen Baum zur&uuml;ck; <br>
 *  stellt den Variablenbaum f&uuml;r das Dialogfenster zusammen, Zeiger
 *  darauf in 'infos' zur&uuml;ck */
{
	infos = &current_infos;
};

void ArpCache::setInfos(Variable_Tree* infos)
/** zur Zeit leer, keine Einstellungen &auml;nderbar; <br>
 *  liest die Werte aus dem Variablenbaum ein, der Zeiger 'infos' sollte
 *  auf den internen Baum zeigen, wird aber nicht beachtet */
{
};

void ArpCache::loadXML(xmlTextReaderPtr infile)
/** zur Zeit werden keine Tags und Attribute geladen (und gespeichert), <br>
 *  l&auml;dt die Einstellungen und Werte des ARP-Caches aus der
 *  XML-Datei 'infile' */
{
	int	state, depth;

	do
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
	} while ((state >= 0) && (depth > 3));
};

void ArpCache::saveXML(xmlTextWriterPtr outfile) const
/** speichert zur Zeit nur den Tag 'xml_syntax_arp_cache', ohne Attribute oder
 *  Subtags; <br>
 *  speichert die Einstellungen und Werte des ARP-Caches in der
 *  XML-Datei 'outfile' */
{
	xmlTextWriterStartElement(outfile, xml_syntax_arp_cache);
	xmlTextWriterEndElement(outfile);
};

