/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include "constants.h"
#include "interface.h"

/** Name der Variablen zum Name der Netzwerkkarte im Info-Variablenbaum zum
 *  Dialog */
const char	interface_varname_name[] = "interface name";
/** Name des Teilbaumes der IP-Warteschlangen im Info-Variablenbaum zum
 *  Dialog */
const char	interface_varname_ip_queue[] = "ip queue";
/** Name des Teilbaumes der Ethernet-Warteschlangen im Info-Variablenbaum zum
 *  Dialog */
const char	interface_varname_eth_queue[] = "eth queue";
/** Name des Teilbaumes des ARP-Cache im Info-Variablenbaum zum Dialog */
const char	interface_varname_arp_cache[] = "arp cache";

Interface::Interface(void) : name("")
/** Konstruktor, Initialisierung des Namens der Karte */
{
};

Interface::~Interface(void)
/** Destruktor, ist leer */
{
	// cout << "Destruktor interface" << endl;
};

void Interface::updatePos(void)
/** Position und Gr&ouml;&szlig;e (nur ARP-Cache) der einzelnen grafischen
 *  Elemente (IP- und Ethernet-Warteschlangen, ARP-Cache) wird neu bestimmt,
 *  abh&auml;ngig von Gr&ouml;&szlig;e und Position der dargestellten
 *  Netzwerkkarte */
{
	ip_queue.setPos(position.getX(),
		position.getY() + graphic_size.getY() / 4.0, position.getZ());
	eth_queue.setPos(position.getX(),
		position.getY() - graphic_size.getY() / 4.0, position.getZ());
	arp_cache.setPos(position.getX(),
		position.getY() + graphic_size.getY() / 8.0,
		position.getZ() + graphic_size.getZ() / 1.5);
	arp_cache.setSize(1.0, 0.5, 0.0);
};

void Interface::setPos(double posx, double posy, double posz)
/** setzt die Position (den Mittelpunkt) der Netzwerkkarte (x, y, z), <br>
 *  ruft updatePos() */
{
	position.setPos(posx, posy, posz);
	updatePos();
};

void Interface::setSize(double new_width, double new_height, double new_depth)
/** setzt die Gr&ouml;&szlig;e der Netzwerkkarte (Bounding-Box),
 *  'new_width' = Breite, 'new_height' = H&ouml;he, 'new_depth' = Tiefe, <br>
 *  Gr&ouml;&szlig;e der IP- und Ethernet-Warteschlangen wird neu berechnet,
 *  ruft updatePos() */
{
	graphic_size.setPos(new_width, new_height, new_depth);
	ip_queue.setSize(new_width, new_height / 2.0, new_depth);
	eth_queue.setSize(new_width, new_height / 2.0, new_depth);
	updatePos();
};

void Interface::drawArpCache(bool level_of_details_on,
				Point_3D camera_pos) const
/** zeichnet den ARP-Cache der Netzwerkkarte, ruft nur ArpCache::draw() */
{
	arp_cache.draw(level_of_details_on, camera_pos);
};

void Interface::drawPackets(bool level_of_details_on,
				Point_3D camera_pos) const
/** zeichnet die Pakete der Warteschlangen, ruft nur IpQueue::drawPackets() und
 *  EthernetQueue::drawPackets() */
{
	ip_queue.drawPackets(level_of_details_on, camera_pos);
	eth_queue.drawPackets(level_of_details_on, camera_pos);
};

void Interface::drawQueue(void) const
/** zeichnet die Quader und Beschriftung der Warteschlangen, ruft nur
 *  IpQueue::drawQueue() und EthernetQueue::drawQueue() */
{
	ip_queue.drawQueue();
	eth_queue.drawQueue();
};

void Interface::init(void)
/** Initialisierung: Name der Karte, des ARP-Caches und der Warteschlangen,
 *  ruft ArpCache::init(), IpQueue::init(), EthernetQueue::init() */
{
	// cout << "Interface::init start" << endl;
	name = "eth0";
	arp_cache.init();
	ip_queue.init();
	eth_queue.init();
	// cout << "Interface::init end" << endl;
};

bool Interface::setIpAddress(const char ip_string[])
/** setzt die IP-Adresse der IP-Schlange der Netzwerkkarte auf die in
 *  'ip_string', gibt 'true' zur&uuml;ck wenn &uuml;bernommen,
 *  'false' bei Fehler; <br>
 *  ruft nur IpQueue::setAddress() und gibt dessen Wert zur&uuml;ck */
{
	return ip_queue.setAddress(ip_string);
};

char* Interface::getIpAddress(void) const
/** gibt die IP-Adresse der IP-Schlange der Netzwerkkarte als Text zur&uuml;ck,
 *  ruft nur IpQueue::getAddress() */
{
	return ip_queue.getAddress();
};

void Interface::getIpAddress(IpAddress &current_ip) const
/** gibt die IP-Adresse der IP-Schlange der Netzwerkkarte in 'current_ip'
 *  zur&uuml;ck, ruft nur IpQueue::getAddress() */
{
	ip_queue.getAddress(current_ip);
};

string Interface::getMacAddress(void) const
/** gibt die Ethernet-Adresse der Ethernet-Schlange der Netzwerkkarte als Text
 *  zur&uuml;ck, ruft nur EthernetQueue::getAddress() */
{
	return eth_queue.getAddress();
};

void Interface::getMacAddress(MacAddress &current_address) const
/** gibt die Ethernet-Adresse der Ethernet-Schlange der Netzwerkkarte in
 *  'current_address' zur&uuml;ck, ruft nur EthernetQueue::getAddress() */
{
	eth_queue.getAddress(current_address);
};

bool Interface::setMacAddress(const char mac_string[])
/** setzt die Ethernet-Adresse der Ethernet-Schlange der Netzwerkkarte auf die
 *  in 'mac_string', gibt 'true' zur&uuml;ck wenn &uuml;bernommen,
 *  'false' bei Fehler; <br>
 *  ruft nur EthernetQueue::setAddress() und gibt dessen Wert zur&uuml;ck */
{
	return eth_queue.setAddress(mac_string);
};

string Interface::getName(void) const
/** gibt den Namen der Netzwerkkarte als Text zur&uuml;ck */
{
	return name;
};

void Interface::setName(string new_name)
/** setzt den Namen der Netzwerkkarte auf 'new_name' */
{
	name = new_name;
};

void Interface::getInfos(Variable_Tree* &infos)
/** stellt den Variablenbaum f&uuml;r das Dialogfenster zusammen, Zeiger
 *  darauf in 'infos' zur&uuml;ck */
{
	String_Variable	*var_name;
	Variable_Tree	*var_tree_ip_queue, *var_tree_eth_queue, 
			*var_tree_arp_cache;

	current_infos.clear();
	var_name = new String_Variable;
	if (var_name)
	{	// name of the interface card
		var_name->setName(interface_varname_name);
		var_name->setLength(255, 0);
		var_name->setValue(name);
		current_infos.addVar(var_name);
	};
	// IpQueue, ArpCache and EthernetQueue
	ip_queue.getInfos(var_tree_ip_queue);
	if (var_tree_ip_queue) 
		var_tree_ip_queue->setName(interface_varname_ip_queue);
	current_infos.addChild(var_tree_ip_queue);
	arp_cache.getInfos(var_tree_arp_cache);
	if (var_tree_arp_cache)
		var_tree_arp_cache->setName(interface_varname_arp_cache);
	current_infos.addChild(var_tree_arp_cache);
	eth_queue.getInfos(var_tree_eth_queue);
	if (var_tree_eth_queue)
		var_tree_eth_queue->setName(interface_varname_eth_queue);
	current_infos.addChild(var_tree_eth_queue);
	infos = &current_infos;
};

void Interface::setInfos(Variable_Tree* infos)
/** liest die Werte aus dem Variablenbaum ein, der Zeiger 'infos' sollte
 *  auf den internen Baum zeigen, wird aber nicht beachtet */
{
	string		value_text;
	Variable	*current_var;
	Variable_Tree	*current_tree;

	if (current_infos.findVar(interface_varname_name, current_var))
	{	// name of the interface card
		value_text = current_var->getValue();
		name = value_text;
	};
	// IpQueue, ArpCache and EthernetQueue
	if (current_infos.findChild(interface_varname_ip_queue, current_tree))
		ip_queue.setInfos(current_tree);
	if (current_infos.findChild(interface_varname_arp_cache, current_tree))
		arp_cache.setInfos(current_tree);
	if (current_infos.findChild(interface_varname_eth_queue, current_tree))
		eth_queue.setInfos(current_tree);
};

void Interface::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt die Einstellungen der Netzwerkkarte aus der XML-Datei 'infile',
 *  Kartenname, Wert von sending_arp_packet, IP-Warteschlange, ARP-Cache und
 *  Ethernet-Warteschlange */
{
	int		state, depth;
	unsigned long	value;
	char		*value_text, *end_ptr;
	xmlChar		*content;

	// cout << "Interface::loadXML start" << endl;
	if (xmlTextReaderHasAttributes(infile))
	{	// name of the interface card, value of sending_arp_packet
		content = xmlTextReaderGetAttribute(infile, 
						xml_syntax_name);
		if (content) name = reinterpret_cast<char*>(content);
		content = xmlTextReaderGetAttribute(infile,
				xml_syntax_interface_sending_arp_packet);
		value_text = NULL;
		if (content) value_text = reinterpret_cast<char*>(content);
		end_ptr = value_text;
		if (value_text)
		{
			value = strtoul(value_text, &end_ptr, 10);
			sending_arp_packet = (value != 0);
		};
	};
	do
	{	// IpQueue, ArpCache and EthernetQueue
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
                                        xml_syntax_ip_queue) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
				ip_queue.loadXML(infile);
                else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
                                        xml_syntax_ethernet_queue) == 0) &&
                        (xmlTextReaderNodeType(infile) == 1))
                                eth_queue.loadXML(infile);
		else if (xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_arp_cache) == 0)
				arp_cache.loadXML(infile);
	} while ((state >= 0) && (depth > 2));
	// cout << "Interface::loadXML end" << endl;
};

void Interface::saveXML(xmlTextWriterPtr outfile) const
/** speichert die Einstellungen der Netzwerkkarte in die XML-Datei 'outfile',
 *  Kartenname, Wert von sending_arp_packet, IP-Warteschlange, ARP-Cache und
 *  Ethernet-Warteschlange */
{
	xmlTextWriterStartElement(outfile, xml_syntax_interface);
	// name of the interface card, value of sending_arp_packet
	xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_name, 
						"%s", name.c_str());
	xmlTextWriterWriteFormatAttribute(outfile,
				xml_syntax_interface_sending_arp_packet,
				"%d", sending_arp_packet);
	// IpQueue, ArpCache and EthernetQueue
	ip_queue.saveXML(outfile);
	arp_cache.saveXML(outfile);
	eth_queue.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
};

void Interface::down_in(IPv4_Packet* incoming_packet)
/** zu sendendes Paket 'incoming_packet' kommt ans Ende der IP-Warteschlange
 *  zum Senden, wenn der TTL-Wert stimmt (time to live); <br>
 *  wenn die IP-Adresse der Karte mit der Zieladresse &uuml;bereinstimmt, kommt
 *  das Paket ans Ende der IP-Warteschlange zum Empfangen; <br>
 *  bei falschem TTL-Wert (gleich 0) wird das Paket gel&ouml;scht, <br>
 *  damit geben Host und L3_Router ihre Pakete zum Senden an die Karte */
{
	IpAddress	destination_ip, my_ip;

	if (incoming_packet != NULL)
	{	// check ttl, delete if too low
		// cout << "Interface::down_in" << endl;
		if (incoming_packet->getTTL() == 0)
		{
			cout << "Interface: ip_packet has TTL = 0. "
				<< "Deleting it." << endl;
			delete incoming_packet;
			incoming_packet = NULL;
		};
	};
	if (incoming_packet != NULL)
	{	// check destination address -> to end of send queue or
		// receive queue
		incoming_packet->getDestinationIpAddress(destination_ip);
		ip_queue.getAddress(my_ip);
		if (!(destination_ip == my_ip))
			ip_queue.down_in(incoming_packet);
		else
		{
			cout << "packet send from me to me" << endl;
			ip_queue.up_in(incoming_packet);
		};
	};
};

void Interface::down_out(Ethernet_Packet* &outgoing_packet)
/** gibt erstes Paket 'outgoing_packet' der Ethernet-Sendeschlange zur&uuml;ck,
 *  wenn es in der Animation am Ende (der Schlange) angekommen ist, sonst NULL,
 *  ruft nur EthernetQueue::down_out(), <br>
 *  Paket geht dann nicht mehr &uuml;ber in das angeschlossene Netzwerkkabel,
 *  sondern wird abgefangen */
{
	eth_queue.down_out(outgoing_packet);
};

void Interface::up_in(Ethernet_Packet* incoming_packet)
/** neues Paket 'incoming_packet' zum Empfangen, kommt ans Ende der
 *  Ethernet-Empfangsschlange, normalerweise kommt Paket &uuml;ber ein Kabel
 *  in diese Schlange, hiermit auch ohne Kabel; <br>
 *  ruft nur EthernetQueue::up_in() */
{
	eth_queue.up_in(incoming_packet);
};

void Interface::up_out(IPv4_Packet* &outgoing_packet)
/** Paket am Anfang der IP-Empfangsschlange wird in 'outgoing_packet'
 *  zur&uuml;ckgegeben, wenn es in der Animation am Ende der Schlange
 *  angekommen ist, sonst NULL; <br>
 *  ruft nur IpQueue::up_out(); <br>
 *  damit holen Host und L3_Router ihre Pakete von den Karten ab */
{
	ip_queue.up_out(outgoing_packet);
};

bool Interface::getFirstPacketPos(Point_3D &packet_pos) const
/** gibt Position 'packet_pos' des ersten Paketes der Sende- bzw.
 *  Empfangsschlangen (falls keines zum Senden) zur&uuml;ck, 'false' wenn
 *  kein Paket da, <br> ruft EthernetQueue::getFirstPacketPos() und evtl.
 *  IpQueue::getFirstPacketPos() (wenn Ethernetschlange keine Pakete hat) */
{
	bool	has_packets;

	has_packets = eth_queue.getFirstPacketPos(packet_pos);
	if (!has_packets)
		has_packets = ip_queue.getFirstPacketPos(packet_pos);
	return has_packets;
};

void Interface::attachCable(Net_Cable *new_cable)
/** verbindet Netzwerkkabel 'new_cable' mit der Netzwerkkarte, &uuml;bernimmt
 *  die Richtung des Kabels am Anfang oder Ende (je nach Ethernet-Adresse)
 *  plus 90 Grad, <br>
 *  bei angegebenem NULL-Zeiger in 'new_cable' wird aktuelle Kabelverbindung
 *  gel&ouml;st */
{
	MacAddress	begin_address, end_address, mac_address;

	// cout << "Interface::attachCable start" << endl;
	// cout << "Attaching Cable: " << new_cable << endl;
	if (new_cable)
	{	// take direction of cable on begin or end (where mac_address
		// matches) 
		getMacAddress(mac_address);
		setAngle(new_cable->getAngle(mac_address) + 90.0);
		new_cable->getAddresses(begin_address, end_address);
		// cout << "Begin: " << begin_address 
		// 	<< " End: " << end_address;
		// cout << " my MAC: " << getMacAddress() << endl;
	};
	attached_cable = new_cable;
};

double Interface::getAngle(void) const
/** gibt den aktuellen Rotationswinkel der Karte zur&uuml;ck, 0.0 - 360.0 Grad
 */
{
	return direction_angle;
};

void Interface::setAngle(double new_angle)
/** setzt den Rotationwinkel der Karte, 0.0 bis 360.0 Grad,
 *  derzeitiger Zielwinkel = aktueller Rotationswinkel = letzter Zielwinkel,
 *  needed_angle = direction_angle = last_needed_angle = new_angle */
{
	// cout << "Interface: current angle: " << direction_angle << " ";
	// cout << "new angle: " << new_angle << endl;
	if (new_angle != direction_angle) last_needed_angle = new_angle;
	direction_angle = new_angle;
	needed_angle = direction_angle;
};

bool Interface::needsTurn(void) const
/** gibt 'true' zur&uuml;ck, wenn die Netzwerkkarte sich drehen will,
 *  damit sie die Pakete in korrekter Richtung abschicken bzw. empfangen kann,
 *  <br> Senderichtung ist entgegengesetzt zu Empfangsrichtung, <br>
 *  Zielwinkel needed_angle ungleich aktuellem Winkel direction_angle */
{
	// cout << "Interface::needsTurn " << direction_angle << " ";
	// cout << needed_angle << endl;
	return (direction_angle != needed_angle);
};

size_t Interface::numPackets(void) const
/** gibt die gesamte Anzahl der Pakete in den Warteschlangen zur&uuml;ck,
 *  ruft nur Interface_ExecuteNet::numPackets() */
{
	return Interface_ExecuteNet::numPackets();
};

size_t Interface::numPacketsOnCable(void) const
// number of packets on cable, if first is for my address
/** gibt die Anzahl der Pakete auf dem angeschlossenen Netzwerkkabel
 *  zur&uuml;ck, wenn das erste Paket im Kabel an diese Netzwerkkarte gerichtet
 *  ist, ruft nur Interface_ExecuteNet::numPacketsOnCable() */
{
	return Interface_ExecuteNet::numPacketsOnCable();
};

void Interface::executeNet(double speed_factor)
/** f&uuml;hrt einen Schritt der Netzwerksimulation in der Karte aus,
 *  'speed_factor' beeinflusst die Geschwindigkeit der Drehung und des
 *  Pakettransports, <br>
 *  ruft nur Interface_ExecuteNet::executeNet() */
{
	Interface_ExecuteNet::executeNet(speed_factor);
};

