/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <math.h>

#include "interface_executenet.h"

/** Anfangswert f&uuml;r die Schrittweite der Drehung der Karte */
const double	interface_executenet_init_angle_step = 5.0;
/** Multiplikationsfaktor f&uuml;r die Drehgeschwindigkeit,
 *  siehe Interface_ExecuteNet::executeNet() */
const double	interface_executenet_angle_step_factor = 20.0;

Interface_ExecuteNet::Interface_ExecuteNet(void) : 
		direction_angle(0.0),
		angle_step(interface_executenet_init_angle_step),
		needed_angle(0.0), last_needed_angle(0.0),
		animation_speed_factor(interface_executenet_angle_step_factor),
		sending_arp_packet(false), attached_cable(NULL)
/** Konstruktor, Initialisierung: direction_angle, angle_step, needed_angle,
 *  last_needed_angle, animation_speed_factor, sending_arp_packet,
 *  attached_cable */
{
};

Interface_ExecuteNet::~Interface_ExecuteNet(void)
/** Destruktor, ist leer */
{
};

bool Interface_ExecuteNet::ipSend(void)
/** Paket am Anfang der IP-Warteschlange zum Senden wird mit intoEthQueue() der
 *  Ethernet-Schlange &uuml;bergeben, aber nur wenn kein ARP-Paket unterwegs
 *  ist (sending_arp_packet) und die IP-Adresse aufgel&ouml;st werden kann
 *  (ArpCache::findMac()); <br>
 *  bei fehlgeschlagener Aufl&ouml;sung kommt das Paket ans Ende der
 *  IP-Warteschlange zum Senden; <br>
 *  liefert 'true', wenn ein IP-Paket zum Senden weitergereicht wurde, sonst
 *  'false' */
{
	string			mac_text;
	IpAddress		next_hop_ip;
	MacAddress		next_hop_mac;
	IPv4_Packet*		ip_packet;
	Ethernet_Packet*	eth_packet;

	ip_packet = NULL;
	eth_packet = NULL;

	if (!sending_arp_packet) // waiting for arp reply to own request ?
		ip_queue.down_out(ip_packet);
	if (ip_packet != NULL)
	{
		ip_packet->getNextHopAddress(next_hop_ip);
		cout << "next hop address: " << next_hop_ip << endl;
		// find ethernet address for ip address
		if (!arp_cache.findMac(next_hop_ip, next_hop_mac))
		{
			cout << "can't resolve IpAddress, so put back" << endl;
			ip_queue.down_in(ip_packet);
			arpSend(ip_packet);
		}
		else
		{	// set address, delete route pointer
			mac_text = next_hop_mac.getAddress();
			cout << "next_hop_mac: " << next_hop_mac << " "
				<< mac_text << endl;
			ip_packet->setDestinationMacAddress(next_hop_mac);
			ip_packet->setRoute(NULL);
			if (ip_packet->isProtocol(ETHERNET))
				eth_packet = reinterpret_cast<Ethernet_Packet*>
						(ip_packet);
		};
	};
	if (eth_packet != NULL) intoEthQueue(eth_packet);
	// return if ip_packet found in ip_queue
	return (ip_packet != NULL);
};

void Interface_ExecuteNet::arpSend(IPv4_Packet* ip_packet)
// send an arp request for target address of 'ip_packet'
/** sendet eine ARP-Anfrage (ARP_Packet::requestIp()) f&uuml;r die Zieladresse
 *  von 'ip_packet', aber nur wenn kein ARP-Paket unterwegs ist
 *  (sending_arp_packet), <br>
 *  entstandenes Paket kommt ans Ende der Ethernet-Sendeschlange
 *  (intoEthQueue()) */
{
	IpAddress	next_hop_ip, my_ip;
	MacAddress	my_mac;
	ARP_Packet	*arp_packet;
	Ethernet_Packet	*eth_packet;

	if (sending_arp_packet) return;	// waiting for an arp reply ?
	sending_arp_packet = true;	// reset by ethernetSend
					// xor by arpReceive
	arp_packet = NULL;
	eth_packet = NULL;
	arp_packet = new ARP_Packet;

	if (arp_packet != NULL)
	{
		ip_queue.getAddress(my_ip);
		cout << my_ip << " Generate and set ARP_Packet" << endl;
		ip_packet->getNextHopAddress(next_hop_ip);
		eth_queue.getAddress(my_mac);
		arp_packet->setSourceMacAddress(my_mac);
		// cout << "Next Hop: " << next_hop_ip << " " 
		//	<< next_hop_ip.getAddress() << endl;
		arp_packet->requestIp(my_ip, next_hop_ip);
		if (arp_packet->isProtocol(ETHERNET))
			eth_packet = reinterpret_cast<Ethernet_Packet*>
					(arp_packet);
	};
	if (eth_packet) intoEthQueue(eth_packet);
};

void Interface_ExecuteNet::intoEthQueue(Ethernet_Packet *eth_packet)
/** Paket 'eth_packet' kommt ans Ende der Ethernet-Schlange zum Senden,
 *  ruft EthernetQueue::down_in() */
{
	if (eth_packet != NULL)
	{
		cout << "Inserting packet " << eth_packet 
			<< " in ethernet queue." << endl;
		eth_queue.down_in(eth_packet);
	};
};

bool Interface_ExecuteNet::turnInterface(void)
/** dreht Netzwerkkarte nach links in die Richtung des Zielwinkels
 *  needed_angle; gibt zur&uuml;ck, ob sich die Karte gedreht hat; <br>
 *  Winkelschritt wird berechnet und addiert, nahe am Zielwinkel wird nur
 *  dessen Wert &uuml;bernommen, */
{
	bool	turned;
	double	x_value, full_distance, min_distance;
	double	distance_to_last, distance_to_needed;

	turned = (direction_angle != needed_angle);	// needs turn?
	if ((needed_angle == 0.0) && (direction_angle > 180.0))
		needed_angle = 360.0;	// better for calculation
	if (turned)
	{	// calculate angle step (rotation speed)
		distance_to_needed = fabs(direction_angle - needed_angle);
		distance_to_last = fabs(direction_angle - last_needed_angle);
		min_distance = (distance_to_needed < distance_to_last)?
				(distance_to_needed):(distance_to_last);
		full_distance = fabs(last_needed_angle - needed_angle);
		x_value = min_distance / full_distance;
		angle_step = 1 / (2 + exp(3.0 - x_value * 12.0)) * 
				animation_speed_factor;
		// cout << "Interface: angle_step: " << angle_step << " "
		//	<< "min_distance: " << min_distance << endl;
	};
	if ((direction_angle + angle_step > 360.0) && 
		(needed_angle < direction_angle))
			needed_angle += 360.0;	// better for next comparison
	if (fabs(direction_angle - needed_angle) > fabs(angle_step))
	{	// add calculated step
		direction_angle = fmodf(direction_angle + angle_step, 360.0);
		needed_angle = fmod(needed_angle, 360.0);
	}
	else
	{	// target angle reached, all same value, reset angle_step
		needed_angle = fmod(needed_angle, 360.0);
		direction_angle = needed_angle;
		last_needed_angle = needed_angle;
		angle_step = interface_executenet_init_angle_step;
	};
	// cout << "Interface_ExecuteNet: " << direction_angle << " "
	//	<< needed_angle << endl;
	return turned;
};

bool Interface_ExecuteNet::ethernetSend(void)
/** Paket am Anfang der Ethernet-Schlange zum Senden wird entnommen und dem
 *  Kabel &uuml;bergeben(Net_Cable::insertPacket()), vorher muss sich die Karte
 *  aber in Senderichtung gedreht haben (turnInterface()) und das Paket durch
 *  die Schlange gelaufen sein (EthernetQueue::down_out() liefert keinen
 *  NULL-Zeiger); <br>
 *  R&uuml;ckgabewert 'true', wenn ein Paket aus der Schlange entnommen wurde,
 *  sonst 'false' */
{
	MacAddress	my_mac;
	Packet		*cable_packet;
	Ethernet_Packet	*eth_packet;

	cable_packet = NULL;
	eth_packet = NULL;
	if ((eth_queue.down_size() > 0) && (attached_cable))
	{
		eth_queue.getAddress(my_mac);
		// don't change needed angle if turning is in progress
		// find send direction and turn to it
		if (direction_angle == needed_angle)
			needed_angle = fmodf(attached_cable->getAngle(my_mac)
						+ 270.0, 360.0);
		if (!turnInterface())	// turn done -> take packet if gone
					// through the queue
			eth_queue.down_out(eth_packet);
		else
		{
			// cout << "Interface angle: " << getAngle() << " ";
			// cout << "needed: " << needed_angle << endl;
		}
	};
	if (eth_packet != NULL)
	{
	//	if (eth_packet->isProtocol(ARP)) sending_arp_packet = false;
		cable_packet = reinterpret_cast<Packet*>(eth_packet);
		cout << "Getting packet " << eth_packet 
			<< " from ethernet queue." << endl;
	};
	// give packet to cable
	if (cable_packet != NULL)
	{
		// cout << "Inserting packet " << cable_packet
		//	<< " in cable: " << attached_cable << endl;
		if (attached_cable)
			attached_cable->insertPacket(my_mac, cable_packet);
	};
	// return if eth_packet found in eth_queue
	return (eth_packet != NULL);
};

bool Interface_ExecuteNet::ethernetReceive(void)
/** Karte empf&auml;ngt ein Ethernetpaket: im Kabel liegt ein Paket f&uuml;r
 *  diese Karte, dann in Empfangsrichtung drehen, Paket ans Ende der
 *  Ethernet-Warteschlange zum Empfangen setzen und im Kabel l&ouml;schen;<br>
 *  liefert 'true', wenn Paket empfangen wurde, sonst 'false' */
{
	bool		is_for_my_pos;
	MacAddress	my_mac, destination_mac;
	Packet		*cable_packet;
	Ethernet_Packet	*eth_packet;

	cable_packet = NULL;
	eth_packet = NULL;
	is_for_my_pos = false;

	// cout << "Interface_ExecuteNet: check packets in cable" << endl;
	eth_queue.getAddress(my_mac);
	// a packet for my address?
	if (attached_cable != NULL) 
		if (attached_cable->numPackets() > 0)
			is_for_my_pos = attached_cable->sniffPacket(my_mac,
								cable_packet);
	// turn to receive direction
	if (is_for_my_pos)
	{
		// don't change needed angle if turning is in progress
		if (direction_angle == needed_angle)
			needed_angle = fmodf(attached_cable->getAngle(my_mac)
						 + 90.0, 360);
		if (turnInterface())
		{
			// cout << "Interface angle: " << getAngle() << " ";
			// cout << "needed: " << needed_angle << endl;
			cable_packet = NULL;
		};
	};
	// cout << "Found: " << cable_packet << endl;
	if (cable_packet != NULL)
	{
		if (cable_packet->isProtocol(ETHERNET))
			eth_packet = reinterpret_cast<Ethernet_Packet*>
					(cable_packet);
	};
	// cout << "Testing eth_packet" << endl;
	if (eth_packet != NULL)
	{
		// cout << "Testing found packet " << eth_packet << endl;
		eth_packet->getDestinationMacAddress(destination_mac);
		cout << "Dest. Address: " << destination_mac << endl;
		// has right destination address ?
		if ((!(destination_mac == my_mac)) &&
			(destination_mac.getAddress() != "ff:ff:ff:ff:ff:ff"))
							// broadcast
		{
			cout << "Packet not for me." << endl;
			eth_packet = NULL;	// ignoring packet
		}
		else
		{	// into ethernet receive queue and delete in cable
			cout << "Received packet " << cable_packet << endl;
			eth_queue.up_in(eth_packet);
			attached_cable->deletePacket(cable_packet);
		}
	};
	// return if packet for me found in cable
	return (eth_packet != NULL);
}


bool Interface_ExecuteNet::ipReceive(void)
/** Paket am Anfang der Ethernet-Schlange zum Empfangen wird entnommen
 *  (EthernetQueue::up_out()) und als ARP-Paket verarbeitet mit arpReceive()
 *  oder als IP-Paket ans Ende der IP-Warteschlange zum Empfangen gesetzt
 *  (IpQueue::up_in()) nachdem TTL reduziert wurde
 *  (IPv4_Packet::decreaseTTL()), bei TTL == 0 wird Paket gel&ouml;scht, <br>
 *  'true' wenn Paket aus der Ethernet-Schlange entnommen wurde, <br>
 *  Weiterverarbeitung dann im Host oder L3_Router per Interface::up_out() */
{
	Ethernet_Packet	*eth_packet;
	ARP_Packet	*arp_packet;
	IPv4_Packet	*ip_packet;

	eth_packet = NULL;
	arp_packet = NULL;
	ip_packet = NULL;

	eth_queue.up_out(eth_packet);	// packet in ethernet queue
	if (eth_packet != NULL)
	{	// ARP or IP
		if (eth_packet->isProtocol(ARP))
			arp_packet = reinterpret_cast<ARP_Packet*>
					(eth_packet);	
		if (eth_packet->isProtocol(IPv4))
			ip_packet = reinterpret_cast<IPv4_Packet*>
					(eth_packet);	
	};
	if (arp_packet != NULL) arpReceive(arp_packet);
	if (ip_packet != NULL)
	{
		cout << "*** Received IP packet. *** " << ip_packet << endl;
		if (ip_packet->getTTL() > 0)
		{
			// cout << "Interface: Inserting in ip_queue" << endl;
			ip_queue.up_in(ip_packet);	// into queue
			ip_packet->decreaseTTL();	// ttl - 1
		}
		else
		{
			cout << "Interface: ip_packet has TTL = 0. "
				<< "Deleting it." << endl;
			delete ip_packet;
			ip_packet = NULL;
		};
	};
	// return if got a packet from eth_queue
	return (eth_packet != NULL);
};

void Interface_ExecuteNet::arpReply(ARP_Packet *arp_packet)
/** beantwortet empfangenes ARP-Paket 'arp_packet' mit Anfrage, Paket kommt ans
 *  Ende der Ethernet-Warteschlange zum Senden */
// answers to received 'arp_packet' with request
{
	string		mac_text, ip_text;
	IpAddress	source_ip, destination_ip, my_ip;
	MacAddress	source_mac, destination_mac, my_mac;
	Ethernet_Packet	*eth_packet;

	eth_packet = NULL;

	if (arp_packet)
	{	// take needed addresses
		arp_packet->getSourceMacAddress(source_mac);
		arp_packet->getSourceIpAddress(source_ip);
		ip_queue.getAddress(my_ip);
		eth_queue.getAddress(my_mac);
		cout << my_ip << " Replying to ARP packet" << endl;
		destination_mac = source_mac;
		destination_ip = source_ip;
		// build packet
		arp_packet->setDestinationMacAddress(destination_mac);
		mac_text = my_mac.getAddress();
		// cout << "My mac: " << my_mac << " " 
		// << mac_text.c_str() << endl;
		arp_packet->setSourceMacAddress(my_mac);
		arp_packet->replyIp(my_ip, destination_ip);
		if (arp_packet->isProtocol(ETHERNET))
			eth_packet = reinterpret_cast<Ethernet_Packet*>
					(arp_packet);
		// to end of queue
		if (eth_packet) intoEthQueue(eth_packet);	
	};
};

void Interface_ExecuteNet::arpReceive(ARP_Packet *arp_packet)
/** verarbeitet ein empfanges ARP-Packet 'arp_packet': Anfrage nach der eigenen
 *  Adresse wird mit arpReply() beantwortet, Antworten an die eigene
 *  Adresse werden in den ARP-Cache eingetragen, <br>
 *  unbeantwortete Pakete werden gel&ouml;scht */
{
	bool		no_delete;
	IpAddress	my_ip, destination_ip, source_ip;
	MacAddress	source_mac;

	// cout << "Testing arp_packet" << endl;
	no_delete = false;
	if (arp_packet != NULL)
	{
		ip_queue.getAddress(my_ip);
		cout << my_ip << " Received ARP Packet" << endl;
		arp_packet->getDestinationIpAddress(destination_ip);
		// cout << "Dest. IP: " << destination_ip << " "
		//	<< "my ip: " << my_ip << endl;
		// cout << "Request?: " << arp_packet->isRequest() << " ";
		// cout << "Reply?: " << arp_packet->isReply() << endl;
		if ((arp_packet->isRequest()) && (destination_ip == my_ip))
		{	// request to my address ?
			arpReply(arp_packet);
			no_delete = true;	// used as reply packet
		}
		if ((arp_packet->isRequest()) && (!(destination_ip == my_ip)))
		{
			cout << "Don't know IP for Request." << endl;
		};
		if ((arp_packet->isReply()) && (destination_ip == my_ip))
		{	// answer to me for an request of me
			sending_arp_packet = false;
			arp_packet->getSourceMacAddress(source_mac);
			arp_packet->getSourceIpAddress(source_ip);
			cout << my_ip << " Received ARP Reply " 
				<< source_mac << endl;
			arp_cache.insertIpMac(source_ip, source_mac);
		};
		if ((arp_packet->isReply()) && (!(destination_ip == my_ip)))
		{
			cout << "ARP Reply not for me." << endl;
		};
	};
	// delete packet
	if ((arp_packet != NULL) && (!no_delete))
	{
		cout << "Interface_ExecuteNet: Deleting ARP_Packet "
			<< arp_packet << endl;
		delete arp_packet;
		arp_packet = NULL;
	};
};

size_t Interface_ExecuteNet::numPackets(void) const
/** Anzahl der Pakete in den Warteschlangen (IP und Ethernet, Sende- und
 *  Empfangsrichtung) wird zur&uuml;ckgegeben */
{
	size_t	count;

	count = ip_queue.down_size() + ip_queue.up_size();
	count += eth_queue.down_size() + eth_queue.up_size();
	return count;
};

size_t Interface_ExecuteNet::numPacketsOnCable(void) const
/** gibt die Anzahl der Pakete auf dem Kabel zur&uuml;ck, wenn das erste
 *  Paket im Kabel an mich gerichtet ist, liefert sonst '0'; <br>
 *  gibt an, ob bald ein Paket empfangen werden kann */
// returns number of Packets on cable if first packet is for me 
// (by direction on cable), else returns 0
{
	bool		is_for_my_pos;
	size_t		count;
	MacAddress	my_mac;
	Packet		*cable_packet;

	count = 0;
	eth_queue.getAddress(my_mac);
	cable_packet = NULL;
	if (attached_cable) count = attached_cable->numPackets();
	is_for_my_pos = false;
	if (count > 0)	// is packet for me ?
		is_for_my_pos = attached_cable->sniffPacket(my_mac,
								cable_packet);
	if (!is_for_my_pos) count = 0;
	cable_packet = NULL;
	return count;
};

void Interface_ExecuteNet::executeNet(double speed_factor)
/** Ausf&uuml;hren eines Schrittes der Netzwerksimulation, <br>
 *  Geschwindigkeitsfaktor 'speed_factor' f&uuml;r Pakettransport wird an die
 *  Schlangen gegeben und in der eigenen Rotation beachtet, <br>
 *  nur eines dieser Teile wird ausgef&uuml;hrt (in der Priorit&auml;t dieser
 *  Reihenfolge): IP/ARP-Paket empfangen, Ethernet-Paket vom Kabel empfangen,
 *  Ethernet-Paket ins Kabel absenden, IP-Paket zum Senden */ 
{
	bool	sent, received, do_receive;

	animation_speed_factor = speed_factor *
					interface_executenet_angle_step_factor;
	ip_queue.setAnimationSpeed(speed_factor);
	eth_queue.setAnimationSpeed(speed_factor);
	sent = false;
	do_receive = false;
	// only receive if there is a packet on cable for me
	if (numPacketsOnCable() > 0) do_receive = true;
	if ((numPackets() > 0) || (do_receive))
		cout << "Interface " << ip_queue.getAddress() << endl;
	// cout << "Interface sending" << endl;
	if (!do_receive)
	{	// send ethernet xor send ip packet
		if (eth_queue.down_size() == 0) 
		{
			if (ipSend()) sent = true;
		}
		else if (ethernetSend()) sent = true;
	};
	// cout << "Interface_ExecuteNet: " << eth_queue.down_size() << " "
	//	<< eth_queue.up_size() << endl;

	received = false;
	// cout << "Interface receiving" << endl;
	// receive ip xor ethernet packet
	if (eth_queue.up_size() == 0)
	{
		if (ethernetReceive()) received = true;
	}
	else if (ipReceive()) received = true;
};

