/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <iostream>
#include <GL/gl.h>

#include "net_cable.h"
#include "constants.h"

/** Name der Variablen zur Ethernet-Adresse, die zum Beginn des Kabels
 *  geh&ouml;rt, im Info-Variablenbaum zum Dialog */
const char		net_cable_varname_beginaddr[] = "begin address";
/** Name der Variablen zur Ethernet-Adresse, die zum Ende des Kabels
 *  geh&ouml;rt, im Info-Variablenbaum zum Dialog */
const char		net_cable_varname_endaddr[] = "end address";
/** Name des Teilbaumes vom Kabel im Info-Variablenbaum zum Dialog */
const char 		net_cable_varname_cable[] = "cable";

/** Farbe des Netzwerkkabels, RGBA 0.0 - 1.0, Farbe der Seitenw&auml;nde etwas
 *  dunkler, siehe net_cable_corners_shade_difference[] in
 *  net_cable_corners.cpp */
const double		net_cable_color[] = {0.9, 0.9, 0.9, 0.9};
/** Anteil / Verh&auml;ltnis des Bodens an der gesamten Kabeltiefe (depth),
 *  bestimmt die Breite des Bodens und die Schr&auml;gheit der
 *  Seitenw&auml;nde */
const double		net_cable_ground_depth_factor = 0.6;
/** Standard-Kabeltiefe (depth) am Anfang */
const double		net_cable_init_depth = 1.0;
/** Standard-Kabelh&ouml;he (height) am Anfang */
const double		net_cable_init_height = 0.1;

Net_Cable::Net_Cable(void):
	depth(0.0), height(0.0), last_level_of_details_on(true),
	display_list(0)
/**
* Konstruktor, setzt Anfangswerte
*/
{
};

Net_Cable::~Net_Cable(void)
/**
* Destruktor, l&ouml;scht die Displayliste des Kabels
*/
{	// cout << "~Net_Cables start" << endl;
	//cout << traffic.begin() << " " << 
	// traffic.end() << " " << traffic.size() << " ";
	if (glIsList(display_list)) glDeleteLists(display_list, 1);
	display_list = 0;
	// cout << "~Net_Cables end" << endl;
};

void Net_Cable::getInfos(Variable_Tree* &infos)
/** stellt den Variablenbaum f&uuml;r das Dialogfenster zusammen, Zeiger
 *  darauf in 'infos' zur&uuml;ck, <br>
 *  begin_address, end_address */
{
	Numeric_Variable<MacAddress>	*var_beginaddress, *var_endaddress;

	current_infos.clear();

	var_beginaddress = new Numeric_Variable<MacAddress>;
	var_beginaddress->setName(net_cable_varname_beginaddr);
	var_beginaddress->setLength(1, 0);
	var_beginaddress->setMaxmin("FF:FF:FF:FF:FF:FF", "00:00:00:00:00:00");
	var_beginaddress->setValue(begin_address);
	current_infos.addVar(var_beginaddress);
	var_endaddress = new Numeric_Variable<MacAddress>;
	var_endaddress->setName(net_cable_varname_endaddr);
	var_endaddress->setLength(1, 0);
	var_endaddress->setMaxmin("FF:FF:FF:FF:FF:FF", "00:00:00:00:00:00");
	var_endaddress->setValue(end_address);
	current_infos.addVar(var_endaddress);
	current_infos.setName(net_cable_varname_cable);
	infos = &current_infos;
};

void Net_Cable::setInfos(Variable_Tree* infos)
/** liest die Werte aus dem Variablenbaum ein, der Zeiger 'infos' sollte
 *  auf den internen Baum zeigen, wird aber nicht beachtet, <br>
 *  begin_address, end_address */
{
	string		value_text;
	Variable*	current_variable;

	// the pointers in &infos should be the same as in current_infos
	// cout << "Net_Cable::setInfos" << endl;
	if (current_infos.findVar(net_cable_varname_beginaddr, 
					current_variable))
	{
		value_text = current_variable->getValue();
		begin_address.setAddress(value_text.c_str());
	};
	if (current_infos.findVar(net_cable_varname_endaddr, current_variable))
	{
		value_text = current_variable->getValue();
		end_address.setAddress(value_text.c_str());
	};
};

void Net_Cable::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt die Einstellungen und Pakete des Kabels aus der XML-Datei
 *  'infile', <br>
 *  Net_Cable_Corners::loadXML(), begin_address, end_address,
 *  Net_Cable_Packets::loadXML(), danach initGraphic() */
{
	int		state, depth, num_corner;

	num_corner = 0;
	do
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_pos) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
			// nodeType: start of node, not end
		{
			Net_Cable_Corners::loadXML(infile, num_corner);
			num_corner++;
		};
		if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_begin) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
			// nodeType: start of node, not end
		{
			begin_address.loadXML(infile);
		};
		if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_end) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
			// nodeType: start of node, not end
		{
			end_address.loadXML(infile);
		};
		Net_Cable_Packets::loadXML(infile);
	} while ((state >= 0) && (depth > 1));
	initGraphic();
};

void Net_Cable::saveXML(xmlTextWriterPtr outfile) const
/** speichert die Einstellungen und Pakete des Kabels in der XML-Datei
 *  'outfile', <br>
 *  Net_Cable_Packets::saveXML(), begin_address, end_address,
 *  Net_Cable_Corners::saveXML() */
{
	xmlTextWriterStartElement(outfile, xml_syntax_cable);
	Net_Cable_Packets::saveXML(outfile);
	xmlTextWriterStartElement(outfile, xml_syntax_begin);
	begin_address.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	xmlTextWriterStartElement(outfile, xml_syntax_end);
	end_address.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	Net_Cable_Corners::saveXML(outfile);
	xmlTextWriterEndElement(outfile);
};

void Net_Cable::insertPacket(MacAddress begin_or_end, Packet *new_packet)
/** f&uuml;gt Paket 'new_packet' in das Kabelende ein, welches durch
 *  'begin_or_end' angegeben ist; <br>
 *  wenn 'begin_or_end' gleich der Ethernet-Adresse 'begin_address' ist,
 *  welche anzeigt, dass der Beginn des Kabels mit der Netzwerkkarte dieser
 *  Ethernet-Adresse verbunden ist, wird es am Beginn eingef&uuml;gt,
 *  sonst am Ende, <br>
 *  ruft getAngle(), Net_Cable_Packets::insertPacket() und
 *  updateDisplaylist() */
{
	bool	is_on_begin;
	double	angle_on_begin_or_end;

	is_on_begin = (begin_or_end == begin_address);
	angle_on_begin_or_end = getAngle(begin_or_end);
	Net_Cable_Packets::insertPacket(is_on_begin, new_packet, depth,
					net_cable_ground_depth_factor,
					angle_on_begin_or_end);
	updateDisplaylist();
};

bool Net_Cable::sniffPacket(MacAddress begin_or_end, Packet* &first_packet)
/** gibt das erste Paket 'first_packet' des Kabels zur&uuml;ck, wenn es 
 *  Richtung Ethernet-Adresse 'begin_or_end' geschickt wurde, <br>
 *  ruft Net_Cable_Packets::sniffPacket() */
{
	bool	is_on_begin;

	is_on_begin = (begin_or_end == begin_address);
	return Net_Cable_Packets::sniffPacket(is_on_begin, first_packet);
};

double Net_Cable::getAngle(MacAddress begin_or_end) const
/** gibt den Winkel zur Horizontalen des Kabelendes, beschrieben durch die
 *  Ethernet-Adresse 'begin_or_end' zur&uuml;ck, 0.0 - 360.0 Grad, <br>
 *  falls begin_or_end == begin_address, dann den Winkel der Geraden
 *  "Erster Punkt" - "Zweiter Punkt" zur Horizontalen, sonst
 *  Winkel der Geraden "letzter Punkt" - "vorletzter Punkt" zur Horizontalen */
{
	vector<Point_3D>::const_iterator	first, last;

	getCurve(first, last);		// first, last
	if (begin_or_end == begin_address)
	{
		last = first + 1;	// one after first
	}
	else
	{
		first = last;		// last
		last--;			// one before last
	};
	return first->getAngleXZ(*last);
};

void Net_Cable::executeNet(double speed_factor)
/** f&uuml;hrt einen Schritt der Netzwerksimulation im Kabel aus,
 *  d.h. verschiebt die Pakete im Kabel,
 *  'speed_factor' beeinflusst die Geschwindigkeit des Pakettransports,<br>
 *  ruft Net_Cable_Packets::executeNet() und updateDisplaylist() */
{
	Net_Cable_Packets::executeNet(speed_factor);
	updateDisplaylist();
};

bool Net_Cable::testExchangeAddresses(Point_3D test_pos,
					MacAddress test_address)
/** pr&uuml;ft, ob zu der Position 'test_pos' (z.B. eines Netzknotens
 *  (Host / L3_Router) mit Netzwerkkarte), welche nah am Kabelbeginn/Kabelende
 *  liegt, auch die Ethernet-Adresse 'test_address' (der Karte des Knotens)
 *  als begin_address/end_address gesetzt ist; <br>
 *  wenn dagegen an Position des Kabelanfangs die Adresse des Kabelendes sein
 *  sollte, bzw. an Position des Kabelendes die Adresse des Kabelanfangs,
 *  werden die Adressen getauscht; <br>
 *  'true' zur&uuml;ck, wenn getauscht wurde, sonst 'false', wenn vorher schon
 *  alles richtig war; <br>
 *  nach der Funktion liegt immer test_address == begin_address am Kabelanfang
 *  bzw. test_address == end_address am Kabelende */
{
	bool		do_exchange;
	size_t		last_index;
	MacAddress	swap_address;

	// cout << "Net_Cable::testExchangeAddresses" << endl;
	do_exchange = false;
	last_index = numCorners() - 1;
	// test_pos is corners[0] -> test_address should be begin_address
	if ((getCorner(0).getDistance(test_pos, false) <= depth) &&
			(test_address == end_address))
		do_exchange = true;
	// test_pos is corners[last] -> test_address should be end_address
	if ((getCorner(last_index).getDistance(test_pos, false) <= depth) &&
			(test_address == begin_address))
		do_exchange = true;
	// cout << "test: " << test_address << " " << test_pos << endl;
	// cout << "begin: " << begin_address << " " << getCorner(0) << endl;
	// cout << "end: " << end_address << " " 
	//	<< getCorner(last_index) << endl;
	if (do_exchange)
	{
		swap_address = begin_address;
		begin_address = end_address;
		end_address = swap_address;
		// cout << "exchanged" << endl;
	}
	return do_exchange;	// exchange done
};

void Net_Cable::getAddresses(MacAddress &begin, MacAddress &end)
/** Gibt die IP-Adressen fr Kabel-Anfang (in 'begin') und -Ende (in 'end')
 *  zur&uuml;ck */
{
	begin = begin_address;
	end = end_address;
};

void Net_Cable::getName(string &current_name)
/** gibt die Bezeichnung des Kabels in 'current_name' zur&uuml;ck,<br>
 *  bestehend aus Textkonstante, der Adresse am Kabelbeginn und der Adresse
 *  am Kabelende, zur Kennzeichnung des Kamerafokus in der Animation */
{
	current_name = net_cable_varname_cable;
	current_name += " " + begin_address.getAddress();
	current_name += " " + end_address.getAddress();
};

void Net_Cable::initGraphic(void)
/** Initialisierung der Grafik nach Start der OpenGL-Umgebung:
 *  setColor(), Displayliste, Net_Cable_Corners::initGraphic(), 
 *  Net_Cable_Corners::updateDisplaylistCorners(), updateDisplaylist() */
{
	depth = net_cable_init_depth; height = net_cable_init_height;
	setColor(net_cable_color[0], net_cable_color[1], net_cable_color[2],
			net_cable_color[3]);
	if (!glIsList(display_list)) display_list = glGenLists(1);
	Net_Cable_Corners::initGraphic();
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

void Net_Cable::init(void)
/** Initialisierung des Kabels: begin_address, end_address, delAllPackets()
 *  (alle Pakete l&ouml;schen) und initGraphic() */
{
	begin_address.init();
	end_address.init();
	delAllPackets();
	initGraphic();
};

void Net_Cable::setColor(double new_red,  double new_green, 
		 double new_blue, double new_alpha)
/**
* Setzt die Farbe des Kabels, Wertebereich 0.0 - 1.0,
* 'new_red' (Rot), 'new_green' (Gr&uuml;n), 'new_blue' (Blau) und
* 'new_alpha' (Transparenz, 1.0 undurchsichtig),
* alpha funktioniert nur mit Blending, <br>
* ruft u.a. Net_Cable_Corners::setColor() 
*/
{
	Net_Cable_Corners::setColor(new_red, new_green, new_blue, new_alpha);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

void Net_Cable::setBegin(double x, double y, double z)
/**
* Position des Kabelanfangs als 3D-Punkt setzen (x, y, z),
* u.a. Net_Cable_Corners::setBegin() */
{
	Net_Cable_Corners::setBegin(x, y, z);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

void Net_Cable::setEnd(double x, double y, double z)
/**
* Position des Kabelendes als 3D-Punkt setzen (x, y, z),
* u.a. Net_Cable_Corners::setEnd() */
{
	Net_Cable_Corners::setEnd(x, y, z);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

void Net_Cable::setBeginNodeRadius(double new_begin)
/** Kabelverk&uuml;rzung am Beginn um 'new_begin' zum richtigen 'Andocken' an
 *  Netzknoten (Host / L3_Router), <br>
 *  u.a. Net_Cable_Corners::setBeginNodeRadius() */
{
	Net_Cable_Corners::setBeginNodeRadius(new_begin);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

void Net_Cable::setEndNodeRadius(double new_end)
/** Kabelverk&uuml;rzung am Ende um 'new_end' zum richtigen 'Andocken' an
 *  Netzknoten (Host / L3_Router), <br>
 *  u.a. Net_Cable_Corners::setEndNodeRadius() */
{
	Net_Cable_Corners::setEndNodeRadius(new_end);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

void Net_Cable::addCorner(double x, double y, double z)
/** f&uuml;gt neuen Hilfspunkt (x, y, z) ins Kabel ein, Bezierkurve des Kabels
 *  wird danach neu berechnet und gezeichnet, <br>
 *  u.a. Net_Cable_Corners::addCorner() */
{
	Net_Cable_Corners::addCorner(x, y, z);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

bool Net_Cable::isOnCable(double x, double y, double z)
/** gibt an, ob Punkt (x, y, z) auf den Hilfslinien zwischen den Hilfspunkten
 *  des Kabels liegt, 'true' wenn ja, sonst 'false'; <br>
 *  ruft nur Net_Cable_Corners::isOnCable() */
{
	return Net_Cable_Corners::isOnCable(x, y, z, depth);
};

void Net_Cable::moveCorner(double old_posxz[2], double new_posxz[2])
/** verschiebt den Punkt (x, y, z) = (old_posxz[0], posy, old_posxz[1]) zum
 *  Punkt (new_posxz[0], posy, new_posxz[1]) im Feld 'corners' (Hilfspunkte),
 *  wobei posy die Standardh&ouml;he aller Eckpunkte des Kabels ist, <br>
 *  ruft u.a. Net_Cable_Corners::moveCorner() */
{
	Net_Cable_Corners::moveCorner(old_posxz, new_posxz);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

size_t Net_Cable::removeCorner(double x, double y, double z)
/** l&ouml;scht den Eckpunkt (x, y, z) aus dem Feld 'corners' (Hilfspunkte),
 *  gibt die neue Anzahl von Punkten in 'corners' zur&uuml;ck, <br>
 *  ruft u.a. Net_Cable_Corners::removeCorner() */
{
	size_t	num_corners;

	num_corners = Net_Cable_Corners::removeCorner(x, y, z);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
	return num_corners;
};

void Net_Cable::setSize(double new_depth, double new_height)
/**
 * Hhe 'new_height' und Tiefe 'new_depth' des Kabels setzen; <br>
 * ruft u.a. Net_Cable_Packets::setSize()
 */
{
	depth = new_depth; height = new_height;
	Net_Cable_Packets::setSize(depth, net_cable_ground_depth_factor);
	Net_Cable_Corners::updateDisplaylistCorners(depth, height,
						net_cable_ground_depth_factor);
	updateDisplaylist();
};

void Net_Cable::draw(bool with_control_points) const
/**
 * Zeichnet das Kabel in ein OpenGL-Fenster, wenn 'with_control_points' auf
 * 'true' ist, werden die Hilfspunkte und die Hilfslinien mit gezeichnet,
 * welche die Anpassung der Bezierkurve vereinfachen, <br>
 * Net_Cable_Corners::draw(), Net_Cable_Corners::drawControlPoints()
 */
{
	// cout << "Net_Cable::draw" << endl;	
	if (glIsList(display_list)) glCallList(display_list);
	Net_Cable_Corners::draw();
	if (with_control_points)
		Net_Cable_Corners::drawControlPoints(depth, height);
};

void Net_Cable::drawSetCameraPos(bool level_of_details_on, Point_3D camera_pos)
/** aktualisiert die internen Werte zu 'last_camera_pos' (letzte 
 *  Kameraposition) und 'last_level_of_details_on' (Level of Detail
 *  an oder aus), Displayliste der Pakete (abh&auml;ngig von Level of Detail)
 *  wird neu gezeichnet (updateDisplaylist()) */
{
	last_camera_pos = camera_pos;
	last_level_of_details_on = level_of_details_on;
	updateDisplaylist();
};

void Net_Cable::updateDisplaylist(void)
/** zeichnet die Displayliste der Pakete neu,
 *  ruft die Zeichenmethode von Net_Cable_Packets (Net_Cable_Packets::draw())
 *  und &uuml;bergibt denen 'last_level_of_details_on' und 'last_camera_pos' */
{
	if (!glIsList(display_list)) return;
	// cout << "Net_Cable::updateDisplaylist start" << endl;
	glNewList(display_list, GL_COMPILE);
	Net_Cable_Packets::draw(last_level_of_details_on, last_camera_pos);
	glEndList();
	// cout << "Net_Cable::updateDisplaylist end" << endl;
};

