/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <iostream>
#include <math.h>

#include "bezier.h"
#include "cylinder.h"
#include "net_cable_corners.h"

/** Anzahl der Kurvenpunkte, die eine Bezierkurve beschreiben, auf der das
 *  Kabel entlangf&uuml;hrt, mit diesen Punkten wird die Kurve gezeichnet
 *  (als gerade Segmente zwischen den Punkten), f&uuml;r Feld 'curve_points' */
const unsigned short	net_cable_corners_num_curve_points = 20;
/** Farbunterschied der Seitenteile des Kabels zum Kabelboden, zur
 *  Verdeutlichung der Kabeldarstellung */
const double		net_cable_corners_shade_difference[] = 
						{-0.1, -0.1, -0.1, 0.0};
/** H&ouml;hen-Abstand der Kreise (zur Darstellung der Hilfspunkte 'corners')
 *  zu den Kabel-Segmenten, sonst treten &Uuml;berlagerungen auf */
const double		net_cable_corners_height_difference = 0.001;
/** Dicke der Linien zur Verbindung der Kreise (Hilfspunkte), mit Hilfe dieser
 *  Linien kann man neue Hilfspunkte hinzuf&uuml;gen */
const double		net_cable_corners_line_width = 3.0;

Net_Cable_Corners::Net_Cable_Corners(void):
	begin_node_radius(0.0), end_node_radius(0.0),
	display_list_corners(0)
/**
* Konstruktor, setzt Anfangswerte
*/
{
};

Net_Cable_Corners::~Net_Cable_Corners(void)
/**
* Destruktor, l&ouml;scht alle Ecken des Kables und die Displayliste
*/
{
	// cout << "Net_Cable_Corners destruktor." << endl;
	corners.erase(corners.begin(), corners.end());
	if (glIsList(display_list_corners))
		glDeleteLists(display_list_corners, 1);
	display_list_corners = 0;
};

void Net_Cable_Corners::initGraphic(void)
/** belegt Speicherplatz f&uuml;r die Displayliste */
{
	if (!glIsList(display_list_corners))
		display_list_corners = glGenLists(1);
};

void Net_Cable_Corners::loadXML(xmlTextReaderPtr infile, int num_corner_read)
/** l&auml;dt die Kabel-Hilfspunkte aus der XML-Datei 'infile' */
{
	Point_3D	new_corner;

	
	if ((num_corner_read <= 1) && (corners.size() >= 2))
	{
		corners[num_corner_read].loadXML(infile);
	}
	else
	{
		new_corner.loadXML(infile);
		corners.insert(corners.end(), new_corner);
	};
};

void Net_Cable_Corners::saveXML(xmlTextWriterPtr outfile) const
/** speichert die Kabel-Hilfspunkte in der XML-Datei 'outfile' */
{
	size_t	i;

	for(i = 0; i < corners.size(); i++) corners[i].saveXML(outfile);
};

void Net_Cable_Corners::setColor(double new_red,  double new_green, 
		 		double new_blue, double new_alpha)
/**
* Setzt die Farbe des Kabels, Wertebereich 0.0 - 1.0,
* 'new_red' (Rot), 'new_green' (Gr&uuml;n), 'new_blue' (Blau) und 'new_alpha'
* (Transparenz, 1.0 undurchsichtig), alpha funktioniert nur bei eingeschaltetem
* Blending
*/
{
	unsigned int	i;

	// cout << "Net_Cable_Corners::setColor: " << new_red << " "
	//	<< new_green << " " << new_blue << " " << new_alpha << " "
	//	<< "size: " << corners.size() << endl;
	if (corners.size() == 0) corners.resize(1, Point_3D());
	for(i = 0; i < corners.size(); i++)
		corners[i].setColor(new_red, new_green, new_blue, new_alpha);
	shade_color.setColor(new_red + net_cable_corners_shade_difference[0],
			new_green + net_cable_corners_shade_difference[1],
			new_blue + net_cable_corners_shade_difference[2],
			new_alpha + net_cable_corners_shade_difference[3]);
};

void Net_Cable_Corners::getBeginEnd(Point_3D &begin, Point_3D &end) const
/** gibt die Positionen vom Kabelanfang ('begin') und Kabelende ('end')
 *  zur&uuml;ck, also den ersten und den letzten Punkt im Feld der
 *  Hilfspunkte */
{
	begin = corners[0];
	end = corners[corners.size() - 1];
};

void Net_Cable_Corners::setBegin(double x, double y, double z)
/**
* Position des Kabelanfangs als 3D-Punkt setzen (x, y, z)
*/
{
	if (corners.size() < 1) corners.resize(1, Point_3D());
	corners[0].setPos(x, y, z);
	setColor(corners[0].getR(), corners[0].getG(), corners[0].getB(),
			corners[0].getA());
};

void Net_Cable_Corners::setEnd(double x, double y, double z)
/**
* Position des Kabelendes als 3D-Punkt setzen (x, y, z)
*/
{
	if (corners.size() < 2) corners.resize(2, Point_3D());
	corners[ corners.size() - 1].setPos(x, y, z);
	setColor(corners[0].getR(), corners[0].getG(), corners[0].getB(),
			corners[0].getA());
};

void Net_Cable_Corners::setBeginNodeRadius(double new_begin)
/** Kabelverk&uuml;rzung 'new_begin' f&uuml;r den Beginn setzen, zum richtigen
 *  'Andocken' an die Netzknoten */
{
	begin_node_radius = new_begin;
};

void Net_Cable_Corners::setEndNodeRadius(double new_end)
/** Kabelverk&uuml;rzung 'new_end' f&uuml;r das Ende setzen, zum richtigen
 *  'Andocken' an die Netzknoten */
{
	end_node_radius = new_end;
};

bool Net_Cable_Corners::isInArea(const Point_3D &first_corner,
					const Point_3D &test_point,
					const Point_3D &second_corner) const
/* returns if test_point is in rectangular area, defined by first and second
 * point */
/** gibt 'true' zur&uuml;ck (sonst 'false'), falls der Punkt 'test_point' im
 *  rechteckigen Bereich  liegt, der durch 'first_corner' und 'second_corner'
 *  angegeben ist (linke obere und rechte untere Ecke bzw.
 *  rechte obere und linke untere Ecke) */
{
	bool	result;
	double	min_x, max_x, min_z, max_z;

	if (first_corner.getX() < second_corner.getX())
	{
		min_x = first_corner.getX(); max_x = second_corner.getX();
	}
	else
	{
		min_x = second_corner.getX(); max_x = first_corner.getX();
	};
	if (first_corner.getZ() < second_corner.getZ())
	{
		min_z = first_corner.getZ(); max_z = second_corner.getZ();
	}
	else
	{
		min_z = second_corner.getZ(); max_z = first_corner.getZ();
	};
	result = (min_x <= test_point.getX());
	result = result && (max_x >= test_point.getX());
	result = result && (min_z <= test_point.getZ());
	result = result && (max_z >= test_point.getZ());
	return result;
};

bool Net_Cable_Corners::findNearestByAngle(const Point_3D &before,
					const Point_3D &current,
					const Point_3D &after,
					const Point_3D &test_point) const
/* returns if current is before test_point (true) or not (false),
 * calculated by angle, not by distance */
/** angegeben ist eine Linie mit Ecke, definiert durch die Punkte 'before',
 *  (Anfang), 'current' (Ecke) und 'after' (Ende); zur&uuml;ckgegeben wird
 *  'true' wenn der Punkt 'test_point' auf/in der N&auml;he des Linienteils
 *  'current' - 'after' liegt (und nicht bei 'before' - 'current'),
 *  'false' wenn nicht. */
{
	bool	result;
	double	angle_before, angle_after, angle_testpoint;

	// triangle 'before', 'current' and horizontal line (look at Point_3D)
	// atan((current.z - before.z) / (current.x - before.x))
	angle_before = before.getAngleXZ(current);
	// triangle 'after', 'current' and horizontal line (look at Point_3D)
	// atan((current.z - after.z) / (current.x - after.x))
	angle_after = after.getAngleXZ(current);
	// triangle 'test_point', 'current' and horizontal line
	// (look at Point_3D)
	// atan((current.z - test_point.z) / (current.x - test_point.x))
	angle_testpoint = test_point.getAngleXZ(current);
	result = false;
	if (fabs(angle_testpoint - angle_before) >
		fabs(angle_testpoint - angle_after))
			result = true;
	return result;	
};

void Net_Cable_Corners::findNearestByDistance(const Point_3D &test_point, 
				vector<Point_3D>::const_iterator &nearest,
				const vector<Point_3D> &point_array) const
/* Nearest Corner/Curve point before (!) the test_point (from begin to end),
 * calculated by distance, not by angle */ 
/** findet den n&auml;chsten Kurven-/Eckpunkt 'nearest' vor dem Punkt
 *  'test_point', in der Reihenfolge wie im Feld 'point_array' definiert;<br>
 *  Anhand des Abstandes wird 'nearest' bestimmt, und auf Linie 'before',
 *  'nearest', 'after' wird anhand des Abstandes bestimmt, ob 'test_point'
 *  zwischen 'before' und 'nearest' (before ist dann nearest in R&uuml;ckgabe)
 *  oder 'nearest' und 'after' liegt, <br>
 *  ungenauer/fehleranf&auml;lliger als findNearestByAngle(), aber wohl
 *  schneller */
{
	double					min_distance, this_distance,
						before_distance, before_near, 
						after_distance, after_near;
	vector<Point_3D>::const_iterator	all_corners, before, after;

	all_corners = point_array.begin();
	min_distance = all_corners->getDistance(test_point, false);
	nearest = all_corners;
	// find nearest to 'test_point'
	while (all_corners != point_array.end())
	{
		this_distance = all_corners->getDistance(test_point, false);
		if (min_distance > this_distance)
		{
			min_distance = this_distance;
			nearest = all_corners;
		};
		all_corners++;
	};
	// line before - nearest - after
	before = nearest - 1;
	if (nearest == point_array.begin()) before = nearest;
	after = nearest + 1;
	if (after != point_array.end())
	{
		before_distance = before->getDistance(test_point, false);
		before_near = before->getDistance(*nearest, false);
		after_distance = after->getDistance(test_point, false);
		after_near = after->getDistance(*nearest, false);
		// change nearest if test_point between 'before' and 'nearest'
		if ((before_distance < before_near) &&
			(after_distance > after_near))
				nearest = before;
		if ((before_distance <= before_near) &&
			(after_distance <= after_near) &&
			(!findNearestByAngle(*before, *nearest,
							*after, test_point)))
				nearest = before;
	}
	else nearest--;
};

void Net_Cable_Corners::findNearest(const Point_3D &test_point, 
				vector<Point_3D>::const_iterator &nearest,
				const vector<Point_3D> &point_array) const
/* Nearest Corner/Curve point before (!) the test_point (from begin to end).
 */ 
/** findet den n&auml;chsten Kurven-/Eckpunkt 'nearest' vor dem Punkt
 *  'test_point', in der Reihenfolge wie im Feld 'point_array' definiert;<br>
 *  ben&ouml;tigt u.a. f&uuml;r die Neueinf&uuml;gung von Hilfspunkten, <br>
 *  nutzt findNearestByAngle() und isInArea(), wenn das fehlschl&auml;gt, wird
 *  findNearestByDistance() genutzt */
{
	bool					in_before_current,
						in_current_after;
	vector<Point_3D>::const_iterator	all_corners, before, after,
						last;

	all_corners = point_array.begin();
	last = point_array.end();
	last--;
	in_current_after = false;
	nearest = point_array.end();
	// between first and last_point of 'point_array' are
	// point_array.size() - 1 rectangular areas, test if test_point are
	// in one of it, if yes call findNearestByAngle()
	while (all_corners != last)
	{
		after = all_corners + 1;
		in_before_current = in_current_after;	// from last loop step
		in_current_after = isInArea(*all_corners, test_point, *after);
		if (test_point.getDistance(*all_corners, false) == 0.0)
			nearest = all_corners - 1;
		if (in_current_after && (!in_before_current))
			if (nearest == point_array.end())   // not set before
				nearest = all_corners;
		if (in_current_after && in_before_current)
		{
			before = all_corners - 1;
			if (findNearestByAngle(*before, *all_corners, *after,
						test_point))
				nearest = all_corners;
		};		
		all_corners++;
	};
	// try findNearestByDistance() if the algorithm before had no result
	if (nearest == point_array.end())	// not found
		findNearestByDistance(test_point, nearest, point_array);
	// cout << "nearest: " << nearest->getX() << " " 
	//	<< nearest->getZ() << endl;
};

void Net_Cable_Corners::findNearestCorner(const Point_3D &test_point,
			vector<Point_3D>::const_iterator &nearest) const
/* wrapper for findNearest with array 'corners' */
/** Hilfsfunktion zur Nutzung von findNearest() mit dem Feld 'corners'
 *  (Hilfspunkte) */
{
	findNearest(test_point, nearest, corners);
};

void Net_Cable_Corners::findNearestCurvePoint(const Point_3D &test_point,
			vector<Point_3D>::const_iterator &nearest) const
/* wrapper for findNearest with array 'curve_points' */
/** Hilfsfunktion zur Nutzung von findNearest() mit dem Feld 'curve_points'
 *  (Kurvenpunkte nach Bezierberechnung) */
{
	findNearest(test_point, nearest, curve_points);
};

void Net_Cable_Corners::curveLength(vector<Point_3D>::const_iterator &before,
				vector<Point_3D>::const_iterator &after,
				double &before_length, double &after_length,
				double &full_length) const
// length of curve begin - before, after - end, begin - end (full_length)
/** berechnet die L&auml;nge der Kurve (Feld 'curve_points') zwischen
 *  Anfang und Punkt 'before' (als Summe der Abst&auml;nde der 
 *  Punkte) - in 'before_length', zwischen 'after' und Ende - in
 *  'after_length' - und zwischen Anfang und Ende - in 'full_length', <br>
 *  f&uuml;r die Geschwindigkeit des Pakettransports im Kabel in
 *  Net_Cable_Packets::calcSpeed() */
{
	double					curve_part_length;
	vector<Point_3D>::const_iterator	all_curve_points, last_point;

	all_curve_points = curve_points.begin();
	before_length = 0.0;
	after_length = 0.0;
	full_length = 0.0;
	last_point = curve_points.end();
	last_point--;
	while(all_curve_points != last_point)
	{
		// distance between the two curve points
		curve_part_length = all_curve_points->getDistance(
						*(all_curve_points + 1));
		// sum up all / only points before / only points after
		full_length += curve_part_length;
		if (all_curve_points <= before)
			before_length += curve_part_length;
		if (all_curve_points >= after)
			after_length += curve_part_length;
		all_curve_points++;
	};
	/* cout << "Net_Cable_Corners::curveLength: " << endl
		<< "before length: " << before_length << " after length: "
		<< after_length << " sum: " << (before_length + after_length)
		<< " full: " << full_length << endl;*/
};

void Net_Cable_Corners::getCurve(vector<Point_3D>::const_iterator &first,
				vector<Point_3D>::const_iterator &last) const
/* iterators (begin) and (end - 1) to private array 'curve_points' */
/** gibt &uuml;ber die Iteratoren zum ersten 'first' und letzten Punkt 'last'
 *  Zugriff auf das private Feld 'curve_points'
 *  (Kurvenpunkte nach Bezierberechnung), nur zum Lesen; <br>
 *  siehe Net_Cable_Packets::movePacket() und Net_Cable::getAngle() */
{
	first = curve_points.begin();
	last = curve_points.end();
	last--;
};

Point_3D Net_Cable_Corners::getCorner(size_t number) const
/** gibt den Eckpunkt 'number' (0 bis size - 1) aus dem Feld 'corners'
 *  (Hilfspunkte) zur&uuml;ck */
{
	Point_3D	number_corner;

	if (corners.size() > number) number_corner = corners[number];
	return number_corner;
};

size_t Net_Cable_Corners::numCorners(void) const
/** gibt die Anzahl der Eckpunkte des Feldes 'corners' (Hilfspunkte)
 *  zur&uuml;ck */
{
	return corners.size();
};

void Net_Cable_Corners::addCorner(double x, double y, double z)
/** f&uuml;gt einen neuen Eckpunkt (x, y, z) in das Feld der Hilfspunkte ein,
 */
{
	Point_3D				new_corner;
	vector<Point_3D>::iterator		insert_pos;
	vector<Point_3D>::const_iterator	all_corners, nearest_corner; 

	new_corner.setPos(x, y, z);
	// if there is a corner on this point, return
	all_corners = corners.begin();
	while (all_corners != corners.end())
	{
		if (all_corners->getDistance(new_corner, false) == 0.0) 
			return;
		all_corners++;
	};
	// find nearest
	findNearestCorner(new_corner, nearest_corner);
	nearest_corner++;		
	// insert after this corner
	insert_pos = corners.begin();
	while (nearest_corner != corners.begin())
	{
		insert_pos++;
		nearest_corner--;
	};
	corners.insert(insert_pos, new_corner);
	// output all corners, for debugging
	all_corners = corners.begin();
	cout << "Net_Cable_Corners: Add corner: " << endl;
	while (all_corners != corners.end())
	{
		cout << all_corners->getX() << " " 
			<< all_corners->getZ() << endl;
		all_corners++;
	};
	// set the color of all corners to that of first point in array
	setColor(corners[0].getR(), corners[0].getG(), corners[0].getB(),
			corners[0].getA());
};

bool Net_Cable_Corners::isOnCable(double x, double y, double z,
					double depth) const
/* 'true' if point (x, y, z) is on help lines between the corners,
 * 'depth' is half of the cable thickness, need to find cable to which do
 *  addCorner() */
/** gibt 'true' zur&uuml;ck, wenn der Punkt (x, y, z) auf den Hilfslinien
 *  zwischen den Hilfspunkten 'corners' liegt, bzw. maximal im Abstand 'depth'
 *  zu diesen; <br> 
 *  zur Auswahl des Kabels, in das dann der Punkt eingef&uuml;gt / verschoben /
 *  gel&ouml;scht (Network::getInfos(), Network::move(), Network::remove())
 *  wird */
{
	bool					result;
	Point_3D				test_point;
	double					before_x, before_z;
	double					after_x, after_z;
	double					m, n, saved_z, saved_x;
	vector<Point_3D>::const_iterator	before_corner, after_corner;

	test_point.setPos(x, y, z);
	// find belonging part of curve
	findNearestCorner(test_point, before_corner);
	after_corner = before_corner + 1;
	before_x = before_corner->getX();
	before_z = before_corner->getZ();
	after_x = after_corner->getX();
	after_z = after_corner->getZ();
	// calculate the function, defining the line between before and after
	// is the test_point on or near it ?
	// z = m * x + n
	result = false;
	if (after_x != before_x)
	{
		m = (after_z - before_z) / (after_x - before_x);
		n = (before_z - m * before_x);
		if (after_x < before_x)	// then exchange values
		{
			saved_x = after_x;
			after_x = before_x;
			before_x = saved_x;
		};
		// cout << "Net_Cable_Corners: x z " << x << " " << z << " "
		//	<< "| x z " << ((z - n) / m) << " " << (m * x + n)
		//	<< " | m n " << m << " " << n << endl
		//	<< "before: " << *before_corner << " "
		//	<< "after: " << *after_corner << endl;
		if (((fabs(z - (m * x + n)) <= depth) &&
			(before_x <= x) && (after_x >= x)) ||
			((fabs(x - (z - n) / m) <= depth) &&
			(((before_z < after_z)?(before_z):(after_z)) <= z) && 
			(((before_z > after_z)?(before_z):(after_z)) >= z)))
				result = true;
	}
	else
	{
		if (after_z < before_z)	// then exchange values
		{
			saved_z = after_z;
			after_z = before_z;
			before_z = saved_z;
		};
		if ((after_x == x) &&	// before_x == x also true
			(before_z <= z) && (after_z >= z))
				result = true;
	};
	return result;
};

void Net_Cable_Corners::moveCorner(double old_posxz[2], double new_posxz[2])
/* move the point (x,y,z) = (old[0], posy, old[1]) to (new[0], posy, new[1]) */
/** verschiebt den Punkt (x, y, z) = (old_posxz[0], posy, old_posxz[1]) zum
 *  Punkt (new_posxz[0], posy, new_posxz[1]) im Feld 'corners' (Hilfspunkte),
 *  wobei posy die Standardh&ouml;he aller Eckpunkte des Kabels ist */
{
	double				posy;
	Point_3D			test_point;
	vector<Point_3D>::iterator	all_corners;

	test_point.setPos(old_posxz[0], 0.0, old_posxz[1]);
	all_corners = corners.begin();
	while (all_corners != corners.end())
	{
		// find corner with the old values, set it to new
		if (all_corners->getDistance(test_point, false) == 0.0)
		{
			posy = all_corners->getY();
			all_corners->setPos(new_posxz[0], posy, new_posxz[1]);
		};
		all_corners++;
	};
};

unsigned int Net_Cable_Corners::removeCorner(double x, double y, double z)
/* delete the corner (x,y,z) from array 'corners', returns new numCorners() */
/** l&ouml;scht den Eckpunkt (x, y, z) aus dem Feld 'corners' (Hilfspunkte),
 *  gibt die neue Anzahl von Punkten in 'corners' zur&uuml;ck */
{
	Point_3D			test_point;
	vector<Point_3D>::iterator	all_corners;

	test_point.setPos(x, y, z);
	all_corners = corners.begin();
	while (all_corners != corners.end())
	{
		if (all_corners->getDistance(test_point, false) == 0.0)
		{
			corners.erase(all_corners);
			all_corners = corners.end();
		}
		else all_corners++;
	};
	// cout << "end of Net_Cable::removeCorner" << endl;
	return corners.size();
};

void Net_Cable_Corners::draw(void) const
/** zeichnet das Kabel, d.h. ruft die Displayliste auf */
{
	if (glIsList(display_list_corners)) glCallList(display_list_corners);
};

void Net_Cable_Corners::drawControlPoints(double depth, double height) const
/** zeichnet die Hilfspunkte 'corners' und die Hilfslinien zwischen diesen,
 *  die Punkte als Zylinder mit dem Durchmesser 'depth' und der H&ouml;he
 *  'height' */
{
	size_t	i;

	if (corners.size() > 1)
		for(i = 0; i < corners.size() - 1; i++) 
		{
			glLineWidth(net_cable_corners_line_width);
			corners[i].colorForDraw();
			// the line between the corners
			glBegin(GL_LINES);
				corners[i].vertexDraw(0,
					-2*net_cable_corners_height_difference,
					0);
				corners[i + 1].vertexDraw(0,
					-2*net_cable_corners_height_difference,
					0);
			glEnd();
			glLineWidth(1.0);
			// the corners himself
			drawCorner(corners[i], depth, height);
		};
	if (corners.size() > 0)
		drawCorner(corners[ corners.size() - 1], depth, height);
};

void Net_Cable_Corners::updateDisplaylistCorners(double depth, double height,
						double ground_depth_factor)
/** aktualisiert die Displayliste, zeichnet das Kabel darin neu mit den
 *  Parametern 'depth' als Kabeldicke in der Tiefe, 'height' als Kabelh&ouml;he
 *  und 'ground_depth_factor' f&uuml;r die Kabelbodendicke (bestimmt die
 *  Schr&auml;gheit der Seitenw&auml;nde), <br>
 *  die Bezierkurvenpunkte 'curve_points' werden aus den Hilfspunkten 'corners'
 *  neu berechnet */
{
	unsigned int		i;
	double			begin_angle, end_angle;
	double			begin_translate, end_translate;

	if (!glIsList(display_list_corners)) return;

	Bezier::calcBezier(corners, net_cable_corners_num_curve_points,
				curve_points);
	glNewList(display_list_corners, GL_COMPILE);
	if (curve_points.size() > 1)
	{
		for(i = 0; i < curve_points.size() - 1; i++) 
		{
			begin_translate = 0.0;
			end_translate = 0.0;
			// begin center point of cable segment with angle
			// defining top left and bottom left point
			if (i > 0)
				begin_angle = curve_points[i].getAngleXZ(
							curve_points[i - 1]);
			else
			{
				begin_angle = curve_points[i + 1].getAngleXZ(
							curve_points[i]);
				begin_translate = -1.0 * begin_node_radius;
			};
			// end center point of cable segment with angle
			// defining top right and bottom right point
			if (i < curve_points.size() - 2)
				end_angle = curve_points[i + 2].getAngleXZ(
							curve_points[i + 1]);
			else
			{
				end_angle = curve_points[i + 1].getAngleXZ(
							curve_points[i]);
				end_translate = end_node_radius;
			};
			// draw it
			if ((curve_points[i + 1].getDistance(corners[0]) > 
				begin_node_radius) &&
				(curve_points[i].getDistance(corners[
				corners.size() - 1]) > end_node_radius))
			{
				drawSegment(curve_points[i],
					curve_points[i + 1],
					begin_angle, end_angle,
					begin_translate, end_translate,
					depth, height, ground_depth_factor);
			};
		};
	};
	glEndList();
};

void Net_Cable_Corners::drawCorner(Point_3D pos, double depth,
					double height) const
/** zeichnet die Hilfspunkte 'corners' des Kabels als Zylinder mit dem
 *  Durchmesser 'depth', der H&ouml;he 'height' am Mittelpunkt 'pos' */
{
	Cylinder	circle;

	circle.setPos(pos.getX(),
		pos.getY() - height / 2 - net_cable_corners_height_difference,
		pos.getZ());
	circle.setSize(depth, height, depth);
	circle.setColor(pos.getR(), pos.getG(), pos.getB(), pos.getA());
	circle.draw();
};

void Net_Cable_Corners::drawSegment(Point_3D begin, Point_3D end,
					double begin_angle, double end_angle,
					double begin_translate,
					double end_translate,
					double depth, double height,
					double ground_depth_factor) const
/** zeichnet einen Teil der Bezierkurve des Kabels als u-f&ouml;rmige Schiene,
 *  mit Mittelpunkt am Beginn 'begin', verschoben um 'begin_translate',
 *  dem Winkel 'begin_angle' am 'begin' f&uuml;r die Berechnung der
 *  linken oberen und linken unteren Ecke, dem Mittelpunkt am Ende 'end',
 *  verschoben um 'end_translate', dem Winkel 'end_angle' am 'end' f&uuml;r die
 *  Berechnung der rechten oberen und rechten unteren Ecke, <br>
 *  der Kabeldicke 'depth' (mit den Seitenw&auml;nden), der Kabelh&ouml;he
 *  'height' und dem Faktor 'ground_depth_factor' zur Dicke des Kabelbodens
 *  (und Bestimmung der Schr&auml;gheit der Seitenw&auml;nde), <br>
 *  alles am Beispiel eines horizontalen Segmentes (von Anfang links zum Ende
 *  rechts) */
{
	bool		change_end_angle;
	double		width_on_x_begin, width_on_z_begin;
	double		width_on_x_end, width_on_z_end, m1, m2;
	Point_3D	left, right;
	double		ground_factor_1, ground_factor_2, my_angle;

	my_angle = fmod(360.0 + end.getAngleXZ(begin), 360.0);
	begin_angle = fmodf(360.0 + begin_angle, 360.0);
	end_angle = fmodf(360.0 + end_angle, 360.0);
	begin_angle = 90.0 + my_angle + (begin_angle - my_angle) / 2.0;
	end_angle = 90.0 + my_angle + (end_angle - my_angle) / 2.0;

	begin_angle = begin_angle * M_PI / 180.0;
	end_angle = end_angle * M_PI / 180.0;
	my_angle = my_angle * M_PI / 180.0;
	change_end_angle = false;
	if (end.getX() == begin.getX())
		change_end_angle = (cos(begin_angle) * cos(end_angle) < 0.0);
	else if (end.getZ() == begin.getZ())
		change_end_angle = (sin(begin_angle) * sin(end_angle) < 0.0);
	else 
	{
		m1 = (end.getX() + cos(end_angle) - 
			begin.getX() - cos(begin_angle)) /
			(end.getZ() - sin(end_angle) - 
			begin.getZ() + sin(begin_angle));
		m2 = (end.getX() - begin.getX()) / (end.getZ() - begin.getZ());
		// cout << "Net_Cable_Corners: m1: " << m1 << " m2: " << m2 
		//	<< " " << begin << " " << end << endl;
		change_end_angle = ((fabs(m2 - m1) > fabs(m2 * 0.3)) && 
				(fabs(m1) > 0.001) && (fabs(m2) > 0.001));
	};
	// change_end_angle for correct drawing
	if (change_end_angle) end_angle = fmod(M_PI + end_angle, 2 * M_PI);

	width_on_x_begin = -cos(begin_angle) * 0.5 * depth;
	width_on_z_begin = sin(begin_angle) * 0.5 * depth;
	width_on_x_end = -cos(end_angle) * 0.5 * depth;
	width_on_z_end = sin(end_angle) * 0.5 * depth;
	left.setPos(begin.getX() - width_on_x_begin + 
				begin_translate * cos(my_angle),
			begin.getY() + height,
			begin.getZ() - width_on_z_begin - 
				begin_translate * sin(my_angle));
	right.setPos(
		end.getX() - width_on_x_end + end_translate * cos(my_angle), 
		end.getY() + height,
		end.getZ() - width_on_z_end - end_translate * sin(my_angle));
	ground_factor_1 = 1.0 - ground_depth_factor;
	ground_factor_2 = 2.0 - ground_factor_1;

	glBegin(GL_QUAD_STRIP);
		shade_color.colorForDraw();
		// left and right top on top of first side wall 
		left.vertexDraw(0, 0, 0);
		right.vertexDraw(0, 0, 0);
		// left and right top on low end of first side wall
		left.vertexDraw(ground_factor_1 * width_on_x_begin, -height,
				ground_factor_1 * width_on_z_begin);
		right.vertexDraw(ground_factor_1 * width_on_x_end, -height, 
				ground_factor_1 * width_on_z_end);	
		begin.colorForDraw();
		// left and right top for ground
		left.vertexDraw(ground_factor_1 * width_on_x_begin, -height,
				ground_factor_1 * width_on_z_begin);
		right.vertexDraw(ground_factor_1 * width_on_x_end, -height, 
				ground_factor_1 * width_on_z_end);	
		// left and right bottom for ground
		left.vertexDraw(ground_factor_2 * width_on_x_begin, - height, 
				ground_factor_2  * width_on_z_begin);
		right.vertexDraw(ground_factor_2 * width_on_x_end, -height, 
				ground_factor_2 * width_on_z_end);
		shade_color.colorForDraw();
		// left and right bottom on low end of second side wall
		left.vertexDraw(ground_factor_2 * width_on_x_begin, - height, 
				ground_factor_2  * width_on_z_begin);
		right.vertexDraw(ground_factor_2 * width_on_x_end, -height, 
				ground_factor_2 * width_on_z_end);
		// left and right bottom on top of second side wall
		left.vertexDraw(2.0 * width_on_x_begin, 0.0, 
				2.0 * width_on_z_begin);
		right.vertexDraw(2.0 * width_on_x_end, 0.0,
				2.0 * width_on_z_end);
	glEnd();
};

