/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <iostream>
#include <math.h>
#include <GL/gl.h>

#include "net_cable_packets.h"
#include "net_queue.h"

/** interner Faktor zur Geschwindigkeit des Pakettransports im Kabel */
const double		net_cable_packets_speed_factor = 0.5;

Net_Cable_Packets::Net_Cable_Packets(void)
/**
* Konstruktor, ist leer
*/
{
};

Net_Cable_Packets::~Net_Cable_Packets(void)
/**
* Destruktor, l&ouml;scht alle Pakete, die zum Kabel geh&ouml;ren
*/
{
	// cout << "Net_Cable_Packets destruktor." << endl;
	delAllPackets();
};

void Net_Cable_Packets::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt ein Paket des Kabels aus der XML-Datei 'infile',
 *  wird in der while-Schleife von Net_Cable::loadXML() je nach Bedarf
 *  mehrfach gerufen, <br>
 *  ruft u.a. Net_Queue::loadPacketClasses() */
{
	Packet		*cable_packet;

	cable_packet = NULL;
	Net_Queue::loadPacketClasses(infile, cable_packet);
	if (cable_packet) traffic.insert(traffic.end(), cable_packet);
};

void Net_Cable_Packets::saveXML(xmlTextWriterPtr outfile) const
/** speichert alle Pakete des Kabels in der XML-Datei 'outfile',
 *  ruft u.a. Packet::saveXML() bzw. dessen &Uuml;berladungen in den
 *  abgeleiteten Klassen */
{
	list<Packet*>::const_iterator	traffic_packets;

	traffic_packets = traffic.begin();
	while (traffic_packets != traffic.end())
	{
		if (*traffic_packets) (*traffic_packets)->saveXML(outfile);
		traffic_packets++;
	};
};

size_t Net_Cable_Packets::numPackets(void) const
/** gibt die derzeitige Anzahl der Pakete im Kabel zur&uuml;ck */
{
	return traffic.size();
};

void Net_Cable_Packets::findEndCorner(Packet* cable_packet,
					Point_3D &current_pos,
					Point_3D &end_corner) const
/** zu einem Paket 'cable_packet' wird die aktuelle Position 'current_pos'
 *  und dessen Zielposition 'end_corner' zur&uuml;ckgegeben, zu der es
 *  hintransportiert wird */
{
	double		posx, posy, posz;
	Point_3D	first_pos, end_pos;

	if (cable_packet == NULL) return;

	cable_packet->getPos(posx, posy, posz);	
	current_pos.setPos(posx, posy, posz);
	cable_packet->getFirstPos(posx, posy, posz);	
	first_pos.setPos(posx, posy, posz);
	if (first_pos.getDistance(getCorner(0), false) == 0.0)
		end_corner = getCorner(numCorners() - 1);
	else	end_corner = getCorner(0);
};


void Net_Cable_Packets::insertPacket(bool is_on_begin, Packet *new_packet,
					double depth,
					double ground_depth_factor,
					double angle_on_begin_or_end)
/** das Paket 'new_packet' wird am Anfang (mit 'is_on_begin' == true) oder
 *  Ende ('is_on_begin' == false) ins Kabel eingef&uuml;gt, das Kabel hat die
 *  Parameter Kabeldicke/Tiefe 'depth',
 *  Verh&auml;ltnis Bodenbreite zu Dicke 'ground_depth_factor' und
 *  Richtung des ersten Kabelst&uuml;cks 'angle_on_begin_or_end', <br>
 *  ruft u.a. movePacket(), weshalb 'angle_on_begin_or_end' zur Zeit nicht mehr
 *  gebraucht wird */
{
	double		packet_width, packet_height;
	double		packet_depth, posx, posy, posz;
	Point_3D	insert_pos;

	if (new_packet != NULL)
	{
		new_packet->getSize(packet_width, packet_height, packet_depth);
		new_packet->setSize(ground_depth_factor * depth,
					packet_height, packet_depth);
		if (is_on_begin)
			insert_pos = getCorner(0);
		else
		{
			insert_pos = getCorner( numCorners() - 1);
		};
		// calculate and set start position
		posx = insert_pos.getX();
		posy = insert_pos.getY() + packet_height / 2.0 + 0.001;
		posz = insert_pos.getZ();
		new_packet->setPos(posx, posy, posz);
		new_packet->setFirstPos(posx, posy, posz);
		movePacket(1.0, new_packet); // to get right angle
		new_packet->setPos(posx, posy, posz);
		new_packet->setFirstPos(posx, posy, posz);
		traffic.insert(traffic.end(), new_packet); // into list
		cout << "Net_Cable::insertPacket: " << new_packet << endl;
	};
};

bool Net_Cable_Packets::sniffPacket(bool is_on_begin,
					Packet* &first_packet) const
/** gibt 'true' zur&uuml;ck, wenn das erste Paket der internen Liste
 *  in Richtung Kabelanfang ('is_on_begin' == true) oder Kabelende
 *  ('is_on_begin' == false) transportiert wird, <br> wenn es am Anfang/Ende
 *  angekommen ist (Distanz zum Ziel <= 0.5), wird auch ein Zeiger auf
 *  das erste Paket in der internen Liste in 'first_packet' zur&uuml;ckgegeben,
 *  in allen anderen F&auml;llen wird 'first_packet' auf 'NULL' gesetzt; <br>
 *  ruft u.a. findEndCorner() */
{
	bool		give_out, is_for_interface;
	Point_3D	current_pos, end_pos;

	// cout << "Net_Cable::sniffPacket start" << endl;
	give_out = false;
	is_for_interface = false;
	if (traffic.size() > 0)		// have packet(s) ?
	{
		first_packet = NULL;
		if (*traffic.begin())
		{
			give_out = false;
			is_for_interface = false;
			// target is begin or end
			findEndCorner(*traffic.begin(), current_pos, end_pos);
			if (is_on_begin)
			{
				if (getCorner(0).getDistance(end_pos,
							false) == 0.0)
					is_for_interface = true;
			}
			else
				if (getCorner(numCorners() - 1).getDistance(
						end_pos, false) == 0.0)
					is_for_interface = true;
			// distance to target
			if (end_pos.getDistance(current_pos, false) <= 0.5)
				give_out = is_for_interface && true;
			else	give_out = false;
			if (give_out)
				first_packet = *traffic.begin();
			// (*traffic.begin())->copy(first_packet);
		};
	}
	else 	first_packet = NULL;
	// cout << "Net_Cable::sniffPacket end" << endl;
	return is_for_interface;
};

void Net_Cable_Packets::deletePacket(Packet *first_packet)
/** l&ouml;scht das erste Paket in der internen Liste, nur wenn 'first_packet'
 *  auch ein Zeiger auf das gleiche Paket ist, <br>
 *  normalerweise nachdem sniffPacket() den Zeiger geliefert hat */
{
	cout << "*** Delete ***: " << first_packet << endl;
	if (*traffic.begin() == first_packet)
		traffic.erase(traffic.begin());
};

void Net_Cable_Packets::delAllPackets(void)
/**
* l&ouml;scht alle Pakete des Kabels
*/
{
	list<Packet*>::iterator	traffic_packets;

	traffic_packets = traffic.begin();
	while (traffic_packets != traffic.end())
	{
		if (*traffic_packets) delete *traffic_packets;
		*traffic_packets = NULL;
		traffic_packets++;
	};
	traffic.erase(traffic.begin(), traffic.end());
};

double Net_Cable_Packets::calcSpeed(bool direction_beginend,
				Point_3D &current_pos,
				vector<Point_3D>::const_iterator &before,
				vector<Point_3D>::const_iterator &after) const
/** gibt die Soll-Geschwindigkeit eines Paketes zur&uuml;ck, das Paket ist an
 *  der Stelle 'current_pos' und hat die Transportrichtung
 *  Kabelanfang zu Kabelende ('direction_beginend' == true) bzw.
 *  Kabelende zu -anfang ('direction_beginend' == false),
 *  es befindet sich zur Zeit zwischen den zwei Kurvenpunkten 'before' und
 *  'after' (wobei 'before' in der Reihenfolge des Feldes der Kurvenpunkte
 *  immer vor 'after' liegt, unabh&auml;ngig von 'direction_beginend'), <br>
 *  ruft u.a. curveLength() */
{
	double	distance_before, distance_after;
	double	min_distance, full_distance, speed;

	// calculate distances
	curveLength(before, after, distance_before, distance_after,
			full_distance);
	if (direction_beginend)
	{
		distance_before += current_pos.getDistance(*before);
		distance_after += current_pos.getDistance(*after);
	}
	else
	{
		distance_before += current_pos.getDistance(*after);
		distance_after += current_pos.getDistance(*before);
	};
	min_distance = (distance_before < distance_after)?
			(distance_before):(distance_after);
	// speed as a hysterese over distance 
	speed = net_cable_packets_speed_factor / 
			(2 + exp(3 - min_distance/full_distance * 12.0));
	// way = 1 / (1 + exp(5.0 - distance_before/full_distance * 10.0));
	// cout << "Net_Cable_Packets: speed: " << speed << endl;
	// cout << "way: " << way << " " << distance_before << endl;
	return speed;
};

void Net_Cable_Packets::movePacket(double speed_factor,
					Packet *cable_packet) const
/** transportiert das Paket 'cable_packet' im Kabel Richtung Ziel mit
 *  dem Geschwindigkeitsfaktor der Animation 'speed_factor' */
{
	bool					direction_beginend;
	double					speed, angle;
	double					current_length, needed_length;
	double					x_pos, y_pos, z_pos, m, n;
	Point_3D				first_pos, current_pos;
	Point_3D				end_pos;
	vector<Point_3D>::const_iterator	before, after, first, last;

	if (cable_packet == NULL) return;
	cout << "Net_Cable::movePacket" << endl;
	// find and direction target of packet
	findEndCorner(cable_packet, current_pos, end_pos);
	cable_packet->getFirstPos(x_pos, y_pos, z_pos);
	first_pos.setPos(x_pos, y_pos, z_pos);
	if (first_pos.getDistance(getCorner(0), false) == 0.0)
		direction_beginend = true;
	else	direction_beginend = false;
	// find the curve points before and after current position
	getCurve(first, last);
	before = first;
	after = before + 1;
	if (!direction_beginend)
	{
		before = last;
		after = before - 1;
		last = first;
	};
	current_length = 0.0;
	needed_length = cable_packet->getMovedWay();
	// cout << "Net_Cable_Packets: current: " << current_length << " "
	//	<< "needed: " << needed_length << endl;
	while ((current_length < needed_length) && (before != last))
	{	// go from first to last until length reached
		if (direction_beginend) after = before + 1;
		else after = before - 1;
		current_length += before->getDistance(*after, false);
		before = after;
	};
	if (needed_length > 0.0)
	{
		if (direction_beginend)  before--;
		else before++;
	};
	// calculate speed and new direction angle
	speed = calcSpeed(direction_beginend, current_pos, before, after);
	speed *= speed_factor;
	angle = before->getAngleXZ(*after);
	// ----------------------------------------------------------
	// calculate new position of packet
	// cout << "angle: " << angle << " ";
	// cout << "x_pos: " << x_pos << endl;
	// cout << "Net_Cable_Packets: before: " << *before
	//	<< " after: " << *after << endl;
	y_pos = current_pos.getY();
	if (after->getX() == before->getX())
	{
		z_pos = current_pos.getZ() - sin(angle * M_PI / 180.0) * speed;
		x_pos = after->getX();
	}
	else
	{	// better with: z = m * x + n;
		m = (after->getZ() - before->getZ()) / 
			(after->getX() - before->getX());
		n = (before->getZ() - m * before->getX());
		if (fabs(after->getX() - before->getX()) >
			fabs(after->getZ() - before->getZ()))
		{
			x_pos = current_pos.getX() + 
					cos(angle * M_PI / 180.0) * speed;
			z_pos = m * x_pos + n;
		}
		else
		{
			z_pos = current_pos.getZ() - 
					sin(angle * M_PI / 180.0) * speed;
			x_pos = (z_pos - n) / m;
		};
	};
	if (2 * before->getDistance(*after) < before->getDistance(current_pos))
	{	// too far away from cable so reset to cable
		x_pos = before->getX();
		z_pos = before->getZ();
	};
	// set to end pos if end reached
	if (end_pos.getDistance(current_pos, false) < speed)
		current_pos.setPos(end_pos.getX(), y_pos, end_pos.getZ());
	else current_pos.setPos(x_pos, y_pos, z_pos);
	// cout << "Pos: " << current_pos << endl;
	cable_packet->setPos(current_pos.getX(), current_pos.getY(),
				current_pos.getZ());
	// cout << "angle: " << angle << endl;
	angle = fmodf(270.0 + angle, 360.0);
	cable_packet->setAngle(angle);
};

bool Net_Cable_Packets::getFirstPacketPosAngle(Point_3D &packet_pos,
						double &packet_angle) const
/** gibt 'true' zur&uuml;ck, wenn das Kabel mindestens ein Paket transportiert,
 *  sonst 'false', zus&auml;tzlich die Position 'packet_pos' des ersten Paketes
 *  in der internen Liste und dessen Transportrichtung / Rotationswinkel 
 *  'packet_angle' (0.0 - 360.0 Grad), diese Werte werden nicht ver&auml;ndert,
 *  wenn die Methode 'false' liefert */
{
	bool	has_packet;
	double	posx, posy, posz;
	Packet	*first_packet;

	first_packet = NULL;
	has_packet = false;
	if (traffic.size() > 0) first_packet = *traffic.begin();
	if (first_packet)
	{
		first_packet->getPos(posx, posy, posz);
		packet_pos.setPos(posx, posy, posz);
		packet_angle = first_packet->getAngle();
		has_packet = true;
	};
	return has_packet;
};

void Net_Cable_Packets::executeNet(double speed_factor)
/** f&uuml;hrt einen Schritt der Netzwerksimulation im Kabel aus, hier
 *  transportiert es einfach alle Pakete ein St&uuml;ck Richtung Ziel, <br>
 *  ruft u.a. movePacket() */
{
	list<Packet*>::iterator	traffic_packets;

	traffic_packets = traffic.begin();
	while (traffic_packets != traffic.end())
	{
		if (*traffic_packets)
			movePacket(speed_factor, *traffic_packets);
		traffic_packets++;
	};
};

void Net_Cable_Packets::setSize(double depth, double ground_depth_factor)
/**
 * passt die Gr&ouml;&szlig;e der Pakete des Kabels an die neue
 * Kabeldicke/Tiefe 'depth' und das neue Verh&auml;ltnis Kabelbodenbreite zu
 * Tiefe 'ground_depth_factor' an
 */
{
	double			packet_width, packet_height, packet_depth;
	list<Packet*>::iterator	traffic_packets;

	traffic_packets = traffic.begin();
	while (traffic_packets != traffic.end())
	{
		if (*traffic_packets)
		{
			(*traffic_packets)->getSize(packet_width,
						packet_height, packet_depth);
			(*traffic_packets)->setSize(
					ground_depth_factor * depth,
					packet_height, packet_depth);
		};
		traffic_packets++;
	};
};

void Net_Cable_Packets::draw(bool level_of_details_on,
				Point_3D last_camera_pos) const
/** zeichnet alle Pakete des Kabels, d.h. ruft dessen Zeichenmethode
 *  Packet::draw() mit den Parametern 'last_camera_pos' (letzte Kameraposition)
 *  und 'last_level_of_details_on' (Level of Detail an oder aus) */
{
	list<Packet*>::const_iterator	traffic_packets;

	traffic_packets = traffic.begin();
	while (traffic_packets != traffic.end())
	{
		if (*traffic_packets)
			(*traffic_packets)->draw(level_of_details_on,
							last_camera_pos);
		traffic_packets++;
	};
};

