/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <iostream>
#include <GL/gl.h>
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>

#include "config.h"
#include "constants.h"
#include "host.h"
#include "l3_router.h"
#include "network.h"

/** Name der Variablen zum Dateinamen des Netzwerkes im Info-Variablenbaum zum
 *  Dialog */
const char		network_varname_filename[] = "file name";
/** Name des Teilbaumes vom Netzwerk im Info-Variablenbaum zum Dialog */
const char		network_varname_network[] = "network";
/** Portnummer, auf dem der Paketserver auf Verbindungen wartet (listen) */
const unsigned short	network_packet_socket_port = 9000;

using namespace std;

Network::Network(void) : node_posy(1.89), cable_posy(0.0), filename("")
/** Konstruktor, Initialisierung von Variablen
 *  (node_posy, cable_posy, filename) */
{
};

Network::~Network(void)
/** Destruktor, ruft clear() zur L&ouml;schung aller Netzwerkelemente */
{
	cout << "destruktor Network." << endl;
	clear();
	// cout << "destruktor Network end." << endl;
};

void Network::clear(void)
/** l&ouml;scht alle Netzwerkelemente, also Knoten und Kabel */
{
	unsigned int			i;

	for(i = 0; i < nodes.size(); i++) delete nodes[i];
	nodes.erase(nodes.begin(), nodes.end());
	for(i = 0; i < cables.size(); i++) delete cables[i];
	cables.erase(cables.begin(), cables.end());
};

void Network::init(unsigned short new_server_port)
/** Initialisierung des Netzwerkes, Paketserver wird auf dem Port
 *  'new_server_port' (wenn > 0) gestartet, bzw. auf Port aus der Konstante
 *  network_packet_socket_port (wenn 'new_server_port' == 0); <br>
 *  zus&auml;tzlich werden nach clear() 10 Host, 1 L3_Router und 1 Kabel
 *  (Net_Cable) als Anfangselemente ins Netzwerk eingef&uuml;gt, wenn kein
 *  Dateiname (filename) bekannt ist, aus dem ein Netzwerk gelesen werden soll
 */
{
	unsigned int 	i;
	double		pos[11][2] = {{0.0, 0.0}, {3.0, 3.0}, {-3.0, 3.0}, 
					{-3.0, -3.0}, {3.0, -3.0}, 
					{-10.0, -10.0},	{-10.0, -6.0}, 
					{-10.0, -14.0}, {-14.0, -10.0}, 
					{-6.0, -10.0}, {10.0, 10.0}};
	double		begin_pos[2] = {-10.0, -10.0},
			end_pos[2] = {10.0, 10.0};

	// cout << "Network::init start" << endl;
	clear();
	if (filename.size() > 0) load();
	else
	{
		for(i = 0; i < 10; i++)
			insert_host(pos[i][0], pos[i][1]);
		insert_router(pos[10][0], pos[10][1]);
		insert_cable(begin_pos, end_pos);
	};
	if (new_server_port == 0) new_server_port = network_packet_socket_port;
	packet_socket_io.init(new_server_port);
	// cout << "Network::init end" << endl;
};

void Network::getInfos(Variable_Tree* &infos)
/** stellt den Variablenbaum f&uuml;r das Dialogfenster zusammen, Zeiger
 *  darauf in 'infos' zur&uuml;ck; <br>
 *  hiermit werden die Einstellungen des Netzwerkes selber (momentan nur der
 *  Dateiname) herausgegeben */
{
	String_Variable			*var_filename;

	current_infos.clear();

	var_filename = new String_Variable;
	var_filename->setName(network_varname_filename);
	var_filename->setLength(255, 1);
	var_filename->setValue(filename);
	current_infos.addVar(var_filename);
	current_infos.setName(network_varname_network);
	infos = &current_infos;
};

void Network::setInfos(Variable_Tree* infos)
/** liest die Werte aus dem Variablenbaum ein, der Zeiger 'infos' sollte
 *  auf den internen Baum zeigen, wird aber nicht beachtet; <br>
 *  hiermit werden die Einstellungen des Netzwerkes selber (momentan nur der
 *  Dateiname) angepasst */
{
	string		value_text;
	Variable*	current_variable;

	// the pointers in &infos should be the same as in current_infos
	if (current_infos.findVar(network_varname_filename, current_variable))
	{
		value_text = current_variable->getValue();
		filename = value_text;
	};
};

void Network::getInfos(double posxz[2], Variable_Tree* &infos)
/** liest die Einstellungen des Netzwerkelementes (Knoten oder Kabel) an
 *  Position (x = posxz[0], y egal, z = posxz[1]) ein und gibt sie &uuml;ber
 *  'infos' zur&uuml;ck (gibt sie an den Dialog weiter); <br>
 *  Node::getPos(), Node::getInfos(), Net_Cable::isOnCable(),
 *  Net_Cable::getInfos() */
{
	bool		node_found;
	unsigned int	i;
	double		x_current, y_current, z_current;

	infos = NULL;
	node_found = false;
	
	for(i = 0; i < nodes.size(); i++)
	{
		nodes[i]->getPos(x_current, y_current, z_current);
		if ((x_current == posxz[0]) && (z_current == posxz[1]))
		{
			nodes[i]->getInfos(infos);
			node_found = true;
		};
	};
	// no node found on that position
	if (!node_found)
		for(i = 0; i < cables.size(); i++)
			if (cables[i]->isOnCable(posxz[0], cable_posy, 
							posxz[1]))
			{
				cables[i]->getInfos(infos);
			};
	// cout << "Network::getInfos: " << node_found << endl;
};

void Network::setInfos(double posxz[2], Variable_Tree* infos)
/** weist die Netzwerkelemente (Knoten oder Kabel) an Position
 *  (x = posxz[0], y egal, z = posxz[1]) an, ihre Einstellungen aus dem
 *  Variablenbaum 'infos' neu einzulesen (nachdem sie per Dialog angepasst
 *  wurden), der Zeiger 'infos' sollte auf den internen Baum des Elementes
 *  zeigen, wird aber nicht beachtet; <br>
 *  ruft attachCablesToNodes() zur Zuordnung der Kabel zu den Netzknoten,
 *  Node::getPos(), Node::setInfos(), Net_Cable::isOnCable(),
 *  Net_Cable::setInfos() */
{
	bool			node_found;
	unsigned int		i;
	double			x_current, y_current, z_current;

	node_found = false;

	for(i = 0; i < nodes.size(); i++)
	{
		nodes[i]->getPos(x_current, y_current, z_current);
		if ((x_current == posxz[0]) && (z_current == posxz[1]))
		{
			nodes[i]->setInfos(infos);
			node_found = true;
		};
	};
	// no node found on that position
	if (!node_found)
		for(i = 0; i < cables.size(); i++)
			if (cables[i]->isOnCable(posxz[0], cable_posy, 
							posxz[1]))
			{
				cables[i]->setInfos(infos);
			};
	attachCablesToNodes();
};

void Network::insert_host(double new_xpos, double new_zpos)
/** f&uuml;gt einen Host an Position (x, y, z) =
 *  ('new_xpos', node_posy, 'new_zpos') ins Netzwerk ein, <br>
 *  node_posy ist eine interne Konstante zur H&ouml;henlage des Knoten,
 *  ruft attachCablesToNodes() zur Zuordnung der Kabel zu den Netzknoten */
{
	Host	*new_computer;

	// cout << "insert" << endl;
	new_computer = new Host();
	if (new_computer != NULL)
	{
		new_computer->init();
		new_computer->setPos(new_xpos, node_posy, new_zpos);
		new_computer->setAddress("192.168.1.1");
		// cout << new_computer->getAddress() << endl;
		nodes.insert(nodes.end(), new_computer);
	};
	attachCablesToNodes();
};

void Network::insert_router(double new_xpos, double new_zpos)
/** f&uuml;gt einen L3_Router an Position (x, y, z) =
 *  ('new_xpos', node_posy, 'new_zpos') ins Netzwerk ein, <br>
 *  node_posy ist eine interne Konstante zur H&ouml;henlage des Knoten,
 *  ruft attachCablesToNodes() zur Zuordnung der Kabel zu den Netzknoten */
{
	L3_Router	*new_router;

	new_router = new L3_Router();
	if (new_router != NULL)
	{
		new_router->init();
		new_router->setPos(new_xpos, node_posy, new_zpos);
		nodes.insert(nodes.end(), new_router);
	};
	attachCablesToNodes();
};

void Network::insert_cable(double begin_posxz[2], double end_posxz[2])
/** f&uuml;gt ein Netzkabel mit Beginn an Position (x, y, z) =
 *  ('begin_posxz[0]', cable_posy, 'begin_posxz[1]') und Ende an Position
 *  (x, y, z) = ('end_posxz[0]', cable_posy, 'end_posxz[1]') ins Netzwerk
 *  ein, <br>
 *  cable_posy ist eine interne Konstante zur H&ouml;henlage des Kables,
 *  ruft attachCablesToNodes() zur Zuordnung der Kabel zu den Netzknoten */
{
	Net_Cable	*new_cable;

	// cout << "insert cable start" << endl;
	new_cable = new Net_Cable();
	if (new_cable)
	{
		// cout << "init" << endl;
		new_cable->init();
		// cout << "Begin, end" << endl;
		new_cable->setBegin(begin_posxz[0], cable_posy, begin_posxz[1]);
		new_cable->setEnd(end_posxz[0], cable_posy, end_posxz[1]);
		/* new_cable->addCorner(begin_posxz[0] + 10.0, cable_posy, 
					begin_posxz[1] + 10.0);
		new_cable->addCorner(begin_posxz[0] + 7.0, cable_posy, 
					begin_posxz[1] + 8.0);
		new_cable->addCorner(begin_posxz[0] + 18.0, cable_posy, 
					begin_posxz[1] + 18.0); */
		// cout << "vector insert" << endl;
		cables.insert(cables.end(), new_cable);
		// cout << "insert cable end" << endl;
	};
	attachCablesToNodes();
};

void Network::move(double old_posxz[2], double new_posxz[2],
			bool move_node_with_cable)
/** verschiebt das Netzwerkelement (Knoten oder Kabelteil) an Position
 *  (x, z) = ('old_posxz[0]', 'old_posxz[1]') auf die neue Position (x, z) =
 *  ('new_posxz[0]', 'new_posxz[1]'), die Position auf der y-Achse wird nicht
 *  verglichen und bleibt erhalten, <br>
 *  mit 'move_node_with_cable' == 'true' wird ein Kabel, das an gleicher Stelle
 *  verl&auml;uft wie ein Netzknoten, zusammen mit diesem Knoten verschoben,
 *  mit 'false' wird nur der Knoten neu positioniert, das Kabel bleibt liegen,
 *  <br> innerhalb eines Kabels wird ein neuer Hilfspunkt (welcher die
 *  Bezierkurve des Kabels beeinflusst) eingef&uuml;gt, wenn an 'old_posxz'
 *  noch kein Hilfspunkt existiert */
{
	bool			found;
	unsigned int		i;
	double			x_current, y_current, z_current;

	found = false;
	for(i = 0; i < nodes.size(); i++)
	{
		nodes[i]->getPos(x_current, y_current, z_current);
		if ((x_current == old_posxz[0]) && 
			(z_current == old_posxz[1]))
		{
			nodes[i]->setPos(new_posxz[0], y_current,
						new_posxz[1]);
			found = true;
		};
	};
	if (!found || move_node_with_cable)
		for(i = 0; i < cables.size(); i++)
			if (cables[i]->isOnCable(old_posxz[0], cable_posy, 
							old_posxz[1]))
			{	// adds only a corner if there is no one
				cables[i]->addCorner(old_posxz[0], cable_posy, 
							old_posxz[1]);
				cables[i]->moveCorner(old_posxz, new_posxz);
			};
};

void Network::remove(double old_xpos, double old_zpos)
/** l&ouml;scht das Netzelement (Knoten oder Kabelteil) an Position (x, z)
 *  = ('old_xpos', 'old_zpos'), <br>
 *  bei Knoten wird die H&ouml;henlage old_ypos nicht beachtet; <br>
 *  bei Kabeln wird der Hilfspunkt (zur Anpassung der Bezierkurve) an der
 *  angegebenen Position und H&ouml;henlage cable_posy gel&ouml;scht,
 *  wenn ein Kabel danach aus weniger als zwei (2) Punkten besteht, wird es
 *  ganz gel&ouml;scht; <br>
 *  cable_posy ist eine interne Konstante zur H&ouml;henlage des Kables, <br>
 *  ruft attachCablesToNodes() zur Zuordnung der Kabel zu den Netzknoten */
{
	bool				node_deleted;
	unsigned int			i, num_corners;
	double				x_current, y_current, z_current;
	vector<Node*>::iterator		last_node;
	vector<Net_Cable*>::iterator	last_cable;

	node_deleted = false;
	for(i = 0; i < nodes.size(); i++)
	{
		nodes[i]->getPos(x_current, y_current, z_current);
		if ((x_current == old_xpos) && (z_current == old_zpos))
		{
			delete nodes[i];
			nodes[i] = NULL;
			/* move pointer from last */
			nodes[i] = nodes[ nodes.size() - 1 ];
			// cout << "count before: " << nodes.size() << endl;
			/* erase last */
			last_node = nodes.end();
			last_node--;
			nodes.erase(last_node);
			node_deleted = true;
			// cout << "count after: " << nodes.size() << endl;	
			// cout << "i: " << i << endl;
		};
	};
	// no node found on that position
	if (!node_deleted)
		for(i = 0; i < cables.size(); i++)
		{
			if (cables[i]->isOnCable(old_xpos, cable_posy, 
							old_zpos))
			{
				num_corners = cables[i]->removeCorner(old_xpos, 
							cable_posy, old_zpos);
				// delete whole cable if it has only 1 corner
				if (num_corners < 2)
				{
					delete cables[i];
					cables[i] = NULL;
					cables[i] = cables[cables.size()-1];
					last_cable = cables.end();
					last_cable--;
					cables.erase(last_cable);
				};
			};
		};
	attachCablesToNodes();
};

void Network::attachCablesToNodes(void)
/** verbindet die Kabel mit den Netzknoten, <br>
 *  d.h. von jedem Kabel werden die zwei Ethernet-Adressen ausgelesen, mit
 *  denen Kabelanfang und -ende verbunden sein sollen; jeder Knoten, der
 *  eine dieser Adressen hat, wird mit diesem Kabel verbunden, <br>
 *  f&uuml;r jedes Kabel wird bei gefundener Verbindung
 *  Net_Cable::testExchangeAddresses() gerufen (siehe dort) */
{
	size_t		i, j;
	double		pos_x, pos_y, pos_z;
	Point_3D	node_pos;
	MacAddress	begin_address, end_address;

	// clear all old connections
	for(j = 0; j< nodes.size(); j++)
	{
		if (nodes[j] != NULL) nodes[j]->detachCables();
	};
	for(i = 0; i < cables.size(); i++)
	{
		// for all cables read their addresses for begin and end
		if (cables[i] == NULL) continue;
		cables[i]->getAddresses(begin_address, end_address);
		// search in all nodes if one has this address
		for(j = 0; j < nodes.size(); j++)
		{
			if (nodes[j] == NULL) continue;
			nodes[j]->getPos(pos_x, pos_y, pos_z);
			node_pos.setPos(pos_x, pos_y, pos_z);
			if (nodes[j]->hasMacAddress(begin_address))
			{	// attach and check if addresses are on right
				// end of the cable
				nodes[j]->attachCable(cables[i]);
				cables[i]->testExchangeAddresses(node_pos,
								begin_address);
			};
			if (nodes[j]->hasMacAddress(end_address))
			{
				nodes[j]->attachCable(cables[i]);
				cables[i]->testExchangeAddresses(node_pos,
								end_address);
			};
		};
	};
};

void Network::load(const char filename_to_load[])
/** l&auml;dt ein Netzwerk aus der XML-Datei mit dem Namen 'filename_to_load'
 *  bzw. wenn dieser String leer ist, mit dem intern gemerkten Namen
 *  'filename', <br>
 *  zuvor werden alle Netzelemente (Knoten/Kabel) entfernt, am Ende wird
 *  wird die Netzstruktur neu aufgebaut:
 *  ruft attachCablesToNodes() zur Zuordnung der Kabel zu den Netzknoten */
{
	bool			found;
	int			state, depth;
	double			begin_posxz[2], end_posxz[2];
	xmlTextReaderPtr	infile;

	// delete all from old network
	clear();

	// load file with name 'filename'
	if (filename_to_load != NULL)
		if (filename_to_load[0] != 0) // strlen > 0
			filename = filename_to_load;
	cout << "Loading: " << filename << endl;
	infile = xmlNewTextReaderFilename(filename.c_str());
	cout << "infile: " << infile << endl;
	// read until tag with name 'netview' where found
	found = false; state = 1;
	while ((!found) && (state >= 0))
	{
		state = xmlTextReaderRead(infile);
		if (xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_netview) == 0)
			found = true;
	};
	depth = xmlTextReaderDepth(infile);
	state = xmlTextReaderRead(infile);	// get next node
	depth = xmlTextReaderDepth(infile);
	// read the whole file and load all cables, computers (Host) and
	// routers (L3_Router) defined in it
	while((state >= 0) && (depth > 0))
	{
		// cout << "state: " << state << " nodeType: " 
		//	<< xmlTextReaderNodeType(infile) << endl;
		// cables
		if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_cable) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
		{
			begin_posxz[0] = 0.0; begin_posxz[1] = 0.0;
			end_posxz[0] = 0.0; end_posxz[1] = 0.0;
			insert_cable(begin_posxz, end_posxz);
			cables[ cables.size() - 1 ] -> loadXML(infile);
		}
		// computers (hosts)
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_computer) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
		{
			insert_host(0.0, 0.0);
			nodes[ nodes.size() - 1 ] -> loadXML(infile);
		}
		// routers
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_router) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
		{
			insert_router(0.0, 0.0);
			nodes[ nodes.size() - 1 ] -> loadXML(infile);
		}
		// read next tag
		else state = xmlTextReaderRead(infile);
		state = xmlTextReaderReadState(infile);
		depth = xmlTextReaderDepth(infile);
	};
	// close file, free memory, reassign connections node - cable
	xmlTextReaderClose(infile);
	xmlFreeTextReader(infile);
	attachCablesToNodes();
};

void Network::save(void)
/** speichert das gesamte Netzwerk in eine XML-Datei mit dem Namen aus der
 *  internen Variable 'filename',
 *  'filename' wurde von der letzten Lade-Prozedur (load()) oder dem Dialog
 *  gesetzt (getInfos(), setInfos()) */
{
	unsigned int		i;
	xmlTextWriterPtr	outfile;

	cout << "Saving: " << filename << endl;
	outfile = xmlNewTextWriterFilename(filename.c_str(), 0);
#ifdef WIN32
#define HAVE_XMLTEXTWRITERSETINDENT 1
#endif
#ifdef HAVE_XMLTEXTWRITERSETINDENT
	xmlTextWriterSetIndent(outfile, 1);
#endif
	xmlTextWriterStartDocument(outfile, "1.0", "ISO-8859-1", "yes");
	xmlTextWriterStartElement(outfile, xml_syntax_netview);
	// all nodes and cables
	for(i = 0; i < nodes.size(); i++) nodes[i]->saveXML(outfile);
	for(i = 0; i < cables.size(); i++) cables[i]->saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	// close file, free memory
	xmlTextWriterEndDocument(outfile);
	xmlFreeTextWriter(outfile);
};

void Network::draw(bool with_cable_control_points)
/** zeichnet das gesamte Netzwerk, d.h. ruft die Zeichenmethoden der Knoten
 *  (Node::draw()) und Kabel (Net_Cable::draw()), <br>
 *  bei den Kabeln wird der Parameter 'with_cable_control_points' angegeben,
 *  auf 'true' werden auch die Hilfspunkte (und die Linien dazwischen)
 *  gezeichnet, mit deren Hilfe man die Bezierkurve des Kabels anpassen kann,
 *  wenn sie dagegen auf 'false' gesetzt ist, werden nur die Bezierkurven
 *  gezeichnet */
{
	Node*				current_node;
	Net_Cable*			current_cable;
	vector<Node*>::iterator		all_nodes, nodes_end;
	vector<Net_Cable*>::iterator	all_cables, cables_end;

	all_nodes = nodes.begin();
	nodes_end = nodes.end();
	while(all_nodes != nodes_end)
	{
		current_node = *all_nodes;
		if (current_node) current_node->draw();
		all_nodes++;
	};
	all_cables = cables.begin();
	cables_end = cables.end();
	while(all_cables != cables_end)
	{
		current_cable = *all_cables;
		if (current_cable)
			current_cable->draw(with_cable_control_points);
		all_cables++;
	};
};

void Network::drawSetCameraPos(bool level_of_details_on, Point_3D camera_pos)
/** aktualisiert die gespeicherten Werte in den Netzelementen zu
 *  Level of Detail ('level_of_details_on') und Kameraposition ('camera_pos'),
 *  mit deren Hilfe einzelne Teile der Elemente und Packete, dem Kameraabstand
 *  und Zustand von Level of Detail angepasst, neu gezeichnet werden; <br>
 *  ruft Node::drawSetCameraPos() und Net_Cable::drawSetCameraPos() */
{
	size_t	i;

	for(i = 0; i < nodes.size(); i++)
		if (nodes[i])
			nodes[i]->drawSetCameraPos(level_of_details_on,
							camera_pos);
	for(i = 0; i < cables.size(); i++)
		if (cables[i])
			cables[i]->drawSetCameraPos(level_of_details_on,
							camera_pos);
};

void Network::executeNetImportPacket(Packet* cable_packet)
/** importiert Paket 'cable_packet' ins Netzwerk, kommt meist &uuml;ber den
 *  Paketserver, <br>
 *  anhand der Quell-IP-Adresse des Paketes wird es in den Knoten importiert,
 *  der auch diese Adresse verwendet (Node::executeNetImportPacket()), <br>
 *  wenn kein Knoten gefunden wird, wird es gel&ouml;scht */
{
	size_t		i;
	IpAddress	source_address;
	IPv4_Packet	*ip_packet;

	if (cable_packet == NULL) return;

	ip_packet = NULL;
	// cout << "Network::executeNetImportPacket: " << cable_packet << endl;
	if (cable_packet->isProtocol(IPv4))
		ip_packet = reinterpret_cast<IPv4_Packet*>(cable_packet);
	if (ip_packet)
	{
		ip_packet->getSourceIpAddress(source_address);
		i = 0;
		// cout << "Network: Searchung node for: "
		//	<< source_address << endl;
		while ((i < nodes.size()) && (cable_packet)) // not found
		{
			if (nodes[i] == NULL) continue;
			if (nodes[i]->hasIpAddress(source_address))
			{
				nodes[i]->executeNetImportPacket(cable_packet);
				cable_packet = NULL;
			};
			i++;
		};
	};
	// delete if no node found
	if (cable_packet) delete cable_packet;
};

bool Network::executeNetCables(double speed_factor, Point_3D &camera_pos,
				double &camera_angle, bool nodes_need_turn,
				string &current_cable_name)
/** f&uuml;hrt einen Schritt der Netzwerksimulation in jedem Kabel aus,
 *  mit dem Geschwindigkeitsfaktor 'speed_factor' (Pakettransport), wenn
 *  'nodes_need_turn' auf 'false' steht;
 *  wenn es auf 'true' steht, sind alle Pakete in den Kabel angehalten und
 *  die Knoten drehen sich (in executeNetNodes()) in Empfangs- oder
 *  Senderichtung; <br>
 *  'camera_pos' (Position zum Fokus auf ein Paket), 'camera_angle' (Richtung
 *  des Paketes zum Fokus auf dieses) und 'current_cable_name' (Adressen
 *  von Kabelanfang und -ende) werden durch jedes Kabel neu gesetzt, wenn sie
 *  ein Paket transportieren; <br>
 *  ruft u.a. Net_Cable::executeNet(), Net_Cable::getFirstPacketPosAngle(),
 *  Net_Cable::getName(), <br>
 *  gibt 'true' zur&uuml;ck, wenn die Kameraposition und der -winkel angepasst
 *  wurde, also mindestens ein Paket in den Kabeln unterwegs ist, sonst 'false'
 */
{
	bool		changed;
	size_t		i;
	double		new_angle;
	Point_3D	new_pos;

	changed = false;
	for(i = 0; i < cables.size(); i++)
	{
		if (cables[i])
		{
			if (!nodes_need_turn) // stop if nodes do a rotation
				// else one step in network simulation
				cables[i]->executeNet(speed_factor);
			// if there is a packet in cable, focus camera on it
			if (cables[i]->getFirstPacketPosAngle(new_pos,
								new_angle))
			{
				camera_pos = new_pos;
				camera_angle = new_angle;
				changed = true;
				cables[i]->getName(current_cable_name); 
			};
		};
	};
	return changed;
};

bool Network::executeNetNodes(double speed_factor, Point_3D &camera_pos,
				double &camera_angle, bool &needs_turn,
				string &current_node_name)
/** f&uuml;hrt einen Schritt der Netzwerksimulation in jedem Netzknoten aus,
 *  mit dem Geschwindigkeitsfaktor 'speed_factor' (f&uuml;r Pakettransport und
 *  Knotenrotation) <br>
 *  'camera_pos' (Position zum Fokus auf ein Paket), 'camera_angle'
 *  (Ausrichtung des Knoten zum Fokus auf diesen), 'needs_turn' (Knoten
 *  befindet sich in einer Rotation zur Sende-/Empfangsrichtung),
 *  'current_node_name' (Name des Knoten mit Adressen) wird dabei von dem
 *  Knoten &uuml;bernommen, dessen (erstes) Paket am weitesten unten ist
 *  (Node::getFirstPacketPos()), <br>
 *  ruft u.a. Node::executeNet(), Node::getPosAngle(), Node::getName(), <br>
 *  gibt 'true' zur&uuml;ck, wenn die Kameraposition und der -winkel angepasst
 *  wurde, also mindestens ein Paket in den Knoten unterwegs ist, sonst 'false'
 */
{
	bool		changed;
	size_t		i;
	double		new_angle, min_angle;
	Point_3D	new_pos, min_pos;

	changed = false;
	needs_turn = false;
	min_angle = 0.0;
	for(i = 0; i < nodes.size(); i++)
	{
		if (nodes[i])
		{
			// one step in network simulation
			nodes[i]->executeNet(speed_factor);
			if (!needs_turn) needs_turn = nodes[i]->needsTurn();
			// focus on packet with lowest position
			if (nodes[i]->getFirstPacketPos(new_pos))
			{
				nodes[i]->getPosAngle(camera_pos,
							new_angle);
				if ((min_pos.getY() >= new_pos.getY()) ||
						(!changed))
				{
					min_pos = new_pos;
					min_angle = new_angle;
					nodes[i]->getName(current_node_name);
				};
				changed = true;
			};
		};
	};
	// cout << "Network: min_pos " << min_pos
	//	<< " min_angle: " << min_angle << endl;
	// cout << "Network::executeNetNodes changed: " << changed << endl;
	// if (needs_turn) cout << "needs turn" << endl;
	camera_pos = min_pos;
	camera_angle = min_angle;
	return changed;
};

bool Network::executeNet(double speed_factor, Point_3D &camera_pos,
				double &camera_angle, string &current_name)
/** f&uuml;hrt einen Schritt des Paketservers aus, importiert (falls vorhanden)
 *  alle Pakete von ihm ins Netzwerk, <br>
 *  f&uuml;hrt einen Schritt der Netzwerksimulation (mit dem
 *  Geschwindigkeitsfaktor 'speed_factor' f&uuml;r Pakettransport und
 *  Knotenrotation) in jedem Netzknoten und Kabel (nicht immer) aus,
 *  &uuml;bernimmt daraus die 'camera_pos' (zum Fokus auf ein Paket),
 *  den 'camera_angle' (Ausrichtung des Paketes/Knotens) und
 *  den 'current_name' (Knoten-/Kabelname, auf dem der Fokus liegt); <br>
 *  gibt 'true' zur&uuml;ck, wenn die Kameraposition und der -winkel angepasst
 *  wurde, also mindestens ein Paket in den Kabeln/Knoten unterwegs ist,
 *  sonst 'false' */
{
	bool		changed, needs_turn;
	Packet		*cable_packet;

	// packet server one step, import packets
	packet_socket_io.serverStep();
	needs_turn = false;
	do
	{
		cable_packet = NULL;
		packet_socket_io.getPacket(cable_packet);
		executeNetImportPacket(cable_packet);
	} while (cable_packet != NULL);
	// one step of network simulation in nodes and cables
	current_name = "";
	changed = executeNetNodes(speed_factor, camera_pos, camera_angle,
					needs_turn, current_name);
	if (changed) cout << "+++++" << endl;
	changed = changed | executeNetCables(speed_factor, camera_pos,
						camera_angle, needs_turn,
						current_name);
	if (changed) 
		cout << "-----------------------------------------" << endl;
	return changed;
};

