/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <iostream>
#include <math.h>
#include <unistd.h>

#include "constants.h"
#include "route_table.h"

/** Name der Variablen zu Routeneintr&auml;gen in der Routentabelle im
 *  Info-Variablenbaum zum Dialog */
const char 		route_table_varname_route[] = "route";
/** Name der Variablen zur Anzahl der Routen in der Tabelle im
 *  Info-Variablenbaum zum Dialog */
const char 		route_table_varname_num_routes[] = "number of routes";

// needed to set Stroke_Font size for output
/** maximale Anzahl Textzeichen je Zeile in der dargestellten
 *  Routentabelle, <br>
 *  3 x IP-Adresse (4*3+3 = 15) = 45 Zeichen plus 3 Leerzeichen und den Namen
 *  der Netzwerkkarte, zu der Pakete der Route geschickt werden sollen */
const unsigned short	route_table_max_text_length = 60;
/** maximale Anzahl Textzeilen in der dargestellten Routentabelle (nur
 *  dargestellter Bereich, Tabelle kann gr&ouml;&szlig;er sein) */
const unsigned short	route_table_max_text_lines = 8;

/** Farbe des Textes der dargestellten Tabelle, RGBA 0.0 - 1.0, <br>
 *  hier: wei&szlig; und undurchsichtig */
const double		route_table_text_color[4] = {0.95, 0.95, 0.95, 1.0};
/** Gr&ouml;&szlig;e des dargestellten Textes gesamt (derzeit 60 Zeichen mal
 *  8 Zeilen), zur Berechnung der Zeilenanfangspositionen */
const double		route_table_text_size[3] = {2.0, 1.0, 0.0};
/** maximaler Abstand der dargestellten Tabelle zur Kamera, bei dem mit Level
 *  of Detail noch etwas dargestellt wird */
const double		route_table_max_detail_distance = 3.0;
/** Farbe des Texthintergrundes einer Zeile, RGBA 0.0 - 1.0, wenn Tabellenzeile
 *  keine passende Route ist, <br> hier: gr&uuml;n und teildurchsichtig */
const double		route_table_color_not_found[4] = {0.0, 0.8, 0.0, 0.8};
/** Farbe des Texthintergrundes einer Zeile, RGBA 0.0 - 1.0, wenn Tabellenzeile
 *  eine PASSENDE Route ist, <br> hier: rot und teildurchsichtig */
const double		route_table_color_found[4] = {1.0, 0.0, 0.0, 0.8};
/** Farbe des Texthintergrundes einer Zeile, RGBA 0.0 - 1.0, wenn Tabellenzeile
 *  gerade verglichen wird, <br> hier: gelb und teildurchsichtig */
const double		route_table_color_current[4] = {0.5, 0.5, 0.0, 0.8};

/** Faktor f&uuml;r Wartezeit in Mikrosekunden zwischen zwei Routenvergleichen
 *  bei der Simulation eines Routers */
const unsigned long	route_table_wait_time_on_exec = 2000000;

Route_Table::Route_Table(void) : 
	search_finished(true),
	last_found_route(0), current_route(0), packet_for_routing(NULL)
/** Konstruktor, Initialisierung der Variablen: search_finished,
 *  last_found_route, current_route, packet_for_routing */
{
};

Route_Table::~Route_Table(void)
/** Destruktor, ruft nur 'clear' zum L&ouml;schen aller Tabelleneintr&auml;ge
 *  und des zu routenden Paketes */
{
	// cout << "Destruktor Route_Table start" << endl;
	clear();
	// cout << "Destruktor Route_Table end" << endl;
};

void Route_Table::clear(void)
/** l&ouml;scht alle Routeneintr&auml;ge in der Tabelle und das Paket, f&uuml;r
 *  das eine Route zu finden war */
{
	size_t	i;

	if (packet_for_routing != NULL) delete packet_for_routing;
	for(i = 0; i < route_array.size(); i++)
	{
		if (route_array[i]) delete route_array[i];
		route_array[i] = NULL;
	};
	route_array.erase(route_array.begin(), route_array.end());
};

void Route_Table::drawColorBackground(size_t current_i,
					class Rectangle &background) const
/** setzt die Farbe des Hintergrundes 'background' zu Tabelleneintrag
 *  'current_i', <br> 
 *  wenn Route passt (last_found_route) dann rot, aktuell verglichene Route
 *  (current_route) auf gelb und sonst gr&uuml;n */
{
	if (current_i == last_found_route)
	{	// color_found
		background.setColor(route_table_color_found[0],
					route_table_color_found[1],
					route_table_color_found[2],
					route_table_color_found[3]);
	}
	else if (current_i == current_route)
	{	// color_current
		background.setColor(route_table_color_current[0],
					route_table_color_current[1],
					route_table_color_current[2],
					route_table_color_current[3]);
	}
	else
	{	// color_not_found
		background.setColor(route_table_color_not_found[0],
					route_table_color_not_found[1],
					route_table_color_not_found[2],
					route_table_color_not_found[3]);
	};
};

void Route_Table::draw(bool level_of_details_on, Point_3D camera_pos) const
/** zeichnet die Routen-Tabelle ins 3D-Fenster, wenn Zustand
 *  'level_of_details_on' auf aus (false) oder Kamera 'camera_pos' nah genug
 *  dran ist, <br>
 *  das Paket wird auch gezeichnet: Packet::draw() mit 'level_of_details_on'
 *  und 'camera_pos', <br>
 *  ruft u.a. drawColorBackground() und Route::getRoute() */
{
	double		x_pos, y_pos, z_pos, sign, char_width;
	size_t		i;
	IpAddress	hop_address;
	Point_3D	start, pos;
	class Rectangle	background;

	// draw packet and take next hop address
	if (packet_for_routing != NULL)
	{
		packet_for_routing->draw(level_of_details_on, camera_pos);
		packet_for_routing->getNextHopAddress(hop_address);
	};
	if ((!level_of_details_on) || (camera_pos.getDistance(text_pos) <= 
					route_table_max_detail_distance))
	{
		// calculate top left position, where begin to draw
		x_pos = text_pos.getX() - text_size.getX() / 2.0;
		y_pos = text_pos.getY() + text_size.getY() / 2.0;
		z_pos = text_pos.getZ();
		start.setPos(x_pos, y_pos, z_pos);
		start.translate(1);
		// character width and line height
		char_width = text_size.getX() / route_table_max_text_length;
		y_pos = - text_size.getY() / route_table_max_text_lines;
		// sign: draw background before or in front of text
		sign = text_pos.getZ() / fabs(text_pos.getZ());
		pos.setPos(0.0, y_pos, 0.0);
		// set size of background rectangle for route lines
		background.setPos(0.0, 0.0, text_size.getX(), y_pos);
		// show address of packet
		if (packet_for_routing != NULL)
			route_text_output.print(hop_address.getAddress());
		for(i = 0; i < route_array.size(); i++)
		{
			pos.translate(i * 1.0);
			glTranslated(0.0, -0.01, -0.001 * sign);
			// choose color and draw background
			drawColorBackground(i, background);
			background.draw();
			glTranslated(0.0, 0.01, 0.001 * sign);
			glTranslated(char_width, 0.0, 0.0);
			pos.translate(1.0);
			// draw route entry
			route_text_output.print(
				route_array[i]->getRoute().c_str());
			pos.translate((i + 1) * (-1.0));
			glTranslated(-char_width, 0.0, 0.0);
		};
		start.translate(-1);
	};
};

void Route_Table::init(void)
/** Initialisierung: eine Route mit Null-Werten wird erzeugt,
 *  Farbe mit setColor(), Gr&ouml;&szlig;e mit setSize() und
 *  Position mit setPos() */
{
	Route	*new_route;

	// cout << "Route_Table::init" << endl;
	new_route = new Route();
	if (new_route) 
	{
		new_route->init();
		route_array.insert(route_array.end(), new_route);
	};
	setColor(route_table_text_color[0], route_table_text_color[1],
		route_table_text_color[2], route_table_text_color[3]);
	setSize(route_table_text_size[0], route_table_text_size[1],
		route_table_text_size[2]);
	route_text_output.setPos(0.0, 0.0, 0.0);
	// cout << "Route_Table::init end" << endl;
};

void Route_Table::setColor(double new_red, double new_green,
				double new_blue, double new_alpha)
/** setzt die Farbe des dargestellten Textes auf 'new_red' (Rot),
 *  'new_green' (Gr&uuml;n), 'new_blue' (Blau) und 'new_alpha'
 *  (Transparenz, 1.0 undurchsichtig), Wertebereich jeweils 0.0 - 1.0 */
{
	route_text_output.setColor(new_red, new_green, new_blue, new_alpha);
};

void Route_Table::setPos(double new_centerx, double new_centery,
				double new_centerz)
/** setzt die Position des dargestellten Textes (Mittelpunkt des rechteckigen
 *  Bereiches) x,y,z */
{
	text_pos.setPos(new_centerx, new_centery, new_centerz);
};

void Route_Table::setSize(double new_width, double new_height,
				double new_depth)
/** setzt die Gr&ouml;&szlig;e des dargestellten Textes (Bounding-Box),
 *  'new_width' = Breite, 'new_height' = H&ouml;he, 'new_depth' = Tiefe, <br>
 *  die Ausma&szlig;e des gesamten Textbereiches, nicht nur eines Zeichens */
{
	route_text_output.setSize(new_width / route_table_max_text_length,
				new_height / route_table_max_text_lines / 1.5);
	text_size.setPos(new_width, new_height, new_depth);
};

void Route_Table::getInfos(Variable_Tree* &infos)
/** stellt den Variablenbaum f&uuml;r das Dialogfenster zusammen, Zeiger
 *  darauf in 'infos' zur&uuml;ck, <br>
 *  Anzahl der Routen und alle Routeneintr&auml;ge (Route::getRoute()) */
{
	size_t				i;
	char				current_i[10];
	string				current_varname;
	Numeric_Variable<size_t>	*num_routes;
	String_Variable			*route_text;

	// cout << "Route_Table::getInfos start" << endl;
	current_infos.clear();
	num_routes = new Numeric_Variable<size_t>;
	if (num_routes)
	{	// number of routes
		num_routes->setName(route_table_varname_num_routes);
		num_routes->setLength(1, 0);
		num_routes->setMaxmin(20, 0);
		num_routes->setValue(route_array.size());
		current_infos.addVar(num_routes);
	};
	for(i = 0; i < route_array.size(); i++)
	{	// all route entries
		if (route_array[i] == NULL) continue;
		route_text = new String_Variable;
		if (route_text)
		{
			snprintf(current_i, sizeof(current_i), "%d", i + 1);
			current_varname = string(route_table_varname_route);
			current_varname += string(" ") + current_i;
			route_text->setName(current_varname);
			route_text->setLength(255, 0);
			route_text->setValue(route_array[i]->getRoute());
			current_infos.addVar(route_text);
		};
	};
	infos = &current_infos;
	// cout << "Route_Table::getInfos end" << endl;
};

void Route_Table::resize(size_t new_number_of_routes)
/** vergr&ouml;&szlig;ert/verkleinert das Feld der Routeneintr&auml;ge:
 *  entsprechende Anzahl an Routen wird hinzugef&uuml;gt (und initialisiert)
 *  oder am Ende beginnend gel&ouml;scht */
{
	size_t	i, max_size_t;
	Route	*new_route;

	while (new_number_of_routes > route_array.size())
	{	// add entries
		new_route = new Route();
		if (new_route) 
		{
			new_route->init();
			route_array.insert(route_array.end(), new_route);
		}
		else return;
	};
	if (new_number_of_routes < route_array.size())
	{	// delete entries
		max_size_t = 0;
		max_size_t--;
		for(i = route_array.size() - 1; 
			(i >= new_number_of_routes) && (i != max_size_t); i--)
		{
			delete route_array[i];
		};
	};
	route_array.resize(new_number_of_routes);		
};

void Route_Table::setInfos(Variable_Tree* infos)
/** liest die Werte aus dem Variablenbaum ein, der Zeiger 'infos' sollte
 *  auf den internen Baum zeigen, wird aber nicht beachtet, <br>
 *  Anzahl der Routen und alle Routeneintr&auml;ge */
{
	size_t				i;
	size_t				value;
	char				current_i[10];
	string				current_varname, value_text;
	Variable			*current_var;
	Numeric_Variable<size_t>	*num_routes;

	// cout << "Route_Table::setInfos start" << endl;
	// number of routes
	if (current_infos.findVar(route_table_varname_num_routes, current_var))
	{
		num_routes = reinterpret_cast< Numeric_Variable<size_t>* >
				(current_var);
		num_routes->getValue(0, value);
		resize(value);
	};
	// cout << "Route_Table::setInfos set routes" << endl;
	for(i = 0; i < route_array.size(); i++)
	{	// route entries
		if (route_array[i] == NULL) continue;
		snprintf(current_i, sizeof(current_i), "%d", i + 1);
		current_varname = string(route_table_varname_route);
		current_varname += string(" ") + current_i;
		if (current_infos.findVar(current_varname, current_var))
		{
			value_text = current_var->getValue();
			route_array[i]->setRoute(value_text);
		};
	};
	// cout << "Route_Table::setInfos end" << endl;
};

void Route_Table::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt die gesamte Routentabelle aus der XML-Datei 'infile', <br>
 *  ruft u.a. resize() und Route::loadXML() */
{
	int	state, depth;

	// cout << "Route_Table::loadXML start" << endl;
	resize(0);
	do
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
				xml_syntax_route) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
		{
			resize( route_array.size() + 1 );
			route_array[route_array.size() - 1]->loadXML(infile);
		};
	} while ((state >= 0) && (depth > 2));
	// cout << "Route_Table::loadXML end" << endl;
};

void Route_Table::saveXML(xmlTextWriterPtr outfile) const
/** speichert die gesamte Routentabelle in der XML-Datei 'outfile', <br>
 *  ruft u.a. Route::saveXML() */
{
	size_t	i;

	xmlTextWriterStartElement(outfile, xml_syntax_route_table);
	for(i = 0; i < route_array.size(); i++)
		if (route_array[i]) route_array[i]->saveXML(outfile);
	xmlTextWriterEndElement(outfile);
};

string Route_Table::findInterface(IPv4_Packet *ip_packet,
					Route* &found_route)
/** sucht zum Paket 'ip_packet' eine passende Route 'found_route', der Name
 *  der Netzwerkkarte (interface name / device name) aus der Route wird
 *  zur&uuml;ckgegeben, <br>
 *  wenn keine Route passt, ist 'found_route' auf 'NULL' und es wird ein leerer
 *  String ("") zur&uuml;ckgegeben, <br>
 *  zur Zeit von Host genutzt, alle Eintr&auml;ge werden auf einmal verglichen,
 *  <br> nutzt Route::matchRoute() und Route::getInterface() */
{
	short		most_matching, current_matching;
	size_t		num_route, i;
	string		interface_name;

	interface_name = "";
	found_route = NULL;
	most_matching = -1;
	num_route = 0;
	for(i = 0; i < route_array.size(); i++)
	{
		if (route_array[i] == NULL) continue;
		current_matching = route_array[i]->matchRoute(ip_packet);
		if (current_matching > most_matching)
		{
			most_matching = current_matching;
			num_route = i;
		};
	};
	cout << "Route: " << (num_route + 1) << " Matching: " 
		<< int(most_matching) << endl;
	if (route_array.size() > 0)
		if ((route_array[num_route]) && (most_matching >= 0))
		{
			interface_name = route_array[num_route]->getInterface();
			found_route = route_array[num_route];
			last_found_route = num_route;
		};
	return interface_name;
};

bool Route_Table::processPacket(IPv4_Packet* ip_packet)
// returns true if there was no packet for routing in here
/** Paket 'ip_packet' wird zum Routing in mehreren Schritten genutzt,
 *  hier intern gespeichert, alle bisherigen Suchergebnisse werden
 *  zur&uuml;ckgesetzt, <br>
 *  gibt true zur&uuml;ck, wenn vor Aufruf kein Paket enthalten war, <br>
 *  nur zusammen mit executeNet(), foundRoute(), getPacketPos() nutzbar */
{
	bool	has_packet;

	has_packet = (packet_for_routing != NULL);
	if (!has_packet)
	{
		if (ip_packet)
		{
			// save packet, reset all former results
			packet_for_routing = ip_packet;
			current_route = static_cast<size_t>(-1);
			last_found_route = static_cast<size_t>(-1);
			search_finished = false;
		};
	};
	return (!has_packet);
};

bool Route_Table::foundRoute(IPv4_Packet* &ip_packet, Route* &found_route)
/** wenn die Suche der Route abgeschlossen ist (alle Eintr&auml;ge verglichen),
 *  wird die gefundene Route 'found_route' hier zusammen mit dem Zeiger auf das
 *  Paket 'ip_packet' zur&uuml;ckgegeben, <br>
 *  bei erfolgloser Suche ist 'found_route' auf 'NULL', ip_packet hat aber
 *  trotzdem den Zeiger aufs Paket, <br>
 *  gibt 'true' zur&uuml;ck, wenn Suche abgeschlossen ist, sonst 'false' und
 *  'ip_packet' und 'found_route' bleiben unver&auml;ndert, <br>
 *  nur zusammen mit executeNet(), getPacketPos(), processPacket() nutzbar */
{
	// cout << "Route_Table::foundRoute" << endl;
	if ((search_finished) && (last_found_route != static_cast<size_t>(-1)))
		if (route_array[last_found_route] != NULL)
		{
			ip_packet = packet_for_routing;
			packet_for_routing = NULL;
			found_route = route_array[last_found_route];
		};
	// no route found
	if ((search_finished) && (last_found_route == static_cast<size_t>(-1)))
	{
		ip_packet = packet_for_routing;
		packet_for_routing = NULL;
		found_route = NULL;
	};
	return search_finished;
};

void Route_Table::executeNet(double speed_factor)
/** vergleicht den ersten/n&auml;chsten Routeneintrag mit dem gespeicherten
 *  Paket, falls die Suche noch nicht abgeschlossen ist (alle verglichen), <br>
 *  von L3_Router genutzt, <br>
 *  nur zusammen mit foundRoute(), getPacketPos(), processPacket() nutzbar */
{
	double	sleep_time;
	short	most_matching, current_matching;

	// search current entry and sleep
	if (current_route == static_cast<size_t>(-1)) current_route = 0;
	else
	{
		current_route++;
		sleep_time = route_table_wait_time_on_exec / speed_factor;
		usleep(static_cast<unsigned int>(rint(sleep_time)));
	};
	// match entry (better than last found ?) or finish
	if (current_route >= route_array.size()) search_finished = true;
	if ((!search_finished) && (packet_for_routing != NULL))
	{
		most_matching = -1;
		if (last_found_route != static_cast<size_t>(-1))
			most_matching = route_array[last_found_route]->
						matchRoute(packet_for_routing);
		current_matching = route_array[current_route]->
						matchRoute(packet_for_routing);
		if (current_matching > most_matching)
			last_found_route = current_route;
	};
};

bool Route_Table::getPacketPos(Point_3D &packet_pos) const
/** gibt die Position 'packet_pos' des gespeicherten Paketes zum Finden einer
 *  Route zur&uuml;ck, au&szlig;erdem 'true', falls solch ein Paket bekannt
 *  ist, sonst 'false' und 'packet_pos' bleibt unver&auml;ndert, <br>
 *  nur zusammen mit executeNet(), foundRoute(), processPacket() nutzbar */
{
	bool	has_packet;
	double	pos_x, pos_y, pos_z;

	has_packet = (packet_for_routing != NULL);
	if (has_packet)
	{
		packet_for_routing->getPos(pos_x, pos_y, pos_z);
		packet_pos.setPos(pos_x, pos_y, pos_z);
	};	 
	return has_packet;
};

