/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include <iostream>

#include "host.h"
#include "http_service.h"
#include "udp_service_packet.h"
#include "constants.h"

using namespace std;

/** Name des Teilbaumes vom Host im Info-Variablenbaum zum Dialog */
const char	host_varname_host[] = "host";
/** Name der Variablen zum Name des Hosts im Info-Variablenbaum zum Dialog */
const char	host_varname_name[] = "hostname";
/** Name des Teilbaumes der Netzwerkkarte im Info-Variablenbaum zum Dialog */
const char	host_varname_network_card[] = "network card";
/** Name des Teilbaumes der Routentabelle im Info-Variablenbaum zum Dialog */
const char	host_varname_route_table[] = "route table";
/** Gr&ouml;&szlig;e des Hosts (Breite, H&ouml;he, Tiefe) */
const double	host_size[3] = {2.0, 4.0, 2.0};
/** Faktor, um den sich die Kabel verk&uuml;rzen sollen, damit sie richtig
 *  am Host 'andocken', normalerweise gehen sie genau bis zum Mittelpunkt des
 *  Hosts */
const double	host_radius_factor_for_cable = 1 / 4.5;
/** Farbe des Hosts, RGBA 0.0 - 1.0, Alpha-Transparenz = 1 = undurchsichtig */
const double	host_color[4] = {1.0, 1.0, 0.0, 1.0};

Host::Host(void) : direction_angle(0.0), last_level_of_details_on(true),
			display_list_host(0), display_list_packets(0)
/** Konstruktor, Initialisierung der internen Variablen (direction_angle,
 *  last_level_of_details_on, display_list_host, display_list_packets) */
{
};

Host::~Host(void)
/** Destruktor, l&ouml;scht die Displaylisten */
{
	// cout << "Destruktor host begin." << endl;
	if (glIsList(display_list_host)) glDeleteLists(display_list_host, 1);
	if (glIsList(display_list_packets))
		glDeleteLists(display_list_packets, 1);
	// cout << "Destruktor host end." << endl;
};

void Host::draw(void) const
/** zeichnet den Host, d.h. ruft die Displaylisten auf */
{
	// cout << "Host draw" << endl;
	if (glIsList(display_list_host)) glCallList(display_list_host);
	if (glIsList(display_list_packets)) glCallList(display_list_packets);
};

void Host::drawSetCameraPos(bool level_of_details_on, Point_3D camera_pos)
/** aktualisiert die internen Werte zu 'last_camera_pos' (letzte Kameraposition
 *  relativ zur Hostposition) und 'last_level_of_details_on' (Level of Detail
 *  an oder aus), Displayliste der Pakete inklusive ARP-Cache
 *  (abh&auml;ngig von Level of Detail) wird neu gezeichnet */
{
	last_camera_pos.setPos(camera_pos.getX() - pos.getX(),
				camera_pos.getY() - pos.getY(),
				camera_pos.getZ() - pos.getZ());
	last_level_of_details_on = level_of_details_on;
	updateDisplaylistPackets();
};

void Host::updateDisplaylistHost(void)
/** zeichnet die Displayliste der Hostelemente neu: W&auml;nde,
 *  Transportschienen, Quader der Netzwerkkarten und TCP/UDP-Schlangen,
 *  Boden und 'Dach',
 *  vor allem nach Positions- oder Gr&ouml;&szlig;en&auml;nderung */
{
	size_t	i;

	if (!glIsList(display_list_host)) return;

	// cout << "Host::updateDisplaylistHost start list:" 
	// 	<< display_list_host << endl;
	glNewList(display_list_host, GL_COMPILE);
	pos.translate(1);
	glScaled(width / 2.0, height / 4.0, depth / 2.0);
	glRotated(direction_angle, 0, 1, 0);

	// cout << "Poles: " << endl; 
	for(i = 0; i < 4; i++) poles[i].draw();
	walls[0].draw();
	walls[1].draw();
	top.draw();
	bottom.draw();

	// cout << "interface " << endl;
	network_card.drawQueue();
	// cout << "tcp queue" << endl;
	tcp_queue.drawQueue();
	udp_queue.drawQueue();

	glRotated(-direction_angle, 0, 1, 0);
	glScaled(2.0 / width, 4.0 / height, 2.0 / depth);
	pos.translate(-1);
	glEndList();
	// cout << "Host::updateDisplaylistHost end" << endl;
};

void Host::updateDisplaylistPackets(void)
/** zeichnet die Displayliste der Pakete neu, inklusive ARP-Cache d.h. ruft die
 *  Zeichenmethoden von Interface, TcpQueue und UdpQueue
 *  und &uuml;bergibt denen 'last_level_of_details_on' und 'last_camera_pos' */
{
	if (!glIsList(display_list_packets)) return;

	glNewList(display_list_packets, GL_COMPILE);
	pos.translate(1);
	glScaled(width / 2.0, height / 4.0, depth / 2.0);
	glRotated(direction_angle, 0, 1, 0);

	// cout << "interface " << endl;
	network_card.drawPackets(last_level_of_details_on, last_camera_pos);
	network_card.drawArpCache(last_level_of_details_on, last_camera_pos);
	// cout << "tcp queue" << endl;
	tcp_queue.drawPackets(last_level_of_details_on, last_camera_pos);
	udp_queue.drawPackets(last_level_of_details_on, last_camera_pos);

	glRotated(-direction_angle, 0, 1, 0);
	glScaled(2.0 / width, 4.0 / height, 2.0 / depth);
	pos.translate(-1);
	glEndList();
};


void Host::init(void)
/** Initialisierung: TcpQueue, UdpQueue, Routentabelle, Netzwerkkarte, Dienste,
 *  Displaylisten, ruft setColor(), setPos(), setSize() */
{
	HTTP_Service	*new_service;

	// cout << "Host::init start" << endl;
	tcp_queue.init();
	udp_queue.init();
	route_table.init();
	network_card.init();
	new_service = NULL;
	new_service = new HTTP_Service;
	if (new_service != NULL)
	{
		new_service->init();
		services.insert(services.end(), new_service);
	};

	// cout << "Host::init graphics" << endl;
	display_list_host = glGenLists(1);
	display_list_packets = glGenLists(1);
	// cout << "Colors" << endl;
	// cout << "Pos Size" << endl;
	setColor(host_color[0], host_color[1], host_color[2], host_color[3]);
	setPos(0.0, 0.0, 0.0);
	setSize(host_size[0], host_size[1], host_size[2]);
	// cout << mac_address << endl;
	// cout << "Host::init end" << endl;
};

void Host::getInfos(Variable_Tree* &infos)
/** stellt den Variablenbaum f&uuml;r das Dialogfenster zusammen, Zeiger
 *  darauf in 'infos' zur&uuml;ck */
{
	String_Variable	*var_name;
	Variable_Tree	*var_tree_network_card, *var_tree_route_table;

	current_infos.clear();
	// hostname
	var_name = new String_Variable;
	if (var_name)
	{
		var_name->setName(host_varname_name);
		var_name->setLength(255, 0);
		var_name->setValue(name);
		current_infos.addVar(var_name);
	};
	// network interface, routing table
	network_card.getInfos(var_tree_network_card);
	if (var_tree_network_card)
		var_tree_network_card->setName(host_varname_network_card);
	current_infos.addChild(var_tree_network_card);
	route_table.getInfos(var_tree_route_table);
	if (var_tree_route_table)
		var_tree_route_table->setName(host_varname_route_table);
	current_infos.addChild(var_tree_route_table);
	current_infos.setName(host_varname_host);
	infos = &current_infos;
};

void Host::setInfos(Variable_Tree* infos)
/** liest die Werte aus dem Variablenbaum ein, der Zeiger 'infos' sollte
 *  auf den internen Baum zeigen, wird aber nicht beachtet */
{
	string		value_text;
	Variable	*current_var;
	Variable_Tree	*current_tree;

	if (current_infos.findVar(host_varname_name, current_var))
	{
		value_text = current_var->getValue();
		name = value_text;
	};
	if (current_infos.findChild(host_varname_network_card, current_tree))
		network_card.setInfos(current_tree);
	if (current_infos.findChild(host_varname_route_table, current_tree))
		route_table.setInfos(current_tree);
};

char* Host::getAddress(void) const
/** gibt die IP-Adresse der Netzwerkkarte zur&uuml;ck */
{
	return network_card.getIpAddress();
};

void Host::setAddress(char *ip_string)
/** setzt die IP-Adresse der Netzwerkkarte auf die in 'ip_string' angegebene */
{
	network_card.setIpAddress(ip_string);
};

void Host::getPos(double &current_centerx, double &current_centery,
			double &current_centerz) const
/** gibt die Position (den Mittelpunkt) des Hosts zur&uuml;ck (x, y, z) */
{
	Cuboid::getPos(current_centerx, current_centery, current_centerz);
};

void Host::getPosAngle(Point_3D &current_pos, double &current_angle) const
/** gibt Position (den Mittelpunkt, in 'current_pos') und den aktuellen
 *  Rotationswinkel (in 'current_angle') zur&uuml;ck */
{
	current_pos = pos;
	current_angle = direction_angle;	
};

void Host::setColor(double new_red,  double new_green, double new_blue,
			double new_alpha)
/** setzt die Farbe des Hosts auf 'new_red' (Rot), 'new_green' (Gr&uuml;n),
 *  'new_blue' (Blau) und 'new_alpha' (Transparenz, 1.0 undurchsichtig),
 *  Wertebereich jeweils 0.0 - 1.0 */
{
	int		i;

	for(i = 0; i < 4; i++)
		poles[i].setColor(new_red, new_green, new_blue, new_alpha);
	for(i = 0; i < 2; i++)
		walls[i].setColor(new_red, new_green, new_blue, new_alpha);
	top.setColor(new_red, new_green, new_blue, new_alpha);
	bottom.setColor(new_red, new_green, new_blue, new_alpha);
	updateDisplaylistHost();
	updateDisplaylistPackets();
};

void Host::setPos(double new_centerx, double new_centery,
                        double new_centerz)
/** setzt die Position (den Mittelpunkt) des Hosts (x, y, z) */
{
	Cuboid::setPos(new_centerx, new_centery, new_centerz);
	updateDisplaylistHost();
	updateDisplaylistPackets();
};

void Host::setSize(double new_width, double new_height, double new_depth)
/** setzt die Gr&ouml;&szlig;e des Hosts (Bounding-Box),
 *  'new_width' = Breite, 'new_height' = H&ouml;he, 'new_depth' = Tiefe, <br>
 *  positioniert die Elemente neu, zeichnet die Displaylisten neu */
{
	Cuboid::setSize(new_width, new_height, new_depth);
	Host::updatePos();
	updateDisplaylistHost();
	updateDisplaylistPackets();

};

void Host::updatePos(void)
/** Position und Gr&ouml;&szlig;e der einzelnen grafischen Elemente (inklusive
 *  Netzwerkkarten, TCP- und UDP-Schlange) wird neu bestimmt,
 *  unabh&auml;ngig von der Bounding-Box des Routers, da sie beim Zeichnen
 *  nochmal skaliert werden */
{
	// x*y*z = 2 * 4 * 2

	int	i;

	tcp_queue.setPos(0.0, 0.475, 0.0);
	tcp_queue.setSize(1.5, 0.95, 1.0);
	udp_queue.setPos(0.0, 0.475, 0.0);
	udp_queue.setSize(1.5, 0.95, 1.0);
	network_card.setPos( 0.0, -0.925, 0.0);
	network_card.setSize(1.5,  1.85,  1.0);
	for(i = 0; i < 4; i++)
		poles[i].setSize(0.05, 3.8, 0.05);
	poles[0].setPos(-0.325, 0.0, -0.4);
	poles[1].setPos( 0.325, 0.0, -0.4);
	poles[2].setPos( 0.325, 0.0,  0.4);
	poles[3].setPos(-0.325, 0.0,  0.4);
	// cout << "walls " << endl;
	for(i = 0; i < 2; i++) 
		walls[i].setSize(0.05, 3.8, 1.0);
	walls[0].setPos(-0.75, 0.0, 0.0);
	walls[1].setPos( 0.75, 0.0, 0.0);
	top.setSize(2.0, 0.1, 2.0);
	top.setPos(0.0, 1.95, 0.0);
	bottom.setSize(2.0, 0.1, 2.0);
	bottom.setPos(0.0, -1.95, 0.0);
};

void Host::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt die Einstellungen des Hosts aus der XML-Datei 'infile',
 *  Hostname, Netzwerkkarte, Routentabelle, TCP- und UDP-Schlangen */
{	
	int	state, depth;
	xmlChar	*content;

	// cout << "Host::loadXML start" << endl; 
	if (xmlTextReaderHasAttributes(infile))
	{	// hostname
		content = xmlTextReaderGetAttribute(infile, 
				xml_syntax_name);
		if (content) name = reinterpret_cast<char*>(content);
	};
	do
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		// position
		if (xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_pos) == 0)
			pos.loadXML(infile);
		// network interface, routing table
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_interface) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
				network_card.loadXML(infile);
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_route_table) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
				route_table.loadXML(infile);
		// TCP queue, UDP queue
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_tcp_queue) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
				tcp_queue.loadXML(infile);
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile), 
					xml_syntax_udp_queue) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
				udp_queue.loadXML(infile);
	} while((state >= 0) && (depth > 1));
	updateDisplaylistHost();
	updateDisplaylistPackets();
	// cout << "Host::loadXML end" << endl; 
};

void Host::saveXML(xmlTextWriterPtr outfile) const
/** speichert die Einstellungen des Hosts in der XML-Datei 'outfile',
 *  d.h. Attribut Name, Position und ruft die saveXML() der Netzwerkkarte,
 *  Routentabelle, TCP-/UDP-Schlangen und der Dienste, <br>
 *  XML-Daten der Dienste werden zur Zeit aber nicht geladen, da es keine
 *  Dienste gibt, welche XML-Daten speichern */
{
	list<Service*>::const_iterator	service_entry;

	xmlTextWriterStartElement(outfile, xml_syntax_computer);
	xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_name, 
					"%s", name.c_str());
	pos.saveXML(outfile);
	network_card.saveXML(outfile);
	route_table.saveXML(outfile);
	tcp_queue.saveXML(outfile);
	udp_queue.saveXML(outfile);
	service_entry = services.begin();
	while(service_entry != services.end())
	{
		if (*service_entry) (*service_entry)->saveXML(outfile);
		service_entry++;
	};
	xmlTextWriterEndElement(outfile);
};

void Host::attachCable(Net_Cable *new_cable)
/** verbindet das Kabel 'new_cable' mit dem Host, also seiner Netzwerkkarte;
 *  <br>
 *  es ruft Interface::attachCable(), passt die Kabell&auml;nge am
 *  entsprechenden Ende an, damit es richtig 'andockt' und &uuml;bernimmt den
 *  Rotationswinkel der Karte (die sich zum Kabel gedreht hat) und
 *  zeichnet sich neu */
{
	Point_3D	cable_begin, cable_end;

	network_card.attachCable(new_cable);
	// attach to cable, take angle of card
        // (has turned himself to cable), redraw
	// resizing cable length on begin or end
	if (new_cable)
	{
		new_cable->getBeginEnd(cable_begin, cable_end);
		if (cable_begin.getDistance(pos, false) == 0.0) 
			new_cable->setBeginNodeRadius(
				width * depth * host_radius_factor_for_cable);
		else	new_cable->setEndNodeRadius(
				width * depth * host_radius_factor_for_cable);
	};
	direction_angle = network_card.getAngle();
	updateDisplaylistHost();
	updateDisplaylistPackets();
};

void Host::detachCables(void)
/** l&ouml;st alle Kabelverbindungen zum Host auf, d.h. ruft
 *  Interface::attachCable() mit Parameter NULL f&uuml;r die eine Karte*/
{
	network_card.attachCable(NULL);
};

bool Host::needsTurn(void) const
/** gibt 'true' zur&uuml;ck, wenn die eine Netzwerkkarte sich drehen will,
 *  damit sie die Pakete in korrekter Richtung abschicken bzw. empfangen kann,
 *  <br> Senderichtung ist entgegengesetzt zu Empfangsrichtung */
{
	return network_card.needsTurn();
};

void Host::getName(string &node_name) const
/** gibt den Namen des Hosts, inklusive der IP-Adresse der Netzwerkkarte
 *  in 'node_name' zur&uuml;ck, zur Anzeige im Fenster */
{
	IpAddress	card_address;

	network_card.getIpAddress(card_address);
	node_name = name + " " + card_address.getAddress();
};

bool Host::hasIpAddress(IpAddress &searched_address) const
/** gibt 'true' zur&uuml;ck, wenn die Netzwerkkarte des Hosts die
 *  gesuchte IP-Adresse 'searched_address' hat, sonst 'false' */
{
	IpAddress	card_address;

	network_card.getIpAddress(card_address);
	return (card_address == searched_address);
};

bool Host::hasMacAddress(MacAddress &searched_address) const
/** gibt 'true' zur&uuml;ck, wenn die Netzwerkkarte des Hosts die
 *  gesuchte Ethernet-Adresse 'searched_address' hat, sonst 'false' */
{
	MacAddress	card_address;

	network_card.getMacAddress(card_address);
	return (card_address == searched_address);
};

bool Host::getFirstPacketPos(Point_3D &packet_pos) const
/** gibt die Position des ersten gefundenen Paketes im Host zur&uuml;ck,
 *  Hostposition wird addiert, also in absoluten Koordinaten; <br>
 *  dazu wird die Netzwerkkarte abgefragt, ob sie ein Paket hat, wenn ja
 *  wird die Position des ersten Paketes dieser Karte
 *  zur&uuml;ckgegeben, siehe Interface::getFirstPacketPos(), <br>
 *  wenn kein Interface ein Paket hat, wird in den TCP- und UDP-Schlangen
 *  gefragt, ob dort gerade ein Paket ist; <br>
 *  gibt 'true' zur&uuml;ck, wenn ein Paket gefunden wurde, sonst 'false' */
{
	bool	has_packets;

	has_packets = network_card.getFirstPacketPos(packet_pos);
	if (!has_packets)
		has_packets = tcp_queue.getFirstPacketPos(packet_pos);
	if (!has_packets)
		has_packets = udp_queue.getFirstPacketPos(packet_pos);
	if (has_packets) packet_pos += pos;		// add Host::pos
	return has_packets;
};

void Host::executeNetImportPacket(Packet* cable_packet)
/** importiert ein Paket in den Router, d.h. wird &uuml;ber Routing an das
 *  Interface weitergegeben,
 *  ruft nur Host_ExecuteNet::importPacket() */
{
	Host_ExecuteNet::importPacket(cable_packet);
};

void Host::executeNet(double speed_factor)
/** f&uuml;hrt einen Schritt der Netzwerksimulation im Host aus,
 *  'speed_factor' beeinflusst die Geschwindigkeit der Drehung und des
 *  Pakettransports, <br>
 *  ruft Host_ExecuteNet::executeNet(), &uuml;bernimmt dessen
 *  Rotationswinkel und zeichnet die Displaylisten neu */
{
	double	new_direction;

	new_direction = direction_angle;
	Host_ExecuteNet::executeNet(speed_factor, new_direction);
	if (new_direction != direction_angle)
	{
		direction_angle = new_direction;
		updateDisplaylistHost();
	};
	updateDisplaylistPackets();
};

