/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "windows32.h"

#include "constants.h"
#include "l3_router.h"

/** Name des Teilbaumes vom Router im Info-Variablenbaum zum Dialog */
const char	l3_router_varname_l3_router[] = "router (l3)";
/** Name der Variable zum (DNS-)Namen des Routers */
const char	l3_router_varname_name[] = "hostname";
/** Name des Teilbaumes zu NAT im Info-Variablenbaum zum Dialog */
const char	l3_router_varname_nat[] = "NAT";
/** Name des Teilbaumes einer Netzwerkkarte im Info-Variablenbaum zum Dialog */
const char	l3_router_varname_network_card[] = "network card";
/** Name der Variablen zur Anzahl der Netzwerkkarten im Info-Variablenbaum
 *  zum Dialog */
const char	l3_router_varname_num_cards[] = "number of cards";
/** Name des Teilbaumes der Routen-Tabelle im Info-Variablenbaum zum Dialog */
const char	l3_router_varname_route_table[] = "route table";
/** Gr&ouml;&szlig;e des Routers (L&auml;nge, Breite, H&ouml;he) */
const double	l3_router_size[3] = {2.0, 4.0, 2.0};
/** Farbe des Routers, RGBA 0.0 - 1.0, alpha = 1.0: undurchsichtig */
const double	l3_router_color[4] = {0.9, 0.6, 0.0, 1.0};
/** Faktor, um den sich die Kabel verk&uuml;rzen sollen, damit sie richtig
 *  am Router 'andocken', normalerweise gehen sie genau bis zum Mittelpunkt des
 *  Routers */
const double	l3_router_radius_factor_for_cable = 1 / 4.5;

L3_Router::L3_Router(void) : direction_angle(0.0), 
				last_level_of_details_on(true),
				display_list_l3_router(0),
				display_list_packets(0)
/** Konstruktor, Initialisierung von direction_angle, last_level_of_details_on,
 *  Displaylisten */
{
};

L3_Router::~L3_Router(void)
/** Destruktor: Displaylisten werden gel&ouml;scht */
{
	// cout << "Destruktor L3_Router start" << endl;
	if (glIsList(display_list_l3_router))
		glDeleteLists(display_list_l3_router, 1);
	if (glIsList(display_list_packets))
		glDeleteLists(display_list_packets, 1);
	// cout << "Destruktor L3_Router end" << endl;
};

void L3_Router::draw(void) const
/** Zeichnet den Router, d.h. ruft die Displaylisten auf */
{
	if (glIsList(display_list_l3_router))
		glCallList(display_list_l3_router);
	if (glIsList(display_list_packets)) glCallList(display_list_packets);
};

void L3_Router::drawSetCameraPos(bool level_of_details_on, Point_3D camera_pos)
/** aktualisiert die internen Werte zu 'last_camera_pos' (letzte Kameraposition
 *  relativ zur Routerposition) und 'last_level_of_details_on' (Level of Detail
 *  an oder aus), Displayliste der Pakete inklusive Routentabelle und ARP-Cache
 *  (abh&auml;ngig von Level of Detail) wird neu gezeichnet */
{
	last_camera_pos.setPos(camera_pos.getX() - pos.getX(),
				camera_pos.getY() - pos.getY(),
				camera_pos.getZ() - pos.getZ());
	last_level_of_details_on = level_of_details_on;
	updateDisplaylistPackets();
};

void L3_Router::updateDisplaylistL3Router(void)
/** zeichnet die Displayliste der Routerelemente neu: W&auml;nde,
 *  Transportschienen, Quader der Netzwerkkarten, Boden und 'Dach',
 *  vor allem nach Positions- oder Gr&ouml;&szlig;en&auml;nderung */
{
	size_t	i;

	if (!glIsList(display_list_l3_router)) return;
	glNewList(display_list_l3_router, GL_COMPILE);
	pos.translate(1);
	glScaled(width / 2.0, height / 4.0, depth / 2.0);
	glRotated(direction_angle, 0, 1, 0);

	for(i = 0; i < 2; i++) walls[i].draw();
	for(i = 0; i < 4; i++) poles[i].draw();
	for(i = 0; i < network_cards.size(); i++)
	{
		if (network_cards[i] != NULL)
		{
			network_cards[i]->drawQueue();
		};
	};
	top.draw();
	bottom.draw();

	glRotated(-direction_angle, 0, 1, 0);
	glScaled(2.0 / width, 4.0 / height, 2.0 / depth);
	pos.translate(-1);
	glEndList();
};

void L3_Router::updateDisplaylistPackets(void)
/** zeichnet die Displayliste der Pakete neu, inklusive Routentabelle und
 *  ARP-Cache, d.h. ruft die Zeichenmethoden von Interface und Route_Table
 *  und &uuml;bergibt denen 'last_level_of_details_on' und 'last_camera_pos' */
{
	bool	arp_drawn;
	size_t	i;

	if (!glIsList(display_list_packets)) return;
	glNewList(display_list_packets, GL_COMPILE);
	pos.translate(1);
	glScaled(width / 2.0, height / 4.0, depth / 2.0);
	glRotated(direction_angle, 0, 1, 0);

	arp_drawn = false;
	for(i = 0; i < network_cards.size(); i++)
	{
		if (network_cards[i] != NULL)
		{
			network_cards[i]->drawPackets(last_level_of_details_on,
							last_camera_pos);
			if ((network_cards[i]->getAngle() == direction_angle)
				&& (!arp_drawn))
			{
				network_cards[i]->drawArpCache(
						last_level_of_details_on,
						last_camera_pos);
				arp_drawn = true;
			};
		};
	};
	route_table.draw(last_level_of_details_on, last_camera_pos);

	glRotated(-direction_angle, 0, 1, 0);
	glScaled(2.0 / width, 4.0 / height, 2.0 / depth);
	pos.translate(-1);
	glEndList();
};

void L3_Router::init(void)
/** Initialisierung der Router-Elemente, nach Starten der OpenGL-Umgebung,<br>
 *  Routentabelle, NAT, Anzahl Netzwerkkarten auf 1, Displaylisten,
 *  Farbe setColor(), Position setPos(), Gr&ouml;&szlig;e setSize() */
{
	route_table.init();
	nat_packet_translator.init();
	resizeInterfaceArray(1);

	// graphic
	display_list_l3_router = glGenLists(1);
	display_list_packets = glGenLists(1);
	setColor(l3_router_color[0], l3_router_color[1], l3_router_color[2],
			l3_router_color[3]);
	setPos(0.0, 0.0, 0.0);
	setSize(l3_router_size[0], l3_router_size[1], l3_router_size[2]);
};

void L3_Router::getPos(double &current_centerx, double &current_centery,
			double &current_centerz) const
/** gibt die Position (den Mittelpunkt) des Routers (x, y, z) zur&uuml;ck */
{
	Cuboid::getPos(current_centerx, current_centery, current_centerz);
};

void L3_Router::getPosAngle(Point_3D &current_pos, double &current_angle) const
/** gibt die Position (den Mittelpunkt) in 'current_pos' und den aktuellen
 *  Rotationswinkel in 'current_angle' des Routers zur&uuml;ck */
{
	current_pos = pos;
	current_angle = direction_angle;	
};

void L3_Router::setColor(double new_red,  double new_green, double new_blue,
				double new_alpha)
/** setzt die Farbe des Routers auf 'new_red' (Rot), 'new_green' (Gr&uuml;n),
 *  'new_blue' (Blau) und 'new_alpha' (Transparenz, 1.0 undurchsichtig),
 *  Wertebereich jeweils 0.0 - 1.0 */
{
	size_t	i;

	for(i = 0; i < 2; i++)
		walls[i].setColor(new_red, new_green, new_blue, new_alpha);
	for(i = 0; i < 4; i++)
		poles[i].setColor(new_red, new_green, new_blue, new_alpha);
	top.setColor(new_red, new_green, new_blue, new_alpha);
	bottom.setColor(new_red, new_green, new_blue, new_alpha);
};

void L3_Router::setPos(double new_centerx, double new_centery,
			double new_centerz)
/** setzt die Position (den Mittelpunkt) des Routers (x, y, z) */
{
	Cuboid::setPos(new_centerx, new_centery, new_centerz);
	updateDisplaylistL3Router();
	updateDisplaylistPackets();
};

void L3_Router::setSize(double new_width, double new_height, double new_depth)
/** setzt die Gr&ouml;&szlig;e des Routers (Bounding-Box),
 *  'new_width' = Breite, 'new_height' = H&ouml;he, 'new_depth' = Tiefe, <br>
 *  positioniert die Elemente neu, zeichnet die Displaylisten neu */
{
	Cuboid::setSize(new_width, new_height, new_depth);
	L3_Router::updatePos();
	updateDisplaylistL3Router();
	updateDisplaylistPackets();

};

void L3_Router::updatePos(void)
/** Position und Gr&ouml;&szlig;e der einzelnen grafischen Elemente (inklusive
 *  Routentabelle, ohne Netzwerkkarten) wird neu bestimmt,
 *  unabh&auml;ngig von der Bounding-Box des Routers, da sie beim Zeichnen
 *  nochmal skaliert werden */ 
{
	int	i;

	for(i = 0; i < 2; i++) walls[i].setSize(0.05, 3.8, 1.0);
	for(i = 0; i < 4; i++) poles[i].setSize(0.05, 3.8, 0.05);
	poles[0].setPos(-0.325, 0.0, -0.4);
	poles[1].setPos(-0.325, 0.0, 0.4);
	poles[2].setPos( 0.325, 0.0, 0.4);
	poles[3].setPos( 0.325, 0.0, -0.4);
	walls[0].setPos(-0.75, 0.0, 0.0);
	walls[1].setPos( 0.75, 0.0, 0.0);
	top.setSize(2.0, 0.1, 2.0);
	top.setPos(0.0, 1.95, 0.0);
	bottom.setSize(2.0, 0.1, 2.0);
	bottom.setPos(0.0, -1.95, 0.0);
	route_table.setPos(0.0, 0.25, 0.8);
	route_table.setSize(1.0, 0.5, 0.0);
};

void L3_Router::getInfos(Variable_Tree* &infos)
/** stellt den Variablenbaum f&uuml;r das Dialogfenster zusammen, Zeiger
 *  darauf in 'infos' zur&uuml;ck */
{
	size_t				i;
	char				current_i[10];
	string				current_varname;
	Numeric_Variable<size_t>	*var_num_cards;
	String_Variable			*var_name;
	Variable_Tree			*var_tree_interface, 
					*var_tree_route_table, 
					*var_tree_nat;

	// cout << "L3_Router::getInfos start" << endl;
	current_infos.clear();
	var_name = new String_Variable;
	if (var_name)
	{	// name of the router
		var_name->setName(l3_router_varname_name);
		var_name->setLength(255, 0);
		var_name->setValue(name);
		current_infos.addVar(var_name);
	};
	var_num_cards = new Numeric_Variable<size_t>;
	if (var_num_cards)
	{	// number of network interfaces 
		var_num_cards->setName(l3_router_varname_num_cards);
		var_num_cards->setLength(1, 0);
		var_num_cards->setMaxmin(8, 0);
		var_num_cards->setValue(network_cards.size());
		current_infos.addVar(var_num_cards);
	};
	for(i = 0; i < network_cards.size(); i++)
	{	// the array of network interfaces
		if (network_cards[i] == NULL) continue;
		var_tree_interface = NULL;
		network_cards[i]->getInfos(var_tree_interface);
		if (var_tree_interface)
		{
			snprintf(current_i, sizeof(current_i), "%2d", i + 1);
			current_varname = 
				string(l3_router_varname_network_card);
			current_varname += string(" ") + current_i;
			var_tree_interface->setName(current_varname);
			current_infos.addChild(var_tree_interface);
		};
	};
	// Routing table and NAT
	route_table.getInfos(var_tree_route_table);
	if (var_tree_route_table)
		var_tree_route_table->setName(l3_router_varname_route_table);
	current_infos.addChild(var_tree_route_table);
	nat_packet_translator.getInfos(var_tree_nat);
	if (var_tree_nat)
		var_tree_nat->setName(l3_router_varname_nat);
	current_infos.addChild(var_tree_nat);
	current_infos.setName(l3_router_varname_l3_router);
	infos = &current_infos;
	// cout << "L3_Router::getInfos end" << endl;
};

void L3_Router::resizeInterfaceArray(size_t new_size)
/** vergr&ouml;&szlig;ert/verkleinert das Feld der Netzwerkkarten:
 *  entsprechende Anzahl an Karten wird hinzugef&uuml;gt (mit Position und
 *  Gr&ouml;&szlig;e) oder am Ende beginnend gel&ouml;scht */
{
	size_t		i, max_size_t;
	Interface	*new_network_card;
	
	// cout << "L3_Router::resize start" << endl;
	while (new_size > network_cards.size())
	{
		new_network_card = new Interface;
		if (new_network_card)
		{
			new_network_card->init();
			new_network_card->setPos( 0.0, -0.925, 0.0);
			new_network_card->setSize(1.5,  1.85,  1.0);
			network_cards.insert(network_cards.end(), 
						new_network_card);
		}
		else return;
	};
	if (new_size < network_cards.size())
	{
		max_size_t = 0;
		max_size_t--;
		for(i = network_cards.size() - 1; 
			(i >= new_size) && (i != max_size_t); i--)
		{
			if (network_cards[i] != NULL) delete network_cards[i];
		};
	};
	network_cards.resize(new_size);
	// cout << "resize end" << endl;
};

void L3_Router::setInfos(Variable_Tree* infos)
/** liest die Werte aus dem Variablenbaum ein, der Zeiger 'infos' sollte
 *  auf den internen Baum zeigen, wird aber nicht beachtet */
{
	size_t				i, value;
	char				current_i[10];
	string				value_text, current_varname;
	Variable			*current_var;
	Variable_Tree			*current_tree;
	Numeric_Variable<size_t>	*var_num_cards;

	// cout << "L3_Router::setInfos start" << endl;
	if (current_infos.findVar(l3_router_varname_name, current_var))
	{	// name of the router
		value_text = current_var->getValue();
		name = value_text;
	};
	// cout << "L3_Router::setInfos num_cards" << endl;
	if (current_infos.findVar(l3_router_varname_num_cards, current_var))
	{	// number of network cards
		var_num_cards = reinterpret_cast<Numeric_Variable<size_t>* >
					(current_var);
		var_num_cards->getValue(0, value);
		resizeInterfaceArray(value);
	};
	// cout << "L3_Router::setInfos interfaces" << endl;
	for(i = 0; i < network_cards.size(); i++)
	{	// the values of the network interfaces, read by them */
		if (network_cards[i] == NULL) continue;
		snprintf(current_i, sizeof(current_i), "%2d", i + 1);
		current_varname = string(l3_router_varname_network_card);
		current_varname += string(" ") + current_i;
		// cout << "current_varname: " << current_varname << endl;
		if (current_infos.findChild(current_varname, current_tree))
			network_cards[i]->setInfos(current_tree);
	};
	// cout << "L3_Router::setInfos route table" << endl;
	if (current_infos.findChild(l3_router_varname_route_table, 
					current_tree))
		route_table.setInfos(current_tree);
	// cout << "L3_Router::setInfos nat" << endl;
	if (current_infos.findChild(l3_router_varname_nat, current_tree))
		nat_packet_translator.setInfos(current_tree);
	// cout << "L3_Router::setInfos end" << endl;
};

void L3_Router::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt die Einstellungen des Routers aus der XML-Datei 'infile',
 *  Router-Name, Netzwerkkarten, Routentabelle, NAT */
{
	size_t	last;
	int	state, depth;
	xmlChar	*content;

	resizeInterfaceArray(0);
	if (xmlTextReaderHasAttributes(infile))
	{
		content = xmlTextReaderGetAttribute(infile, xml_syntax_name);
		if (content) name = reinterpret_cast<char*>(content);
	};
	do
	{
		// a (nearly) infinte amount of network cards can be loaded
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if (xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_pos) == 0)
			pos.loadXML(infile);
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_interface) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
		{
				resizeInterfaceArray(network_cards.size() + 1);
				last = network_cards.size() - 1;
				if (network_cards[last])
					network_cards[last]->loadXML(infile);
		}
		// route table and NAT
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_route_table) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
				route_table.loadXML(infile);
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_nat) == 0) &&
			(xmlTextReaderNodeType(infile) == 1))
				nat_packet_translator.loadXML(infile);
	} while ((state >= 0) && (depth > 1));
	updateDisplaylistL3Router();
	updateDisplaylistPackets();
};

void L3_Router::saveXML(xmlTextWriterPtr outfile) const
/** speichert die Einstellungen des Routers in der XML-Datei 'outfile',
 *  d.h. Attribut Name, Position und ruft die saveXML() der Netzwerkkarten,
 *  Routentabelle und von NAT */
{
	size_t	i;

	xmlTextWriterStartElement(outfile, xml_syntax_router);
	xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_name, 
						"%s", name.c_str());
	pos.saveXML(outfile);
	for(i = 0; i < network_cards.size(); i++)
	{
		if (network_cards[i] != NULL) 
			network_cards[i]->saveXML(outfile);
	};
	route_table.saveXML(outfile);
	nat_packet_translator.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
};

void L3_Router::attachCable(Net_Cable* new_cable)
/** verbindet das Kabel 'new_cable' mit dem Router, also einer seiner
 *  Netzwerkkarten; <br>
 *  es passt die Kabell&auml;nge am entsprechenden Ende (je nach MacAdress) an,
 *  damit es richtig 'andockt', sucht die Netzwerkkarte, an die ein Ende des
 *  Kabels geh&ouml;rt und ruft dessen Interface::attachCable(), dann
 *  &uuml;bernimmt er den Rotationswinkel dieser Karte (sie hat sich zum Kabel
 *  gedreht) und zeichnet sich neu */
{
	size_t		i;
	Point_3D	cable_begin, cable_end;
	MacAddress	begin_address, end_address, interface_mac;

	if (new_cable == NULL) return;

	new_cable->getBeginEnd(cable_begin, cable_end);
	// resizing cable length on begin or end 
	if (cable_begin.getDistance(pos, false) == 0.0) 
		new_cable->setBeginNodeRadius(
			width * depth * l3_router_radius_factor_for_cable);
	else	new_cable->setEndNodeRadius(
			width * depth * l3_router_radius_factor_for_cable);
	for(i = 0; i < network_cards.size(); i++)
	{	// find the network interface, to which the cable addresses
		// match
		if (network_cards[i] == NULL) continue;

		network_cards[i]->getMacAddress(interface_mac);
		new_cable->getAddresses(begin_address, end_address);
		if ((begin_address == interface_mac) || 
			(end_address == interface_mac))
			{	// attach to cable, take angle of card
				// (has turned himself to cable), redraw
				network_cards[i]->attachCable(new_cable);
				direction_angle = network_cards[i]->getAngle();
				updateDisplaylistL3Router();
			};
	};
};

void L3_Router::detachCables(void)
/** l&ouml;st alle Kabelverbindungen zum Router auf, d.h. ruft
 *  Interface::attachCable() mit Parameter NULL f&uuml;r alle Karten */
{
	size_t	i;

	for(i = 0; i < network_cards.size(); i++)
	{
		if (network_cards[i] != NULL) 
			network_cards[i]->attachCable(NULL);
	};
};

void L3_Router::getName(string &node_name) const
/** gibt den Namen des Routers, inklusive aller IP-Adressen der Netzwerkkarten
 *  in 'node_name' zur&uuml;ck, zur Anzeige im Fenster */
{
	size_t		i;
	IpAddress	interface_address;

	node_name = "";
	for(i = 0; i < network_cards.size(); i++)
	{
		if (network_cards[i] == NULL) continue;
		network_cards[i]->getIpAddress(interface_address);
		node_name += interface_address.getAddress() + string(" ");
	};
	node_name = name + " " + node_name;
};

bool L3_Router::hasIpAddress(IpAddress &searched_address) const
/** gibt 'true' zur&uuml;ck, wenn eine der Netzwerkkarten des Routers die
 *  gesuchte IP-Adresse 'searched_address' hat, sonst 'false' */
{
	bool		found;
	size_t		i;
	IpAddress	interface_address;

	found = false;
	for(i = 0; i < network_cards.size(); i++)
	{
		if (network_cards[i] == NULL) continue;
		network_cards[i]->getIpAddress(interface_address);
		if (!found) found = (interface_address == searched_address);
	};
	return found;
};

bool L3_Router::hasMacAddress(MacAddress &searched_address) const
/** gibt 'true' zur&uuml;ck, wenn eine der Netzwerkkarten des Routers die
 *  gesuchte Ethernet-Adresse 'searched_address' hat, sonst 'false' */
{
	bool		found;
	size_t		i;
	MacAddress	interface_mac;

	found = false;
	for(i = 0; i < network_cards.size(); i++)
	{
		if (network_cards[i] == NULL) continue;
		network_cards[i]->getMacAddress(interface_mac);
		if (!found) found = (interface_mac == searched_address);
	};
	return found;
};

bool L3_Router::needsTurn(void) const
/** gibt 'true' zur&uuml;ck, wenn die aktuell bevorzugte Netzwerkkarte
 *  (gleicher Rotationswinkel wie der eigene) sich drehen will,
 *  damit sie die Pakete in korrekter Richtung abschicken bzw. empfangen kann,
 *  <br> Senderichtung ist entgegengesetzt zu Empfangsrichtung */
{
	bool		needs_turn;
	size_t		i;

	needs_turn = false;
	for(i = 0; i < network_cards.size(); i++)
	{
		if ((network_cards[i] != NULL) && 
			(network_cards[i]->getAngle() == direction_angle))
		{
			needs_turn = network_cards[i]->needsTurn();
			i = network_cards.size(); 	// exit for-loop
		};
	};
	return needs_turn;
};

bool L3_Router::getFirstPacketPos(Point_3D &packet_pos) const
/** gibt die Position des ersten gefundenen Paketes im Router zur&uuml;ck,
 *  Routerposition wird addiert, also in absoluten Koordinaten; <br>
 *  dazu wird jede Netzwerkkarte abgefragt, ob sie ein Paket hat, wenn ja
 *  wird die Position des ersten Paketes dieser (ersten gefundenen) Karte
 *  zur&uuml;ckgegeben, siehe Interface::getFirstPacketPos(), <br>
 *  wenn kein Interface ein Paket hat, wird in der die Routentabelle gefragt,
 *  ob es gerade ein Paket verarbeitet; <br>
 *  gibt 'true' zur&uuml;ck, wenn ein Paket gefunden wurde, sonst 'false' */
{
	bool		has_packets;
	size_t		i;
	Point_3D	new_pos;

	has_packets = false;
	for(i = 0; i < network_cards.size(); i++)
	{	// the first interface, which has at least one packet defines
		// the packet_pos
		if ((network_cards[i] != NULL) && (!has_packets))
			has_packets = network_cards[i]->getFirstPacketPos(
								packet_pos);
	};
	// else ask Route_Table if it has a packet for which a route is searched
	if (!has_packets)
		if (has_packets = route_table.getPacketPos(new_pos))
			packet_pos = new_pos;
	if (has_packets) packet_pos += pos;	// add L3_Router::pos
	return has_packets;	
};

void L3_Router::executeNetImportPacket(Packet* cable_packet)
/** importiert ein Paket in den Router, d.h. wird &uuml;ber Routing an das
 *  passende Interface weitergegeben,
 *  ruft nur L3_Router_ExecuteNet::importPacket() */ 
{
	L3_Router_ExecuteNet::importPacket(cable_packet);
};

void L3_Router::executeNet(double speed_factor)
/** f&uuml;hrt einen Schritt der Netzwerksimulation im Router aus,
 *  'speed_factor' beeinflusst die Geschwindigkeit der Drehung und des
 *  Pakettransports, <br>
 *  ruft L3_Router_ExecuteNet::executeNet(), &uuml;bernimmt dessen
 *  Rotationswinkel und zeichnet die Displaylisten neu */
{
	double	new_direction;

	new_direction = direction_angle;
	L3_Router_ExecuteNet::executeNet(speed_factor, new_direction);
	if (new_direction != direction_angle)
	{
		direction_angle = new_direction;
		updateDisplaylistL3Router();
	};
	// cout << "direction angle: " << direction_angle << endl;
	updateDisplaylistPackets();
};

