/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "arp_packet.h"
#include "constants.h"

/** Text der Paketbeschriftung bei Typ Request */
const char	text_arp_request[] = "ARP Request";
/** Text der Paketbeschriftung bei Typ Reply */
const char	text_arp_reply[] = "ARP Reply";
/** Farbe des ARP-Paketes, RGBA 0.0 - 1.0 */
const double	arp_packet_color[4] = {0.3, 0.0, 1.0, 1.0};

ARP_Packet::ARP_Packet(void)
/** Konstruktor, Initialisierung: ETHERTYPE_ARP, setProtocol(), setColor() */
{
	eth_header.ether_type = htons(ETHERTYPE_ARP);
	setProtocol(2, ARP);
	setColor(arp_packet_color[0], arp_packet_color[1], arp_packet_color[2],
			arp_packet_color[3]);
};

ARP_Packet::~ARP_Packet(void)
/** Destruktor, ist leer */
{
	// cout << "Destruktor ~ARP_Packet" << endl;
};

void ARP_Packet::copy(Packet* &copy_of_packet)
/** kopiert den eigenen Paketkopf in den von copy_of_packet;
 *  copy_of_packet wird neu angelegt, wenn es ein NULL-Zeiger ist;
 *  sollte als Objekt der Klasse ARP_Packet &uuml;bergeben werden, falls
 *  nicht NULL */
{
	IpAddress	ip_address;
	ARP_Packet*	arp_packet;

	if (copy_of_packet == NULL) copy_of_packet = new ARP_Packet;
	arp_packet = reinterpret_cast<ARP_Packet*>(copy_of_packet);
	memcpy(&(arp_packet->arp_header), &arp_header, sizeof(arp_header));
	Ethernet_Packet::copy(copy_of_packet);
	cout << "ARP_Packet::copy" << endl;
	cout << "Original: ";
	getSourceIpAddress(ip_address);
	cout << "Source: " << ip_address << " ";
	getDestinationIpAddress(ip_address);
	cout << "Dest.: " << ip_address << " "
		<< "Op (1 request, 2 reply): "
		<< ntohs(arp_header.ea_hdr.ar_op) << endl;
	cout << "Copy: ";
	arp_packet->getSourceIpAddress(ip_address);
	cout << "Source: " << ip_address << " ";
	arp_packet->getDestinationIpAddress(ip_address);
	cout << "Dest.: " << ip_address << " "
		<< "Op: "
		<< ntohs(arp_packet->arp_header.ea_hdr.ar_op) << endl;
};

void ARP_Packet::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt Paketkopf aus XML-Datei 'infile', ruft Ethernet_Packet::loadXML()
 */
{
	int		state, depth, return_depth;
	char		*value_text, *end_ptr;
	xmlChar		*content;
	IpAddress	source_ip, destination_ip;

	return_depth = xmlTextReaderDepth(infile);
	if (xmlTextReaderHasAttributes(infile))
	{
		content = xmlTextReaderGetAttribute(infile,
						xml_syntax_arp_packet_opcode);
		value_text = NULL;
		if (content) value_text = reinterpret_cast<char*>(content);
		end_ptr = value_text;
		if (value_text)
			arp_header.ea_hdr.ar_op = htons(strtoul(value_text,
								&end_ptr, 10));
	};
	do
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if (xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_source_ip) == 0)
			source_ip.loadXML(infile);
		else if (xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_destination_ip) == 0)
			destination_ip.loadXML(infile);
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_ethernet_packet) == 0) &&
				(xmlTextReaderNodeType(infile) == 1))
			Ethernet_Packet::loadXML(infile);
	} while ((state >= 0) && (depth > return_depth));
	if (isRequest()) requestIp(source_ip, destination_ip);
	if (isReply()) replyIp(source_ip, destination_ip);
};

void ARP_Packet::saveXML(xmlTextWriterPtr outfile) const
/** speichert Paketkopf in XML-Datei 'outfile', ruft Ethernet_Packet::saveXML()
 */
{
	IpAddress	address;

	xmlTextWriterStartElement(outfile, xml_syntax_arp_packet);
	xmlTextWriterWriteFormatAttribute(outfile,
						xml_syntax_arp_packet_opcode,
						"%d",
						ntohs(arp_header.ea_hdr.ar_op)
						);
	xmlTextWriterStartElement(outfile, xml_syntax_source_ip);
	getSourceIpAddress(address);
	address.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	xmlTextWriterStartElement(outfile, xml_syntax_destination_ip);
	getDestinationIpAddress(address);
	address.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	Ethernet_Packet::saveXML(outfile);
	xmlTextWriterEndElement(outfile);
};

bool ARP_Packet::isRequest(void) const
/** gibt 'true' zur&uuml;ck, wenn Paket laut ARP-Opcode ein Request (Anfrage)
 *  ist, sonst 'false' */
{
	return (ntohs(arp_header.ea_hdr.ar_op) == ARPOP_REQUEST);
};

bool ARP_Packet::isReply(void) const
/** gibt 'true' zur&uuml;ck, wenn Paket laut ARP-Opcode ein Reply (Antwort)
 *  ist, sonst 'false' */
{
	return (ntohs(arp_header.ea_hdr.ar_op) == ARPOP_REPLY);
};

void ARP_Packet::updatePacketText(void)
/** aktualisiert die dargestellte Paketbeschriftung im 3D-Fenster, aus
 *  den derzeitigen Daten im Paketkopf */
{
	IpAddress	source, destination;

	getSourceIpAddress(source);
	getDestinationIpAddress(destination);
	setTextLine(2, destination.getAddress());
	setTextLine(3, source.getAddress());
	if (isRequest()) setTextLine(4, text_arp_request);
	if (isReply()) setTextLine(4, text_arp_reply);
};

void ARP_Packet::importHeader(struct ether_arp *new_arp_header)
/** kopiert den Paketkopf von Zeiger 'new_arp_header' in den intern
 *  gespeicherten Kopf, Beschriftung wird aktualisiert */
{
	if (new_arp_header)
	{
		memcpy(&arp_header, new_arp_header, sizeof(struct ether_arp));
		updatePacketText();
	};
};

void ARP_Packet::exportHeader(struct ether_arp *current_arp_header) const
/** kopiert den internen Kopf in den Speicherbereich, auf den
 *  'current_arp_header' zeigt */
{
	if (current_arp_header)
	{
		memcpy(current_arp_header, &arp_header,
			sizeof(struct ether_arp));
	};
};

bool ARP_Packet::setDestinationMacAddress(const char mac_string[])
/** setzt die Ethernet-Zieladresse des Paketes aus dem Text 'mac_string', auch
 *  im ARP-Kopf,
 *  bei Request Broadcast-Adresse, bei Reply die Adresse des Anfragenden */
{
	bool	no_error;

	no_error = Ethernet_Packet::setDestinationMacAddress(mac_string);
	if (no_error) 
		memcpy(arp_header.arp_tha, eth_header.ether_dhost, ETH_ALEN);
	updatePacketText();
	return no_error;
};

bool ARP_Packet::setSourceMacAddress(const char mac_string[])
/** setzt die Ethernet-Quelladresse des Paketes aus dem Text 'mac_string',
 *  auch im ARP-Kopf,
 *  immer die Adresse, von der das Paket geschickt wurde */
{
	bool	no_error;

	no_error = Ethernet_Packet::setSourceMacAddress(mac_string);
	if (no_error)
		memcpy(arp_header.arp_sha, eth_header.ether_shost, ETH_ALEN);
	updatePacketText();
	return no_error;
};

void ARP_Packet::setDestinationMacAddress(MacAddress mac_address)
/** &uuml;bernimmt die Adresse aus dem Objekt 'mac_address' als
 *  Ethernet-Zieladresse, auch im ARP-Kopf, ist der Methode mit mac_string
 *  vorzuziehen wenn m&ouml;glich; <br>
 *  bei Request Broadcast-Adresse, bei Reply die Adresse des Anfragenden */
{
	struct ether_addr	ether_address;
	
	mac_address.getAddress(ether_address);
	Ethernet_Packet::setDestinationMacAddress(mac_address);
	memcpy(arp_header.arp_tha, eth_header.ether_dhost, ETH_ALEN);
	updatePacketText();
};

void ARP_Packet::setSourceMacAddress(MacAddress mac_address)
/** &uuml;bernimmt die Adresse aus dem Objekt 'mac_address' als
 *  Ethernet-Quelladresse, auch im ARP-Kopf, ist der Methode mit mac_string
 *  vorzuziehen wenn m&ouml;glich; <br>
 *  immer die Adresse, von der das Paket geschickt wurde */
{
	struct ether_addr	ether_address;
	
	mac_address.getAddress(ether_address);
	Ethernet_Packet::setSourceMacAddress(mac_address);
	memcpy(arp_header.arp_sha, eth_header.ether_shost, ETH_ALEN);
	updatePacketText();
};

void ARP_Packet::getSourceIpAddress(IpAddress &address) const
/** gibt die Quell-IP-Adresse in 'address' zur&uuml;ck, 
 *  immer die Adresse, von der das Paket geschickt wurde */
{
	struct in_addr	source_inaddress;

	memcpy(&source_inaddress, arp_header.arp_spa, sizeof(struct in_addr));
	address.setAddress(source_inaddress);
};

void ARP_Packet::getDestinationIpAddress(IpAddress &address) const
/** gibt die Ziel-IP-Adresse in 'address' zur&uuml;ck,
 *  bei Request die gesuchte, bei Reply die Adresse des Fragenden */
{
	struct in_addr	destination_inaddress;

	memcpy(&destination_inaddress, arp_header.arp_tpa,
		sizeof(struct in_addr));
	address.setAddress(destination_inaddress);
	// cout << arp_header.arp_tpa << " " 
	//	<< destination_inaddress.s_addr << " " << address << endl;
};

void ARP_Packet::requestIp(IpAddress my_ip, IpAddress searched_ip)
/** stellt das Anfrage-Paket zusammen, suche nach 'searched_ip', geschickt
 *  von Quelladresse 'my_ip'; Ethernet-Quelladresse wird in EthernetQueue
 *  gesetzt, kann aber auch vorher gesetzt werden */
{
	struct in_addr	source_inaddress, destination_inaddress;
	IpAddress	source, destination;

	// cout << "ARP_Packet::requestIp " << my_ip << " ";
	// cout << searched_ip << endl;
	arp_header.ea_hdr.ar_hrd = htons(ARPHRD_ETHER);
	arp_header.ea_hdr.ar_pro = htons(ETHERTYPE_IP);
	arp_header.ea_hdr.ar_hln = ETH_ALEN;
	arp_header.ea_hdr.ar_pln = sizeof(in_addr_t);
	arp_header.ea_hdr.ar_op = htons(ARPOP_REQUEST);
	setDestinationMacAddress("ff:ff:ff:ff:ff:ff");
	source = my_ip;
	destination = searched_ip;
	// cout << "Destination: " << destination << " ";
	source.getAddress(source_inaddress);
	memcpy(arp_header.arp_spa, &source_inaddress, sizeof(struct in_addr));
	destination.getAddress(destination_inaddress);
	memcpy(arp_header.arp_tpa, &destination_inaddress, 
		sizeof(struct in_addr));
	// cout << destination_inaddress.s_addr << endl;
	updatePacketText();
};

void ARP_Packet::replyIp(IpAddress my_ip, IpAddress destination_ip)
/** stellt das Antwort-Paket zusammen, geschickt von 'my_ip' (die gesuchte
 *  in der Anfrage), geht zu 'destination_ip' (der Anfragende),
 *  Ethernet-Adressen (Quelle und Ziel) sind selbst zu setzen */
{
	struct in_addr	source_inaddress, destination_inaddress;
	IpAddress	source, destination;

	arp_header.ea_hdr.ar_hrd = htons(ARPHRD_ETHER);
	arp_header.ea_hdr.ar_pro = htons(ETHERTYPE_IP);
	arp_header.ea_hdr.ar_hln = ETH_ALEN;
	arp_header.ea_hdr.ar_pln = sizeof(in_addr_t);
	arp_header.ea_hdr.ar_op = htons(ARPOP_REPLY);
	source = my_ip;
	destination = destination_ip;
	source.getAddress(source_inaddress);
	memcpy(arp_header.arp_spa, &source_inaddress, sizeof(struct in_addr));
	destination.getAddress(destination_inaddress);
	memcpy(arp_header.arp_tpa, &destination_inaddress, 
		sizeof(struct in_addr));
	updatePacketText();
};

