/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "checksums.h"
#include "constants.h"
#include "ipv4_packet.h"

/** der initiale Wert des TTL-Feldes (time to live) */
const u_int8_t	ipv4_packet_standard_ttl = 15;

IPv4_Packet::IPv4_Packet(void) : next_hop_address(INADDR_ANY),
				route_to_destination(NULL)
/** Konstruktor, Initialisierung: next_hop_address, route_to_destination,
 *  setProtocol(), initPacketHeaders(), setSourceIpAddress(),
 *  setDestinationIpAddress() */
{
	setProtocol(2, IPv4);
	initPacketHeaders();
	setSourceIpAddress(IpAddress(INADDR_ANY));
	setDestinationIpAddress(IpAddress(INADDR_ANY));
};

IPv4_Packet::~IPv4_Packet(void)
/** Destruktor, ist leer */
{
};

void IPv4_Packet::copy(Packet* &copy_of_packet)
/** kopiert den eigenen Paketkopf in den von copy_of_packet;
 *  copy_of_packet wird neu angelegt, wenn es ein NULL-Zeiger ist;
 *  sollte als Objekt der Klasse IPv4_Packet &uuml;bergeben werden, falls
 *  nicht NULL */
{
	IPv4_Packet	*ip_packet;

	if (copy_of_packet == NULL) copy_of_packet = new IPv4_Packet;
	ip_packet = reinterpret_cast<IPv4_Packet*>(copy_of_packet);
	memcpy(&(ip_packet->ipv4_header), &ipv4_header, sizeof(ipv4_header));
	Ethernet_Packet::copy(copy_of_packet);
	cout << "IPv4::copy" << endl;
};

void IPv4_Packet::initPacketHeaders(void)
/** setzt die Felder des Paketkopfes auf initiale Werte, setzt ETHERTYPE_IP
 *  im Ethernet-Kopf */
{
	eth_header.ether_type = htons(ETHERTYPE_IP);
	ipv4_header.ip_v = IPVERSION;
	ipv4_header.ip_hl = 5;
	ipv4_header.ip_tos = 0;
	ipv4_header.ip_len = htons(sizeof(ipv4_header));
	ipv4_header.ip_id = htons(0);
	ipv4_header.ip_off = 0;
	ipv4_header.ip_ttl = ipv4_packet_standard_ttl;
	ipv4_header.ip_p = IPPROTO_IP;
	updateIpChecksum();
};

void IPv4_Packet::updateIpChecksum(void)
/** berechnet die Pr&uuml;fsumme des Kopfes neu, ruft nur
 *  Checksums::updateIpChecksum(), passiert bei jeder &Auml;nderung im Kopf */
{
	Checksums::updateIpChecksum(&ipv4_header);
};

void IPv4_Packet::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt Paketkopf aus XML-Datei 'infile', ruft Ethernet_Packet::loadXML()
 */
{
	int		state, depth, return_depth;
	IpAddress	address;

	return_depth = xmlTextReaderDepth(infile);
	do
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if (xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_source_ip) == 0)
		{
			address.loadXML(infile);
			setSourceIpAddress(address);
		}
		else if (xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_destination_ip) == 0)
		{
			address.loadXML(infile);
			setDestinationIpAddress(address);
		}
		else if (xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_next_hop_ip) == 0)
		{
			address.loadXML(infile);
			next_hop_address = address;
		}
		else if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_ethernet_packet) == 0) &&
				(xmlTextReaderNodeType(infile) == 1))
			Ethernet_Packet::loadXML(infile);
	} while ((state >= 0) && (depth > return_depth));
};

void IPv4_Packet::saveXML(xmlTextWriterPtr outfile) const
/** speichert Paketkopf in XML-Datei 'outfile', ruft Ethernet_Packet::saveXML()
 */
{
	IpAddress	address;

	xmlTextWriterStartElement(outfile, xml_syntax_ipv4_packet);
	xmlTextWriterStartElement(outfile, xml_syntax_source_ip);
	getSourceIpAddress(address);
	address.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	xmlTextWriterStartElement(outfile, xml_syntax_destination_ip);
	getDestinationIpAddress(address);
	address.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	xmlTextWriterStartElement(outfile, xml_syntax_next_hop_ip);
	getNextHopAddress(address);
	address.saveXML(outfile);
	xmlTextWriterEndElement(outfile);
	Ethernet_Packet::saveXML(outfile);
	xmlTextWriterEndElement(outfile);
};

unsigned short IPv4_Packet::getSourcePort(void) const
/** Quellport, gibt 0 zur&uuml;ck, benutzt von NAT */
{
	return 0;
};

unsigned short IPv4_Packet::getDestinationPort(void) const
/** Zielport, gibt 0 zur&uuml;ck, benutzt von NAT */
{
	return 0;
};

void IPv4_Packet::setSourcePort(unsigned short source_port)
/** Quellport, es wird nichts ver&auml;ndert, benutzt von NAT */
{
	// do nothing
};

void IPv4_Packet::setDestinationPort(unsigned short dest_port)
/** Zielport, es wird nichts ver&auml;ndert, benutzt von NAT */
{
	// do nothing
};

void IPv4_Packet::updatePacketText(void)
/** aktualisiert die dargestellte Paketbeschriftung im 3D-Fenster, aus
 *  den derzeitigen Daten im Paketkopf */
{
	IpAddress	address;

	getDestinationIpAddress(address);
	setTextLine(2, address.getAddress());
};

void IPv4_Packet::importHeader(struct ip *new_ipv4_header)
/** kopiert den Paketkopf von Zeiger 'new_ipv4_header' in den intern
 *  gespeicherten Kopf, Beschriftung wird aktualisiert */
{
	if (new_ipv4_header)
	{
		memcpy(&ipv4_header, new_ipv4_header, sizeof(struct ip));
		updatePacketText();
	};
};

void IPv4_Packet::exportHeader(struct ip *current_ipv4_header) const
/** kopiert den internen Kopf in den Speicherbereich, auf den
 *  'current_ipv4_header' zeigt */
{
	if (current_ipv4_header)
		memcpy(current_ipv4_header, &ipv4_header, sizeof(struct ip));
};

bool IPv4_Packet::setDestinationIpAddress(const char *ip_string)
/** setzt die IP-Zieladresse anhand von ip_string, gibt bei Fehler 'true'
 *  zur&uuml;ck (alte Adresse bleibt dann stehen), sonst 'false' */
{
	bool		error;
	IpAddress	new_destination;

	error = false;
	if (!new_destination.setAddress(ip_string)) error = true;
	else
	{
		new_destination.getAddress(ipv4_header.ip_dst);
		updateIpChecksum();
		updatePacketText();
	};
	return error;
};

bool IPv4_Packet::setSourceIpAddress(const char *ip_string)
/** setzt die IP-Quelladresse anhand von ip_string, gibt bei Fehler 'true'
 *  zur&uuml;ck (alte Adresse bleibt dann stehen), sonst 'false' */
{
	bool		error;
	IpAddress	new_source;

	error = false;
	if (!new_source.setAddress(ip_string)) error = true;
	else
	{
		new_source.getAddress(ipv4_header.ip_src);
		updateIpChecksum();
		updatePacketText();
	};
	return error;
};

void IPv4_Packet::setDestinationIpAddress(IpAddress address)
/** setzt die IP-Zieladresse anhand von 'address' */
{
	address.getAddress(ipv4_header.ip_dst);
	updateIpChecksum();
	updatePacketText();
};

void IPv4_Packet::setSourceIpAddress(IpAddress address)
/** setzt die IP-Quelladresse anhand von 'address' */
{
	address.getAddress(ipv4_header.ip_src);
	updateIpChecksum();
	updatePacketText();
	// cout << "IPv4_Packet::setSourceIpAddress: " << address << endl;
};

void IPv4_Packet::getDestinationIpAddress(IpAddress &address) const
/** gibt die derzeitige IP-Zieladresse in 'address' zur&uuml;ck */
{
	address.setAddress(ipv4_header.ip_dst);
};

void IPv4_Packet::getSourceIpAddress(IpAddress &address) const
/** gibt die derzeitige IP-Quelladresse in 'address' zur&uuml;ck */
{
	address.setAddress(ipv4_header.ip_src);
};

u_int8_t IPv4_Packet::getTTL(void) const
/** gibt den Wert des TTL-Feldes (time to live) im Kopf zur&uuml;ck */
{
	return ipv4_header.ip_ttl;
};

void IPv4_Packet::setTTL(u_int8_t new_ttl)
/** setzt das TTL-Feld (time to live) im Kopf auf 'new_ttl' */
{
	ipv4_header.ip_ttl = new_ttl;
	updateIpChecksum();
};

void IPv4_Packet::decreaseTTL(void)
/** vermindert den Wert des TTL-Feldes in Kopf um '1' */
{
	ipv4_header.ip_ttl--;
	updateIpChecksum();
};

void IPv4_Packet::setRoute(Route *new_route_to_destination)
/** setzt den internen Zeiger auf eine Route auf 'new_route_to_destination',
 *  die Gatewayadresse darin wird als 'next_hop_address' verwendet, wenn nicht
 *  '0.0.0.0'; bei &Uuml;bergabe eines NULL-Zeigers ist die Zieladresse des
 *  Paketes die 'next_hop_address' */
{
	IpAddress	gateway_address;

	route_to_destination = new_route_to_destination;
	if (route_to_destination)
	{
		route_to_destination->getGatewayAddress(gateway_address);
		if (!(gateway_address == IpAddress(INADDR_ANY)))
			next_hop_address = gateway_address;
	}
	else next_hop_address = IpAddress(INADDR_ANY);
	updatePacketText();
};

void IPv4_Packet::getNextHopAddress(IpAddress &hop_address) const
/** gibt die 'next_hop_address' in 'hop_address' zur&uuml;ck, benutzt f&uuml;r
 *  die Adressaufl&ouml;sung  */
{
	if (next_hop_address == IpAddress(INADDR_ANY))
		hop_address.setAddress(ipv4_header.ip_dst);
	else	hop_address = next_hop_address;
	
};

