/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include "checksums.h"
#include "constants.h"
#include "stringstream.h"
#include "udp_packet.h"

/** Text der Beschriftung des Paketes im 3D-Fenster */
const char	udp_packet_text[] = " UDP";

UDP_Packet::UDP_Packet(void)
/** Konstruktor, Initialisierung: setProtocol(), initPacketHeaders(),
 *  setSourcePort(), setDestinationPort() */
{
	setProtocol(3, UDP);
	initPacketHeaders();
	setSourcePort(0);
	setDestinationPort(0);
};

UDP_Packet::~UDP_Packet(void)
/** Destruktor, ist leer */
{
};

void UDP_Packet::copy(Packet* &copy_of_packet)
/** kopiert den eigenen Paketkopf in den von copy_of_packet;
 *  copy_of_packet wird neu angelegt, wenn es ein NULL-Zeiger ist;
 *  sollte als Objekt der Klasse UDP_Packet &uuml;bergeben werden, falls
 *  nicht NULL */
{
	UDP_Packet*	udp_packet;

	if (copy_of_packet == NULL) copy_of_packet = new UDP_Packet;
	udp_packet = reinterpret_cast<UDP_Packet*>(copy_of_packet);
	memcpy(&(udp_packet->udp_header), &udp_header, sizeof(udp_header));
	IPv4_Packet::copy(copy_of_packet);
	cout << "UDP_Packet::copy" << endl;
};

void UDP_Packet::initPacketHeaders(void)
/** setzt die Felder des Paketkopfes auf initiale Werte, Pr&uuml;fsumme
 *  wird neu berechnet, Quellport/Zielport wird nicht ver&auml;ndert;
 *  zus&auml;tzlich im IP-Kopf: setzt IPPROTO_UDP, Paketl&auml;nge und
 *  Pr&uuml;fsumme */
{
	ipv4_header.ip_p = IPPROTO_UDP;
	ipv4_header.ip_len = htons(sizeof(ipv4_header) + sizeof(udp_header));
	updateIpChecksum();
	udp_header.len = sizeof(udp_header);
	updateUdpChecksum();
};

void UDP_Packet::updateUdpChecksum(void)
/** berechnet die Pr&uuml;fsumme des Kopfes neu, ruft nur
 *  Checksums::updateUdpChecksum(), passiert bei jeder &Auml;nderung im Kopf */
{
	Checksums::updateUdpChecksum(&ipv4_header, &udp_header);
};

void UDP_Packet::loadXML(xmlTextReaderPtr infile)
/** l&auml;dt Paketkopf (nur Portnummern) aus XML-Datei 'infile',
 *  ruft IPv4_Packet::loadXML() */
{
	int	state, depth, return_depth;
	char	*value_text, *end_ptr;
	xmlChar	*content;

	return_depth = xmlTextReaderDepth(infile);
	if (xmlTextReaderHasAttributes(infile))
	{	// attributes source_port, destination_port
		content = xmlTextReaderGetAttribute(infile,
						xml_syntax_source_port);
		value_text = 0;
		if (content) value_text = reinterpret_cast<char*>(content);
		end_ptr = value_text;
		if (value_text)
			setSourcePort(strtoul(value_text, &end_ptr, 10));
		content = xmlTextReaderGetAttribute(infile,
						xml_syntax_destination_port);
		value_text = 0;
		if (content) value_text = reinterpret_cast<char*>(content);
		end_ptr = value_text;
		if (value_text)
			setDestinationPort(strtoul(value_text, &end_ptr, 10));
	};
	do
	{
		state = xmlTextReaderRead(infile);
		depth = xmlTextReaderDepth(infile);
		if ((xmlStrcasecmp( xmlTextReaderConstName(infile),
					xml_syntax_ipv4_packet) == 0) &&
				(xmlTextReaderNodeType(infile) == 1))
			IPv4_Packet::loadXML(infile);
	} while ((state >= 0) && (depth > return_depth));
};

void UDP_Packet::saveXML(xmlTextWriterPtr outfile) const
/** speichert Paketkopf (nur Portnummern) in XML-Datei 'outfile',
 *  ruft IPv4_Packet::saveXML() */
{
	xmlTextWriterStartElement(outfile, xml_syntax_udp_packet);
	xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_source_port,
						"%d", getSourcePort());
	xmlTextWriterWriteFormatAttribute(outfile, xml_syntax_destination_port,
						"%d", getDestinationPort());
	IPv4_Packet::saveXML(outfile);
	xmlTextWriterEndElement(outfile);
};

void UDP_Packet::updatePacketText(void)
/** aktualisiert die dargestellte Paketbeschriftung im 3D-Fenster, aus
 *  den derzeitigen Daten im Paketkopf */
{
	usable_stringstream	value_buffer;
	string			value_text, old_text_line, new_text_line;

	value_buffer << getSourcePort();
	value_buffer >> value_text;
	new_text_line = "S: " + value_text;
	value_buffer.clear();
	value_buffer << getDestinationPort();
	value_buffer >> value_text;
	new_text_line += " D: " + value_text;
	setTextLine(3, new_text_line);
	old_text_line = getTextLine(4);
	if (old_text_line.find(udp_packet_text) > old_text_line.size())
		// not found
		setTextLine(4, old_text_line + udp_packet_text);
};

void UDP_Packet::importHeader(struct udphdr *new_udp_header)
/** kopiert den Paketkopf von Zeiger 'new_udp_header' in den intern
 *  gespeicherten Kopf, Beschriftung wird aktualisiert */
{
	if (new_udp_header)
	{
		memcpy(&udp_header, new_udp_header, sizeof(struct udphdr));
		updatePacketText();
	};
};

void UDP_Packet::exportHeader(struct udphdr *current_udp_header) const
/** kopiert den internen Kopf in den Speicherbereich, auf den
 *  'current_udp_header' zeigt */
{
	if (current_udp_header)
		memcpy(current_udp_header, &udp_header, sizeof(struct udphdr));
};

unsigned short UDP_Packet::getDestinationPort(void) const
/** gibt den aktuellen Zielport zur&uuml;ck (in host byte order) */
{
	return ntohs(udp_header.dest);
};

unsigned short UDP_Packet::getSourcePort(void) const
/** gibt den aktuellen Quellport zur&uuml;ck (in host byte order) */
{
	return ntohs(udp_header.source);
};

void UDP_Packet::setDestinationPort(unsigned short dest_port)
/** setzt den Zielport auf den Wert von dest_port (sollte in host byte order
 *  sein) */
{
	udp_header.dest = htons(dest_port);
	updateUdpChecksum();
	updatePacketText();
};

void UDP_Packet::setSourcePort(unsigned short source_port)
/** setzt den Quellport auf den Wert von source_port (sollte in host byte order
 *  sein) */
{
	udp_header.source = htons(source_port);
	updateUdpChecksum();
	updatePacketText();
};

