/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <math.h>

#include "arp_packet.h"
#include "constants.h"
#include "net_queue.h"
#include "tcp_service_packet.h"
#include "udp_service_packet.h"

/** voreingestellter Geschwindigkeitsfaktor */
const double	net_queue_packet_speed_factor = 0.150;
/** Position auf der X-Achse, auf der Paketk&ouml;pfe nicht sichtbar sind,
 *  sollte au&szlig;erhalb der sichtbaren 3D-Welt liegen */
const double	net_queue_packet_x_unvisible_header = 10000.0;
/** voreingestellte Farbe der dargestellten Quader der Schlange */
const double	net_queue_color[4] = {0.0, 1.0, 1.0, 1.0};
/** voreingestellte Textgr&ouml;&szlig;e der Beschriftung */
const double	net_queue_text_size[2] = {0.1, 0.1};

Net_Queue::Net_Queue(void) :
/** Konstruktor, Initialisierung der Variablen, ruft setColor() */
	animation_speed_factor(net_queue_packet_speed_factor), label_text("")
{
	setColor(net_queue_color[0], net_queue_color[1], net_queue_color[2],
			net_queue_color[3]);
};

Net_Queue::~Net_Queue(void)
/** Destruktor, ist leer */
{
};

void Net_Queue::init(void)
/** Initialisierung, ruft nur Initialisierung der Beschriftung */
{
	label.init();
};

void Net_Queue::setAnimationSpeed(double speed_factor)
/** setzt den Geschwindigkeitsfaktor f&uuml;r Pakete etc. auf 'speed_factor'
 */
{
	animation_speed_factor = speed_factor * net_queue_packet_speed_factor;
};

void Net_Queue::setColor(double new_red, double new_green,
				double new_blue, double new_alpha)
/** setzt die Farbe f&uuml;r die sichtbaren Quader und die Beschriftung,
 *  Wertebereich 0.0 - 1.0, Beschriftung bekommt Kontrastfarbe (1 - new_red,
 *  1 - new_green, 1 - new_blue, alpha) */
{
	left.setColor(new_red, new_green, new_blue, new_alpha);
	right.setColor(new_red, new_green, new_blue, new_alpha);
	label.setColor(1 - new_red, 1 - new_green, 1 - new_blue, new_alpha);
};

void Net_Queue::updatePos(void)
/** berechnet die Position und Gr&ouml;&szlig;e der Quader und Beschriftung
 *  neu, gerufen von setSize() der abgeleiteten Klassen */ 
{
	left.setSize(9 * width / 32.0, height / 1.5, depth * 0.95);
	right.setSize(9 * width / 32.0, height / 1.5, depth * 0.95);
	left.setPos(-23 * width / 64.0, -height / 6.0, 0.0);
	right.setPos(23 * width / 64.0, -height / 6.0, 0.0);
	label.setSize(net_queue_text_size[0], net_queue_text_size[1]);
	label.setPos(-depth / 2.0, height / 6.0 + net_queue_text_size[1] / 10.0,
			-width / 3.0);
};

void Net_Queue::loadPacketClasses(xmlTextReaderPtr infile, Packet* &new_packet)
/** laden eines Paketes aus XML-Datei 'infile', zur&uuml;ckgegeben in
 *  'new_packet', ruft internalLoadXMLPacket(), neue Paketklassen sollten hier
 *  eingetragen werden */
{
	internalLoadXMLPacket<Packet>(infile,
			xml_syntax_packet, new_packet);
	internalLoadXMLPacket<Ethernet_Packet>(infile,
			xml_syntax_ethernet_packet, new_packet);
	internalLoadXMLPacket<ARP_Packet>(infile,
			xml_syntax_arp_packet, new_packet);
	internalLoadXMLPacket<IPv4_Packet>(infile,
			xml_syntax_ipv4_packet, new_packet);
	internalLoadXMLPacket<TCP_Packet>(infile,
			xml_syntax_tcp_packet, new_packet);
	internalLoadXMLPacket<UDP_Packet>(infile,
			xml_syntax_udp_packet, new_packet);
	internalLoadXMLPacket<Service_Packet>(infile,
			xml_syntax_service_packet, new_packet);
	internalLoadXMLPacket<TCP_Service_Packet>(infile,
			xml_syntax_tcp_service_packet, new_packet);
	internalLoadXMLPacket<UDP_Service_Packet>(infile,
			xml_syntax_udp_service_packet, new_packet);
};

void Net_Queue::drawQueue(void) const
/** zeichnet die Schlange, also die Quader und deren Beschriftung */
{
	pos.translate(1);
	glRotated(90.0, 0, 1, 0);
	label.print(label_text.c_str());
	glRotated(-90.0, 0, 1, 0);
	left.draw();
	right.draw();
	pos.translate(-1);
};

void Net_Queue::down_in(Packet* incoming_packet, unsigned char level)	// end
/** Paket 'incoming_packet' wird f&uuml;r das Einf&uuml;gen in die
 *  Sendeschlange der Schicht 'level' (1 = Ethernet, 2 = IP usw. siehe Packet)
 *  vorbereitet: Header <= level werden an unsichtbare Stelle geschoben,
 *  andere zum Mittelpunkt 0, Paketwinkel auf 0.0, Position ans obere
 *  Ende der Schlange */
{
	unsigned char	i;
	double		packet_width, packet_height, packet_depth;
	double		posy, x_header;

	if (incoming_packet)
	{
		for(i = 4; i >= 1; i--)
		{
			x_header = incoming_packet->getHeaderXPos(i);
			// move to center pos (x_header == 0)
			incoming_packet->moveHeader(i, -x_header);
			// cout << "Net_Queue: i: " << static_cast<size_t>(i)
			//	<< " " << static_cast<size_t>(level) << endl;
			if (i <= level)
			{
				incoming_packet->moveHeader(i,
					-net_queue_packet_x_unvisible_header);
			};
		}
		incoming_packet->getSize(packet_width, packet_height,
					packet_depth);
		posy = pos.getY() + height / 2.0 + packet_height / 2.0;
		incoming_packet->setPos(pos.getX(), posy, pos.getZ());
		incoming_packet->setAngle(0.0);
	};
};

bool Net_Queue::down_out(Packet* outgoing_packet)	// begin
/** pr&uuml;ft, ob Sende-Paket 'outgoing_packet' (in der Animation) am Anfang
 *  der Schlange angekommen ist, gibt dann 'true' zur&uuml;ck, sonst 'false',
 *  wenn angekommen, ruft es outgoing_packet->clearOldTextLines() */
{
	bool	is_on_end_pos;
	double	packet_width, packet_height, packet_depth;
	double	posx, posy, posz;

	is_on_end_pos = false;
	if (outgoing_packet)
	{
		outgoing_packet->getSize(packet_width, packet_height,
						packet_depth);
		outgoing_packet->getPos(posx, posy, posz);
		posy -= packet_height / 2.0;
		// cout << "w h d: " << width << " " << height << " " << depth;
		// cout << " posy: " << posy << " needed: " << 
		//	pos.getY() - height / 2.0 << endl;
		if (posy <= pos.getY() - height / 2.0) is_on_end_pos = true;
	};
	if (is_on_end_pos) outgoing_packet->clearOldTextLines();
	return is_on_end_pos;
};

void Net_Queue::up_in(Packet* incoming_packet) 		// end of queue
/** Paket 'incoming_packet' wird f&uuml;r das Einf&uuml;gen in die
 *  Empfangsschlange der Schicht 'level' (1 = Ethernet, 2 = IP usw.
 *  siehe Packet) vorbereitet: Paketwinkel auf 0.0, Position ans untere Ende
 *  der Schlange */
{
	double	packet_width, packet_height, packet_depth, posy;

	if (incoming_packet)
	{
		incoming_packet->getSize(packet_width, packet_height,
					packet_depth);
		posy = pos.getY() - height / 2.0 + packet_height / 2.0;
		incoming_packet->setPos(pos.getX(), posy, pos.getZ());
		incoming_packet->setAngle(0.0);
	};
};

bool Net_Queue::up_out(Packet* outgoing_packet)		// begin
/** pr&uuml;ft, ob Empfangs-Paket 'outgoing_packet' (in der Animation) am
 *  Anfang der Schlange angekommen ist, gibt dann 'true' zur&uuml;ck,
 *  sonst 'false',
 *  wenn angekommen, ruft es outgoing_packet->clearOldTextLines() */
{
	bool	is_on_end_pos;
	double	packet_width, packet_height, packet_depth;
	double	posx, posy, posz;

	is_on_end_pos = false;
	if (outgoing_packet)
	{
		outgoing_packet->getSize(packet_width, packet_height,
						packet_depth);
		outgoing_packet->getPos(posx, posy, posz);
		posy -= packet_height / 2.0;
		// cout << "posy: " << posy << " needed: " << 
		//	pos.getY() - height / 2.0 << endl;
		if (posy >= pos.getY() + height / 2.0) is_on_end_pos = true;
	};
	if (is_on_end_pos) outgoing_packet->clearOldTextLines();
	return is_on_end_pos;
};

void Net_Queue::movePacket(Packet* moving_packet, char direction,
				unsigned char level)
/** verschiebt Paket 'moving_packet' in Richtung 'direction' (nur Werte '1'
 *  oder '-1') innerhalb der Schlange, in Eintrittsposition wird erst nur
 *  der Kopf 'level' verschoben, erst dann das ganze Paket <br>
 *  direction -1: down, abw&auml;rts, senden <br>
 *  direction 1: up, aufw&auml;rts, empfangen */
{
	bool	vertical, horizontal;
	double	posx, posy, posz, speed, hysterese_value, posy_middle;
	double	packet_width, packet_height, packet_depth, queue_bottom;
	double	end_pos, x_header, whole_way;

	x_header = 0.0;
	if (moving_packet)
	{
		moving_packet->getPos(posx, posy, posz);
		moving_packet->getSize(packet_width, packet_height,
						packet_depth);
		posy_middle = pos.getY() + packet_height / 2.0;
		hysterese_value = fabs(posy_middle - posy) / height * 4;
		speed = 1 / (1 + exp(hysterese_value)) * 
				animation_speed_factor;
		vertical = false; horizontal = false;
		queue_bottom = pos.getY() - height / 2.0 + packet_height / 2.0;
		posy += speed * height * direction;
		if (((posy <= queue_bottom) && (direction == -1)) ||
			((posy >= queue_bottom) && (direction == 1)))
		{
			x_header = moving_packet->getHeaderXPos(level);
			if (direction == -1) end_pos = 0.0;
			else end_pos = -width / 2.0 + packet_width / 2.0;
			if (((x_header >= end_pos) && (direction == -1))
 				|| ((x_header <= end_pos) && (direction == 1)))
					vertical = true;
			else horizontal = true;
			if (vertical && (direction == 1))
			{
				moving_packet->moveHeader(level, -x_header);
				moving_packet->moveHeader(level,
					-net_queue_packet_x_unvisible_header);
			};
			if ((direction == -1) && (x_header < -width / 2.0))
			{
				moving_packet->moveHeader(level, -x_header);
				x_header = -width / 2.0 + packet_width / 2.0;
				moving_packet->moveHeader(level, x_header);
			};
		}
		else vertical = true;
		if (vertical)
		{
			moving_packet->setPos(posx, posy, posz);
		};
		if (horizontal)
		{
			whole_way = -width / 2.0 + packet_width / 2.0; 
			hysterese_value = fabs(whole_way / 2.0 - x_header)
						/ fabs(whole_way) * 4.0;
			// cout << "Net_Queue: " << hysterese_value << endl;
			speed = 1 / (1 + exp(hysterese_value)) * 
				animation_speed_factor;
			x_header = speed * -width / 2.0 * direction;
			// cout << "Net_Queue: move header: " << x_header
			//	<< endl;
			moving_packet->moveHeader(level, x_header);
		};
		// cout << "Net_Queue::movePacket: " << speed << " " << posy;
		// cout << " middle: " << posy_middle << endl;
	};
};

