/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <iomanip>

#include "stringstream.h"

template <class ValueType>
Numeric_Variable<ValueType>::Numeric_Variable(void) : 
	min_length(0), max_length(0)
/** Konstruktor, setzt minimale und maximale L&auml;nge auf '0' */
{
};

template <class ValueType>
Numeric_Variable<ValueType>::~Numeric_Variable(void)
/** Destruktor, ist leer */
{
};

template <class ValueType>
bool Numeric_Variable<ValueType>::isConstant(void)
/** gibt 'true' zur&uuml;ck, wenn der Wert (Inhalt) der Variablen konstant
 *  ist, also unver&auml;nderbar, sonst 'false' */
{
	return constant;
};

template <class ValueType>
string Numeric_Variable<ValueType>::getName(void)
/** gibt den Namen der Variable zur&uuml;ck */
{
	return name;
};

template <class ValueType>
void Numeric_Variable<ValueType>::getValue(size_t index, 
						ValueType &current_value)
/** gibt den Wert (Inhalt) des Elementes 'index'  im Feld  der Variablen
 *  mit 'current_value' zur&uuml;ck (current_value = feld[index]) */
{
	current_value = value[index];
};

template <class ValueType>
string Numeric_Variable<ValueType>::getValue(void)
/** gibt den Wert (den Inhalt) der Variablen als String zur&uuml;ck, die
 *  einzelnen Elemente des Feldes werden durch Leerzeichen (' ') getrennt */
{
	size_t			i;
	usable_stringstream	value_buffer;
	string			value_text, all_values_text;

	all_values_text = "";
	// concat all values to a string, divided by spaces
	for(i = 0; i < value.size(); i++)
	{
		value_buffer.clear();
		value_buffer << value[i];
		value_buffer >> value_text;
		all_values_text += value_text;
		if (i < value.size() - 1) all_values_text += " ";
	};
	// cout << "Numeric_Variable::getValue: " << all_values_text << endl;
	return all_values_text;
};

template <class ValueType>
void Numeric_Variable<ValueType>::setConstant(bool is_constant)
/**
 *  - mit 'is_constant' auf 'true' wird der Wert (Inhalt) der Variablen als
 *    konstant angesehen; solange, bis es &uuml;ber diese Methode wieder
 *    ge&auml;ndert wird
 *  - mit 'is_constant' auf 'false' ist der Wert (Inhalt) innerhalb der
 *    L&auml;ngen- und Wertebereichsbeschr&auml;nkungen beliebig anpassbar */
{
	constant = is_constant;
};

template <class ValueType>
bool Numeric_Variable<ValueType>::setLength(size_t new_max, size_t new_min)
/** setzt die neue maximale 'new_max' und minimale 'new_min' L&auml;nge des
 *  Feldes, die es annehmen kann; <br>
 *  der aktuelle Inhalt wird nicht gepr&uuml;ft, nur die &Auml;nderungen
 *  &uuml;ber die Methode setValue(); <br>
 *  funktioniert auch, wenn der Inhalt als konstant markiert ist; <br>
 *  gibt 'true' zur&uuml;ck, wenn die neuen L&auml;ngen &uuml;bernommen wurden,
 *  sonst 'false' (Fehler, zur Zeit nur, wenn 'new_min > 'new_max') */
// min and max length of content array
{
	if (new_max >= new_min)
	{
		min_length = new_min;
		max_length = new_max;
		return true;
	}
	else	return false;
};

template <class ValueType>
bool Numeric_Variable<ValueType>::setMaxmin(ValueType new_max, 
						ValueType new_min)
/** setzt den Wertebereich der Variablen: alle Elemente des enthaltenen Feldes
 *  m&uuml;ssen gr&ouml;&szlig;er/gleich 'new_min' und kleiner/gleich 'new_max'
 *  sein, <br>
 *  der aktuelle Inhalt wird nicht gepr&uuml;ft, nur die &Auml;nderungen
 *  &uuml;ber die Methode setValue(); <br>
 *  funktioniert auch, wenn der Inhalt als konstant markiert ist; <br>
 *  gibt 'true' zur&uuml;ck, wenn der neue Wertebereich &uuml;bernommen wurde,
 *  sonst 'false' (Fehler, zur Zeit nur, wenn 'new_min > 'new_max') */
// min and max value of elements in content array
{
	// cout << "new_min: " << new_min << " ";
	// cout << "new_max: " << new_max << endl;
	if (new_max >= new_min)
	{
		min_value = new_min;
		max_value = new_max;
		return true;
	}
	else	return false;
}

template <class ValueType>
void Numeric_Variable<ValueType>::setName(string new_name)
/** setzt den Namen der Variable auf 'new_name' */
{
	name = new_name;
}

template <class ValueType>
bool Numeric_Variable<ValueType>::setValue(ValueType new_value)
/** setzt den Wert des ersten Feldelementes in der Variable auf 'new_value',
 *  falls die Maximall&auml;nge mindestens auf '1' gesetzt ist und 'new_value'
 *  innerhalb des vorgegebenen Wertebereichs ist, <br>
 *  gibt 'true' zur&uuml;ck, wenn &uuml;bernommen,
 *  sonst 'false' (wegen L&auml;ngenbeschr&auml;nkungen oder Wertebereich
 *  oder Variable ist konstant) */
// set value of first element in array
{
	if (constant) return false;
	if ((new_value >= min_value) && (new_value <= max_value) &&
		(max_length >= 1))
	{
		value.assign(1, new_value);
		return true;
	}
	else	return false;
}

template <class ValueType>
bool Numeric_Variable<ValueType>::setValue(size_t index, ValueType new_value)
/** setzt den Wert des Feldelementes 'index' (erlaubte Indexwerte: 0 bis
 *  Maximall&auml;nge - 1) in der Variable auf 'new_value',
 *  falls die Maximall&auml;nge mindestens auf 'index' + 1 gesetzt ist und
 *  'new_value' innerhalb des vorgegebenen Wertebereichs ist, <br>
 *  gibt 'true' zur&uuml;ck, wenn &uuml;bernommen,
 *  sonst 'false' (wegen L&auml;ngenbeschr&auml;nkungen oder Wertebereich oder
 *  Variable ist konstant) */
{
	bool	test_minmax;

	test_minmax = !constant;	// not changeable if constant == true
	if (!((new_value >= min_value) && (new_value <= max_value)))
		test_minmax = false;
	// index should be smaller than max_length
	if (index > max_length - 1) test_minmax = false;
	// set value
	if (test_minmax)
	{
		if (index >= value.size()) value.resize(index + 1);
		value[index] = new_value;
	};
	return test_minmax;
}

template <class ValueType>
bool Numeric_Variable<ValueType>::setValue(vector<ValueType> new_value)
/** &uuml;bernimmt die Werte des Feldes 'new_value' in das interne Feld,
 *  falls jedes Element des Feldes innerhalb des vorgegebenen Wertebereiches
 *  liegt und das Feld 'new_value' mindestens 'min_length' Elemente und maximal
 *  'max_length' Elemente hat (nicht zu kurz oder zu lang ist), <br>
 *  gibt 'true' zur&uuml;ck, wenn &uuml;bernommen,
 *  sonst 'false' (wegen L&auml;ngenbeschr&auml;nkungen oder Wertebereich
 *  oder Variable ist konstant) */
{
	bool	test_minmax;
	size_t	i;
	
	test_minmax = !constant;	// not changeable if constant==true
	// test each element for min_value and max_value
	for(i = 0; i < new_value.length(); i++)
		if (!((new_value[i] >= min_value) && 
			(new_value[i] <= max_value)))
				test_minmax = false;
	// right length ?
	if ((new_value.length() < min_length) || 
		(new_value.length() > max_length))
			test_minmax = false;
	// set value
	if (test_minmax) value = new_value;
	return test_minmax;
}

template <class ValueType>
bool Numeric_Variable<ValueType>::setValue(string new_value)
/** ersetzt die Elemente des internen Feldes durch die in 'new_value',
 *  diese sollten durch Leerzeichen (' ') getrennt sein und innerhalb des
 *  Wertebereiches liegen (sonst 'false' zur&uuml;ck); <br>
 *  es wird versucht, soviele Elemente einzulesen, wie das interne Feld gerade
 *  hat, f&uuml;r Elemente au&szlig;erhalb des Wertebereichs bleibt der alte
 *  Wert erhalten; <br>
 *  wenn weniger Elemente in 'new_value' sind als im internen Feld, bleiben
 *  die zus&auml;tzlichen Elemente erhalten, wenn mehr Elemente in 'new_value'
 *  als im internen Feld sind, werden die &uuml;berz&auml;hligen ignoriert,
 *  jeweils ohne Fehlermeldung; <br>
 *  gibt 'true' zur&uuml;ck, wenn alle Werte &uuml;bernommen,
 *  sonst 'false' (wegen Wertebereich oder Variable ist konstant) */
{
	bool			test_minmax;
	size_t			i, found;
	usable_stringstream	value_buffer;
	ValueType		cache;

	// cout << "Numeric_Variable::setValue(string) begin" << endl;
	test_minmax = !constant;
	if (constant) return false;	// not changeable if constant==true
	for(i = 0; i < value.size(); i++)
	{
		found = new_value.find(' ');	// search for space
		value_buffer.clear();
		value_buffer << new_value.substr(0, found); // take string part
		// cout << "new_value.substr: " << new_value.substr(0, found);
		cache = value[i];		// former value
		value_buffer >> cache;
		// cout << "cache: " << cache << endl;
		// test for min_value and max_value, then set
		if ((cache >= min_value) && (cache <= max_value))
			value[i] = cache;
		else	test_minmax = false;
		if (found + 1 == 0) new_value = ""; // not found -> end
		else new_value.erase(0, found + 1); // delete until found ' '
		// cout << "new_value: " << new_value << endl;
	};
	// cout << "Numeric_Variable::setValue(string) end" << endl;
	return test_minmax;
};

template <class ValueType>
ostream& operator << (ostream& output, Numeric_Variable<ValueType> &var)
/** gibt alle Elemente des Feldes der Variable 'var' auf dem Stream 'output'
 *  aus (ohne Trennzeichen), <br>
 *  Referenz auf den Stream wird zur&uuml;ckgegeben */
{
	size_t	i;

	for(i = 0; i < var.value.size(); i++)
		output << var.value[i];
	return output;
};

