/******************************************************************************
* This file is part of netview 
* (an program to animate TCP/IP networks in 3D with OpenGL).
*
*    Copyright (C) 2004,2005 Stefan Ziegler.
*
*    netview is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    netview is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with netview; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/
#include <iostream>

#include "variable_tree.h"

Variable_Tree::Variable_Tree(void) :
		name_was_given_out(false),
		current_var(variables.begin()), current_child(children.begin())
/** Konstruktor, Initialisierung der Variablen: name_was_given_out,
 *  current_var, current_child */
{
};

Variable_Tree::~Variable_Tree(void)
/** Destruktor, ruft nur clear() zum L&ouml;schen aller Variablen und
 *  Unterb&auml;ume */
{
	clear();
};

void Variable_Tree::clear(void)
/** l&ouml;scht alle hier gespeicherten Variablen und Unterb&auml;ume */
{
	list<Variable*>::iterator	varlist_entries;
	list<Variable_Tree*>::iterator	treelist_entries;
	
	// cout << "Variable_Tree::clear begin" << endl;
	varlist_entries = variables.begin();
	while (varlist_entries != variables.end())
	{
		if (*varlist_entries) delete *varlist_entries;
		*varlist_entries = NULL;
		varlist_entries++;
	};
	variables.erase(variables.begin(), variables.end());
	// cout << "Variable_Tree::clear children del" << endl;
	treelist_entries = children.begin();
	while (treelist_entries != children.end())
	{
		// deleting done by tree creator
		// if (*treelist_entries) delete *treelist_entries;
		*treelist_entries = NULL;
		treelist_entries++;
	};
	children.erase(children.begin(), children.end());
	// cout << "Variable_Tree::clear end" << endl;
};

void Variable_Tree::addVar(Variable* new_variable)
/** f&uuml;gt die Variable 'new_variable' in den Baum ein, falls sie noch
 *  nicht enthalten ist (Vergleich der Adresse im Zeiger) */
// adds 'new_variable' to tree
{
	bool				found;
	list<Variable*>::iterator	varlist_entries;

	found = false;
	varlist_entries = variables.begin();
	while (varlist_entries != variables.end())
	{	// is var with this address already here ?
		if (*varlist_entries == new_variable) found = true;
		varlist_entries++;
	};
	if (!found) variables.insert(variables.end(), new_variable);
	if (variables.size() == 1) current_var = variables.begin();
};

void Variable_Tree::addChild(Variable_Tree* new_child)
/** f&uuml;gt den Unterbaum 'new_child' in den Baum ein, falls er noch
 *  nicht enthalten ist (Vergleich der Adresse im Zeiger) */
// adds 'new_child' to tree
{
	bool				found;
	list<Variable_Tree*>::iterator	treelist_entries;

	found = false;
	treelist_entries = children.begin();
	while (treelist_entries != children.end())
	{	// is child with this address already here ?
		if (*treelist_entries == new_child) found = true;
		treelist_entries++;
	};
	if (!found) children.insert(children.end(), new_child);
	if (children.size() == 1) current_child = children.begin();
};

bool Variable_Tree::findVar(string varname, Variable* &found_var)
/** sucht im Baum nach Variable mit den Namen 'varname', gibt die gefundene
 *  in 'found_var' zur&uuml;ck (die letzte mit dem Namen); <br>
 *  wenn keine Variable gefunden wird, bleibt 'found_var' unver&auml;ndert;
 *  <br>gibt 'true' zur&uuml;ck, wenn gefunden, sonst 'false' */
// search for var with name 'varname'
{
	bool				found;
	list<Variable*>::iterator	varlist_entries;

	found = false;
	varlist_entries = variables.begin();
	// for all known vars
	while (varlist_entries != variables.end())
	{
		if (*varlist_entries)
			// same name ?
			if ((*varlist_entries)->getName() == varname)
			{
				found = true;
				found_var = *varlist_entries;
			};
		varlist_entries++;
	};
	return found;
};

bool Variable_Tree::findChild(string childname, Variable_Tree* &found_child)
/** sucht im Baum nach Unterbaum mit den Namen 'childname', gibt den gefundenen
 *  in 'found_child' zur&uuml;ck (den letzten mit dem Namen); <br>
 *  wenn kein Unterbaum gefunden wird, bleibt 'found_child' unver&auml;ndert;
 *  <br>gibt 'true' zur&uuml;ck, wenn gefunden, sonst 'false' */
// search for subtree with name 'childname'
{
	bool				found;
	list<Variable_Tree*>::iterator	treelist_entries;

	found = false;
	treelist_entries = children.begin();
	// for all subtrees
	while(treelist_entries != children.end())
	{
		if (*treelist_entries)
			// same name ?
			if ((*treelist_entries)->getName() == childname)
			{
				found = true;
				found_child = *treelist_entries;
			};
		treelist_entries++;
	};
	return found;
};

void Variable_Tree::gotoBegin(void)
/** gehe zum Anfang der Variablenliste zur&uuml;ck, die n&auml;chsten
 *  getNextVar() liefern wieder zuerst den Namen dieses Baums als konstante
 *  Stringvariable, dann die Variablen des eigenen Baums und dann die Namen und
 *  Variablen der Unterb&auml;ume, <br>
 *  wird jeweils auch f&uuml;r alle Unterb&auml;ume aufgerufen */
// goto begin of sequence in getNextVar
{
	list<Variable_Tree*>::iterator	treelist_entries;

	name_was_given_out = false;
	current_var = variables.begin();
	current_child = children.begin();
	treelist_entries = children.begin();
	while (treelist_entries != children.end())
	{
		if (*treelist_entries) (*treelist_entries)->gotoBegin();
		treelist_entries++;
	};

};

bool Variable_Tree::getNextVar(Variable* &next_var, size_t &depth)
/** liefert zuerst den Namen dieses Baums als konstante
 *  Stringvariable, in den folgenden Aufrufen dann die Variablen des eigenen
 *  Baums und dann die Namen und Variablen der Unterb&auml;ume
 *  (&uuml;ber deren getNextVar()) in 'next_var';<br>
 *  'depth' gibt die Baumtiefe zur&uuml;ck, aus der die Variable kommt
 *  (0: Baumname, 1: Variablen dieses Baums / Name eines Unterbaums,
 *  >= 2: Variablen der Unterb&auml;ume); <br>
 *  siehe auch gotoBegin(); <br>
 *  gibt 'true' zur&uuml;ck, wenn eine Variable 'next_var' gefunden wurde,
 *  sonst 'false' (alle Variablen schon herausgegeben) */
// give out vars in q sequence: first my tree name, then my vars, then the
// names and vars of the subtrees 
{
	bool	is_not_end, found;

	// cout << "getNextVar start" << endl;
	is_not_end = true;
	found = false; 
	if (!name_was_given_out)	// name given out as variable
	{
		next_var = &name;
		name_was_given_out = true;
		depth = 0;
	}
	// not all my variables given out
	else if (current_var != variables.end())
		while ((current_var != variables.end()) && (!found))
		{
			if (*current_var != NULL)
			{
				next_var = *current_var;
				depth = 1;
				found = true;
			};
			current_var++;
	}
	// not all subtrees asked
	else if (current_child != children.end())
	{
		found = false;
		// cout << "child: " << *current_child << endl;
		while ((current_child != children.end()) && (!found))
		{
			if (*current_child)
			{
				// ask subtree for his next var
				if ((*current_child)->getNextVar(next_var, 
									depth))
					found = true;
				else current_child++;	// next subtree
			}
			else current_child++;		// next subtree
			if (current_child != children.end())
				// goto begin of tree
				if ((*current_child) && (!found))
					(*current_child)->gotoBegin();
		};
		if (!found) is_not_end = false;
		depth++;
	}
	else is_not_end = false;
	// if (next_var) 
	//	cout << "variable name: " << next_var->getName() << endl;
	// cout << "is not end " << is_not_end << endl;
	return is_not_end;
};

size_t Variable_Tree::getNumVars(void)
/** gibt die Anzahl der Variablen dieses und der Unterb&auml;ume zur&uuml;ck,
 *  jeweils inklusive deren Namen (als eine Variable) */
// number of vars in this tree and all subtrees (with name as a var)
{
	size_t				num_vars;
	list<Variable_Tree*>::iterator	treelist_entries;

	num_vars = variables.size() + 1;	// my vars + 1 for name
	treelist_entries = children.begin();
	// ask all children for her number of vars
	while(treelist_entries != children.end())
	{
		if (*treelist_entries) 
			num_vars += (*treelist_entries)->getNumVars();
		treelist_entries++;
	};
	return num_vars;
};

string Variable_Tree::getName(void)
/** gibt den Namen dieses Baumes als String zur&uuml;ck */
{
	return name.getValue();
};

void Variable_Tree::setName(string new_name)
/** setzt den Namen dieses Baumes auf 'new_name' */
{
	name.setName("Part");
	name.setValue(new_name);
	name.setConstant(true);
};

